=== modified file 'BRANCH.TODO'
--- BRANCH.TODO	2006-08-14 16:16:53 +0000
+++ BRANCH.TODO	2006-09-13 06:45:03 +0000
@@ -1,3 +1,11 @@
 # This file is for listing TODOs for branches that are being worked on.
 # It should ALWAYS be empty in the mainline or in integration branches.
 # 
+
+Support use cases:
+ 1) bzr log over bzr+ssh to a machine without bzr 0.11 installed, without
+editing remote bashrc files.
+
+ 2) bzr log to a inet running bzr://
+
+Put both in the tutorial.

=== modified file 'HACKING'
--- HACKING	2006-08-01 18:27:42 +0000
+++ HACKING	2006-09-13 02:09:37 +0000
@@ -209,6 +209,65 @@
 > factory, then yes, foo_factory is what I would use.
 
 
+Lazy Imports
+------------
+
+To make startup time faster, we use the ``bzrlib.lazy_import`` module to
+delay importing modules until they are actually used. ``lazy_import`` uses
+the same syntax as regular python imports. So to import a few modules in a
+lazy fashion do::
+
+  from bzrlib.lazy_import import lazy_import
+  lazy_import(globals(), """
+  import os
+  import subprocess
+  import sys
+  import time
+
+  from bzrlib import (
+     errors,
+     transport,
+     foo as bar,
+     )
+  import bzrlib.transport
+  import bzrlib.xml5
+  """)
+
+At this point, all of these exist as a ``ImportReplacer`` object, ready to
+be imported once a member is accessed.
+
+
+Modules versus Members
+~~~~~~~~~~~~~~~~~~~~~~
+
+While it is possible for ``lazy_import()`` to import members of a module
+wehn using the ``from module import member`` syntax, it is recommended to
+only use that syntax to load sub modules ``from module import submodule``.
+This is because variables and classes can frequently be used without
+needing a sub-member for example::
+
+  lazy_import(globals(), """
+  from module import MyClass
+  """)
+
+  def test(x):
+      return isinstance(x, MyClass)
+
+This will incorrectly fail, because ``MyClass`` is a ``ImportReplacer``
+object, rather than the real class.
+
+
+Passing to other variables
+~~~~~~~~~~~~~~~~~~~~~~~~~~
+
+It also is incorrect to assign ``ImportReplacer`` objects to other variables.
+Because the replacer only knows about the original name, it is unable to
+replace other variables. The ``ImportReplacer`` class will raise an
+``IllegalUseOfScopeReplacer`` exception if it can figure out that this
+happened. But it requires accessing a member more than once from the new
+variable, so some bugs are not detected right away.
+
+
 Writing output
 ==============
 

=== modified file 'NEWS'
--- NEWS	2006-09-11 01:03:59 +0000
+++ NEWS	2006-09-17 21:03:04 +0000
@@ -24,8 +24,22 @@
       two branches. This is closely related to the fix for bug #39542.
       (Robert Collins)
 
+    * ``bzrlib.lazy_import.lazy_import`` function to create on-demand 
+      objects.  This allows all imports to stay at the global scope, but
+      modules will not actually be imported if they are not used.
+      (John Arbash Meinel)
+
+    * Support bzr:// and bzr+ssh:// urls to work with the new RPC-based
+      transport which will be used with the upcoming high-performance smart
+      server. The new command ``bzr serve`` will invoke bzr in server mode,
+      which processes these requests. (Andrew Bennetts, Robert Collins, Martin
+      Pool)
+
   BUG FIXES:
 
+    * Don't access e.code for generic URLErrors, only HTTPErrors have .code.
+      (Vincent Ladeuil, #59835)
+
     * Handle boundary="" lines properly to allow access through a Squid proxy.
       (John Arbash Meinel, #57723)
 
@@ -43,6 +57,18 @@
     * Don't use preexec_fn on win32, as it is not supported by subprocess.
       (John Arbash Meinel)
 
+    * Skip specific tests when the dependencies aren't met. This includes
+      some ``setup.py`` tests when ``python-dev`` is not available, and
+      some tests that depend on paramiko. (John Arbash Meinel, Mattheiu Moy)
+
+    * Fallback to Paramiko properly, if no ``ssh`` executable exists on
+      the system. (Andrew Bennetts, John Arbash Meinel)
+
+    * ``Branch.bind(other_branch)`` no longer takes a write lock on the
+      other branch, and will not push or pull between the two branches.
+      API users will need to perform a push or pull or update operation if they
+      require branch synchronisation to take place. (Robert Collins, #47344)
+
   INTERNALS:
 
     * TestCaseInTempDir now creates a separate directory for HOME, rather
@@ -102,12 +128,33 @@
       penalty. It also provides the basis for allowing readonly checkouts.
       (Robert Collins)
 
-    * ``Branch.bind(other_branch)`` no longer takes a write lock on the
-      other branch, and will not push or pull between the two branches.
-      API users will need to perform a push or pull or update operation if they
-      require branch synchronisation to take place. (Robert Collins)
-
-bzr 0.10.0RC1  2006-08-28
+    * Special case importing the standard library 'copy' module. This shaves
+      off 40ms of startup time, while retaining compatibility. See:
+      ``bzrlib/inspect_for_copy.py`` for more details. (John Arbash Meinel)
+
+    * WorkingTree has a new parent class MutableTree which represents the 
+      specialisations of Tree which are able to be altered. (Robert Collins)
+
+    * New methods mkdir and put_file_bytes_non_atomic on MutableTree that
+      mutate the tree and its contents. (Robert Collins)
+
+    * Transport behaviour at the root of the URL is now defined and tested.
+      (Andrew Bennetts, Robert Collins)
+
+  TESTING:
+
+    * New test helper classs MemoryTree. This is typically accessed via
+      ``self.make_branch_and_memory_tree()`` in test cases. (Robert Collins)
+      
+    * Add start_bzr_subprocess and stop_bzr_subprocess to allow test code to
+      continue running concurrently with a subprocess of bzr. (Andrew Bennetts,
+      Robert Collins)
+
+    * Add a new method ``Transport.get_smart_client()``. This is provided to
+      allow upgrades to a richer interface than the VFS one provided by
+      Transport. (Andrew Bennetts, Martin Pool)
+
+bzr 0.10  2006-08-29
   
   IMPROVEMENTS:
     * 'merge' now takes --uncommitted, to apply uncommitted changes from a
@@ -166,6 +213,9 @@
       in handling of revision properties. (John Arbash Meinel, Holger Krekel,
       #54723)
 
+    * The bzr selftest was failing on installed versions due to a bug in a new
+      test helper. (John Arbash Meinel, Robert Collins, #58057)
+
   INTERNALS:
 
     * ``bzrlib.cache_utf8`` contains ``encode()`` and ``decode()`` functions

=== modified file 'README'
--- README	2006-03-16 20:20:32 +0000
+++ README	2006-09-12 21:56:31 +0000
@@ -1,8 +1,8 @@
-*******************************
-Release notes for Bazaar-NG 0.8
-*******************************
+*****************************
+Release notes for Bazaar 0.10
+*****************************
 
-bazaar-ng is a distributed/decentralized version control system.
+Bazaar is a distributed/decentralized version control system.
 
 Features include:
 

=== modified file 'bzr'
--- bzr	2006-08-28 01:09:52 +0000
+++ bzr	2006-09-13 01:52:31 +0000
@@ -61,8 +61,6 @@
 
 try:
     import bzrlib
-    import bzrlib.commands
-    import bzrlib.trace
 except ImportError, e:
     sys.stderr.write("bzr: ERROR: "
                      "Couldn't import bzrlib and dependencies.\n"
@@ -70,6 +68,12 @@
                      "\n")
     raise
 
+import bzrlib.inspect_for_copy
+bzrlib.inspect_for_copy.import_copy_with_hacked_inspect()
+
+import bzrlib.commands
+import bzrlib.trace
+
 if bzrlib.version_info[:3] != (0, 11, 0):
     sys.stderr.write("bzr: WARNING: bzrlib version doesn't match the bzr program.\n"
             "This may indicate an installation problem.\n"

=== modified file 'bzrlib/add.py'
--- bzrlib/add.py	2006-08-16 15:31:34 +0000
+++ bzrlib/add.py	2006-09-12 06:30:01 +0000
@@ -296,7 +296,7 @@
                 else:
                     this_ie = inv[this_id]
 
-            for subf in os.listdir(abspath):
+            for subf in sorted(os.listdir(abspath)):
                 # here we could use TreeDirectory rather than 
                 # string concatenation.
                 subp = bzrlib.osutils.pathjoin(directory.raw_path, subf)

=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2006-09-11 01:03:59 +0000
+++ bzrlib/branch.py	2006-09-17 21:03:04 +0000
@@ -294,7 +294,7 @@
 
     def revision_history(self):
         """Return sequence of revision hashes on to this branch."""
-        raise NotImplementedError('revision_history is abstract')
+        raise NotImplementedError(self.revision_history)
 
     def revno(self):
         """Return current revision number for this branch.
@@ -1185,7 +1185,7 @@
         try:
             old_count = len(self.revision_history())
             try:
-                self.update_revisions(source,stop_revision)
+                self.update_revisions(source, stop_revision)
             except DivergedBranches:
                 if not overwrite:
                     raise

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2006-09-10 23:56:57 +0000
+++ bzrlib/builtins.py	2006-09-16 01:57:02 +0000
@@ -827,7 +827,7 @@
         tree = WorkingTree.open_containing(dir)[0]
         tree.lock_write()
         try:
-            existing_pending_merges = tree.pending_merges()
+            existing_pending_merges = tree.get_parent_ids()[1:]
             last_rev = tree.last_revision()
             if last_rev == tree.branch.last_revision():
                 # may be up to date, check master too.
@@ -839,7 +839,7 @@
             conflicts = tree.update()
             revno = tree.branch.revision_id_to_revno(tree.last_revision())
             note('Updated to revision %d.' % (revno,))
-            if tree.pending_merges() != existing_pending_merges:
+            if tree.get_parent_ids()[1:] != existing_pending_merges:
                 note('Your local commits will now show as pending merges with '
                      "'bzr status', and can be committed with 'bzr commit'.")
             if conflicts != 0:
@@ -1350,7 +1350,12 @@
         else:
             # local dir only
             # FIXME ? log the current subdir only RBC 20060203 
-            dir, relpath = bzrdir.BzrDir.open_containing('.')
+            if revision is not None \
+                    and len(revision) > 0 and revision[0].get_branch():
+                location = revision[0].get_branch()
+            else:
+                location = '.'
+            dir, relpath = bzrdir.BzrDir.open_containing(location)
             b = dir.open_branch()
 
         if revision is None:
@@ -1359,6 +1364,12 @@
         elif len(revision) == 1:
             rev1 = rev2 = revision[0].in_history(b).revno
         elif len(revision) == 2:
+            if revision[1].get_branch() != revision[0].get_branch():
+                # b is taken from revision[0].get_branch(), and
+                # show_log will use its revision_history. Having
+                # different branches will lead to weird behaviors.
+                raise BzrCommandError(
+                    "Log doesn't accept two revisions in different branches.")
             if revision[0].spec is None:
                 # missing begin-range means first revision
                 rev1 = 1
@@ -1652,6 +1663,8 @@
 
         if tree is None:
             b, relpath = Branch.open_containing(filename)
+        if revision is not None and revision[0].get_branch() is not None:
+            b = Branch.open(revision[0].get_branch())
         if revision is None:
             revision_id = b.last_revision()
         else:
@@ -2137,7 +2150,9 @@
                 else:
                     return 1
 
-        branch = self._get_remembered_parent(tree, branch, 'Merging from')
+        if revision is None \
+                or len(revision) < 1 or revision[0].needs_branch():
+            branch = self._get_remembered_parent(tree, branch, 'Merging from')
 
         if revision is None or len(revision) < 1:
             if uncommitted:
@@ -2151,6 +2166,7 @@
             if uncommitted:
                 raise BzrCommandError('Cannot use --uncommitted and --revision'
                                       ' at the same time.')
+            branch = revision[0].get_branch() or branch
             if len(revision) == 1:
                 base = [None, None]
                 other_branch, path = Branch.open_containing(branch)
@@ -2160,11 +2176,16 @@
                 assert len(revision) == 2
                 if None in revision:
                     raise BzrCommandError(
-                        "Merge doesn't permit that revision specifier.")
-                other_branch, path = Branch.open_containing(branch)
+                        "Merge doesn't permit empty revision specifier.")
+                base_branch, path = Branch.open_containing(branch)
+                branch1 = revision[1].get_branch() or branch
+                other_branch, path1 = Branch.open_containing(branch1)
+                if revision[0].get_branch() is not None:
+                    # then path was obtained from it, and is None.
+                    path = path1
 
-                base = [branch, revision[0].in_history(other_branch).revno]
-                other = [branch, revision[1].in_history(other_branch).revno]
+                base = [branch, revision[0].in_history(base_branch).revno]
+                other = [branch1, revision[1].in_history(other_branch).revno]
 
         if tree.branch.get_parent() is None or remember:
             tree.branch.set_parent(other_branch.base)
@@ -2746,6 +2767,56 @@
             pass
         
 
+class cmd_wait_until_signalled(Command):
+    """Test helper for test_start_and_stop_bzr_subprocess_send_signal.
+
+    This just prints a line to signal when it is ready, then blocks on stdin.
+    """
+
+    hidden = True
+
+    def run(self):
+        sys.stdout.write("running\n")
+        sys.stdout.flush()
+        sys.stdin.readline()
+
+
+class cmd_serve(Command):
+    """Run the bzr server.
+    """
+    takes_options = [
+        Option('inet',
+               help='serve on stdin/out for use from inetd or sshd'),
+        Option('port',
+               help='listen for connections on nominated port of the form '
+                    '[hostname:]portnumber. Passing 0 as the port number will '
+                    'result in a dynamically allocated port.',
+               type=str),
+        Option('directory',
+               help='serve contents of directory',
+               type=unicode),
+        ]
+
+    def run(self, port=None, inet=False, directory=None):
+        from bzrlib.transport import smart
+        from bzrlib.transport import get_transport
+        if directory is None:
+            directory = os.getcwd()
+        t = get_transport(directory)
+        if inet:
+            server = smart.SmartStreamServer(sys.stdin, sys.stdout, t)
+        elif port is not None:
+            if ':' in port:
+                host, port = port.split(':')
+            else:
+                host = '127.0.0.1'
+            server = smart.SmartTCPServer(t, host=host, port=int(port))
+            print 'listening on port: ', server.port
+            sys.stdout.flush()
+        else:
+            raise BzrCommandError("bzr serve requires one of --inet or --port")
+        server.serve()
+
 
 # command-line interpretation helper for merge-related commands
 def merge(other_revision, base_revision,

=== modified file 'bzrlib/bundle/serializer/__init__.py'
--- bzrlib/bundle/serializer/__init__.py	2006-09-08 18:46:29 +0000
+++ bzrlib/bundle/serializer/__init__.py	2006-09-13 10:36:03 +0000
@@ -106,6 +106,14 @@
 
 
 def _write_bundle(repository, revision_id, base_revision_id, out):
+    """Write a bundle of revisions.
+
+    :param repository: Repository containing revisions to serialize.
+    :param revision_id: Head revision_id of the bundle.
+    :param base_revision_id: Revision assumed to be present in repositories
+         applying the bundle.
+    :param out: Output file.
+    """
     if base_revision_id is NULL_REVISION:
         base_revision_id = None
     base_ancestry = set(repository.get_ancestry(base_revision_id))

=== modified file 'bzrlib/bzrdir.py'
--- bzrlib/bzrdir.py	2006-09-09 18:52:57 +0000
+++ bzrlib/bzrdir.py	2006-09-14 04:18:19 +0000
@@ -22,6 +22,12 @@
 
 # TODO: remove unittest dependency; put that stuff inside the test suite
 
+# TODO: The Format probe_transport seems a bit redundant with just trying to
+# open the bzrdir. -- mbp
+#
+# TODO: Can we move specific formats into separate modules to make this file
+# smaller?
+
 from copy import deepcopy
 from cStringIO import StringIO
 import os
@@ -298,7 +304,7 @@
         This will use the current default BzrDirFormat, and use whatever 
         repository format that that uses for bzrdirformat.create_repository.
 
-        ;param shared: Create a shared repository rather than a standalone
+        :param shared: Create a shared repository rather than a standalone
                        repository.
         The Repository object is returned.
 
@@ -319,7 +325,7 @@
         repository format that that uses for bzrdirformat.create_workingtree,
         create_branch and create_repository.
 
-        The WorkingTree object is returned.
+        :return: The WorkingTree object.
         """
         t = get_transport(safe_unicode(base))
         if not isinstance(t, LocalTransport):
@@ -466,10 +472,18 @@
         _unsupported is a private parameter to the BzrDir class.
         """
         t = get_transport(base)
-        # mutter("trying to open %r with transport %r", base, t)
-        format = BzrDirFormat.find_format(t)
+        return BzrDir.open_from_transport(t, _unsupported=_unsupported)
+
+    @staticmethod
+    def open_from_transport(transport, _unsupported=False):
+        """Open a bzrdir within a particular directory.
+
+        :param transport: Transport containing the bzrdir.
+        :param _unsupported: private.
+        """
+        format = BzrDirFormat.find_format(transport)
         BzrDir._check_supported(format, _unsupported)
-        return format.open(t, _found=True)
+        return format.open(transport, _found=True)
 
     def open_branch(self, unsupported=False):
         """Open the branch object at this BzrDir if one is present.
@@ -509,11 +523,9 @@
         url = a_transport.base
         while True:
             try:
-                format = BzrDirFormat.find_format(a_transport)
-                BzrDir._check_supported(format, False)
-                return format.open(a_transport), urlutils.unescape(a_transport.relpath(url))
+                result = BzrDir.open_from_transport(a_transport)
+                return result, urlutils.unescape(a_transport.relpath(url))
             except errors.NotBranchError, e:
-                ## mutter('not a branch in: %r %s', a_transport.base, e)
                 pass
             new_t = a_transport.clone('..')
             if new_t.base == a_transport.base:

=== modified file 'bzrlib/commit.py'
--- bzrlib/commit.py	2006-09-09 18:52:57 +0000
+++ bzrlib/commit.py	2006-09-14 04:10:40 +0000
@@ -61,6 +61,8 @@
 
 # TODO: If commit fails, leave the message in a file somewhere.
 
+# TODO: Change the parameter 'rev_id' to 'revision_id' to be consistent with
+# the rest of the code; add a deprecation of the old name.
 
 import os
 import re

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-09-09 18:52:57 +0000
+++ bzrlib/errors.py	2006-09-16 13:36:47 +0000
@@ -132,9 +132,14 @@
             if isinstance(s, unicode):
                 return s.encode('utf8')
             return s
-        except (NameError, ValueError, KeyError), e:
-            return 'Unprintable exception %s: %s' \
-                % (self.__class__.__name__, str(e))
+        except (TypeError, NameError, ValueError, KeyError), e:
+            return 'Unprintable exception %s(%r): %s' \
+                % (self.__class__.__name__,
+                   self.__dict__, str(e))
+
+
+class AlreadyBuilding(BzrNewError):
+    """The tree builder is already building a tree."""
 
 
 class BzrCheckError(BzrNewError):
@@ -191,6 +196,10 @@
         self.base = base
 
 
+class NotBuilding(BzrNewError):
+    """Not currently building a tree."""
+
+
 class NotLocalUrl(BzrNewError):
     """%(url)s is not a local path."""
     
@@ -285,6 +294,18 @@
         PathError.__init__(self, url, extra=extra)
 
 
+class ShortReadvError(PathError):
+    """readv() read %(actual)s bytes rather than %(length)s bytes at %(offset)s for %(path)s%(extra)s"""
+
+    is_user_error = False
+
+    def __init__(self, path, offset, length, actual, extra=None):
+        PathError.__init__(self, path, extra=extra)
+        self.offset = offset
+        self.length = length
+        self.actual = actual
+
+
 class PathNotChild(BzrNewError):
     """Path %(path)r is not a child of path %(base)r%(extra)s"""
 
@@ -790,6 +811,13 @@
         BzrNewError.__init__(self)
 
 
+class SmartProtocolError(TransportError):
+    """Generic bzr smart protocol error: %(details)s"""
+
+    def __init__(self, details):
+        self.details = details
+
+
 # A set of semi-meaningful errors which can be thrown
 class TransportNotPossible(TransportError):
     """Transport operation not possible: %(msg)s %(orig_error)%"""
@@ -1168,6 +1196,13 @@
         BadInventoryFormat.__init__(self, msg=msg)
 
 
+class NoSmartServer(NotBranchError):
+    """No smart server available at %(url)s"""
+
+    def __init__(self, url):
+        self.url = url
+
+
 class UnknownSSH(BzrNewError):
     """Unrecognised value for BZR_SSH environment variable: %(vendor)s"""
 
@@ -1182,3 +1217,39 @@
     def __init__(self, revision_id):
         BzrNewError.__init__(self)
         self.revision_id = revision_id
+
+
+class IllegalUseOfScopeReplacer(BzrNewError):
+    """ScopeReplacer object %(name)r was used incorrectly: %(msg)s%(extra)s"""
+
+    is_user_error = False
+
+    def __init__(self, name, msg, extra=None):
+        BzrNewError.__init__(self)
+        self.name = name
+        self.msg = msg
+        if extra:
+            self.extra = ': ' + str(extra)
+        else:
+            self.extra = ''
+
+
+class InvalidImportLine(BzrNewError):
+    """Not a valid import statement: %(msg)\n%(text)s"""
+
+    is_user_error = False
+
+    def __init__(self, text, msg):
+        BzrNewError.__init__(self)
+        self.text = text
+        self.msg = msg
+
+
+class ImportNameCollision(BzrNewError):
+    """Tried to import an object to the same name as an existing object. %(name)s"""
+
+    is_user_error = False
+
+    def __init__(self, name):
+        BzrNewError.__init__(self)
+        self.name = name

=== added file 'bzrlib/inspect_for_copy.py'
--- bzrlib/inspect_for_copy.py	1970-01-01 00:00:00 +0000
+++ bzrlib/inspect_for_copy.py	2006-09-13 01:52:31 +0000
@@ -0,0 +1,74 @@
+# Copyright (C) 2005, 2006 Canonical Development Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""A version of inspect that includes what 'copy' needs.
+
+Importing the python standard module 'copy' is far more expensive than it
+needs to be, because copy imports 'inspect' which imports 'tokenize'.
+And 'copy' only needs 2 small functions out of 'inspect', but has to
+load all of 'tokenize', which makes it horribly slow.
+
+This module is designed to use tricky hacks in import rules, to avoid this
+overhead.
+"""
+
+
+####
+# These are the only 2 functions that 'copy' needs from 'inspect'
+# As you can see, they are quite trivial, and don't justify the
+# 40ms spent to import 'inspect' because it is importing 'tokenize'
+# These are copied verbatim from the python standard library.
+
+# ----------------------------------------------------------- class helpers
+def _searchbases(cls, accum):
+    # Simulate the "classic class" search order.
+    if cls in accum:
+        return
+    accum.append(cls)
+    for base in cls.__bases__:
+        _searchbases(base, accum)
+
+
+def getmro(cls):
+    "Return tuple of base classes (including cls) in method resolution order."
+    if hasattr(cls, "__mro__"):
+        return cls.__mro__
+    else:
+        result = []
+        _searchbases(cls, result)
+        return tuple(result)
+
+
+def import_copy_with_hacked_inspect():
+    """Import the 'copy' module with a hacked 'inspect' module"""
+    # We don't actually care about 'getmro' but we need to pass
+    # something in the list so that we get the direct module,
+    # rather than getting the base module
+    import sys
+
+    # Don't hack around if 'inspect' already exists
+    if 'inspect' in sys.modules:
+        import copy
+        return
+
+    mod = __import__('bzrlib.inspect_for_copy',
+                     globals(), locals(), ['getmro'])
+
+    sys.modules['inspect'] = mod
+    try:
+        import copy
+    finally:
+        del sys.modules['inspect']

=== added file 'bzrlib/lazy_import.py'
--- bzrlib/lazy_import.py	1970-01-01 00:00:00 +0000
+++ bzrlib/lazy_import.py	2006-09-13 02:09:37 +0000
@@ -0,0 +1,379 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Functionality to create lazy evaluation objects.
+
+This includes waiting to import a module until it is actually used.
+
+Most commonly, the 'lazy_import' function is used to import other modules
+in an on-demand fashion. Typically use looks like:
+    from bzrlib.lazy_import import lazy_import
+    lazy_import(globals(), '''
+    from bzrlib import (
+        errors,
+        osutils,
+        branch,
+        )
+    import bzrlib.branch
+    ''')
+
+    Then 'errors, osutils, branch' and 'bzrlib' will exist as lazy-loaded
+    objects which will be replaced with a real object on first use.
+
+    In general, it is best to only load modules in this way. This is because
+    it isn't safe to pass these variables to other functions before they
+    have been replaced. This is especially true for constants, sometimes
+    true for classes or functions (when used as a factory, or you want
+    to inherit from them).
+"""
+
+import re
+import sys
+
+from bzrlib import (
+    errors,
+    )
+
+
+class ScopeReplacer(object):
+    """A lazy object that will replace itself in the appropriate scope.
+
+    This object sits, ready to create the real object the first time it is
+    needed.
+    """
+
+    __slots__ = ('_scope', '_factory', '_name')
+
+    def __init__(self, scope, factory, name):
+        """Create a temporary object in the specified scope.
+        Once used, a real object will be placed in the scope.
+
+        :param scope: The scope the object should appear in
+        :param factory: A callable that will create the real object.
+            It will be passed (self, scope, name)
+        :param name: The variable name in the given scope.
+        """
+        self._scope = scope
+        self._factory = factory
+        self._name = name
+        scope[name] = self
+
+    def _replace(self):
+        """Actually replace self with other in the given scope"""
+        name = object.__getattribute__(self, '_name')
+        try:
+            factory = object.__getattribute__(self, '_factory')
+            scope = object.__getattribute__(self, '_scope')
+        except AttributeError, e:
+            # Because ScopeReplacer objects only replace a single
+            # item, passing them to another variable before they are
+            # replaced would cause them to keep getting replaced
+            # (only they are replacing the wrong variable). So we
+            # make it forbidden, and try to give a good error.
+            raise errors.IllegalUseOfScopeReplacer(
+                name, msg="Object already cleaned up, did you assign it"
+                          " to another variable?",
+                extra=e)
+        obj = factory(self, scope, name)
+        scope[name] = obj
+        return obj
+
+    def _cleanup(self):
+        """Stop holding on to all the extra stuff"""
+        del self._factory
+        del self._scope
+        # We keep _name, so that we can report errors
+        # del self._name
+
+    def __getattribute__(self, attr):
+        _replace = object.__getattribute__(self, '_replace')
+        obj = _replace()
+        _cleanup = object.__getattribute__(self, '_cleanup')
+        _cleanup()
+        return getattr(obj, attr)
+
+    def __call__(self, *args, **kwargs):
+        _replace = object.__getattribute__(self, '_replace')
+        obj = _replace()
+        _cleanup = object.__getattribute__(self, '_cleanup')
+        _cleanup()
+        return obj(*args, **kwargs)
+
+
+class ImportReplacer(ScopeReplacer):
+    """This is designed to replace only a portion of an import list.
+
+    It will replace itself with a module, and then make children
+    entries also ImportReplacer objects.
+
+    At present, this only supports 'import foo.bar.baz' syntax.
+    """
+
+    # '_import_replacer_children' is intentionally a long semi-unique name
+    # that won't likely exist elsewhere. This allows us to detect an
+    # ImportReplacer object by using
+    #       object.__getattribute__(obj, '_import_replacer_children')
+    # We can't just use 'isinstance(obj, ImportReplacer)', because that
+    # accesses .__class__, which goes through __getattribute__, and triggers
+    # the replacement.
+    __slots__ = ('_import_replacer_children', '_member', '_module_path')
+
+    def __init__(self, scope, name, module_path, member=None, children={}):
+        """Upon request import 'module_path' as the name 'module_name'.
+        When imported, prepare children to also be imported.
+
+        :param scope: The scope that objects should be imported into.
+            Typically this is globals()
+        :param name: The variable name. Often this is the same as the 
+            module_path. 'bzrlib'
+        :param module_path: A list for the fully specified module path
+            ['bzrlib', 'foo', 'bar']
+        :param member: The member inside the module to import, often this is
+            None, indicating the module is being imported.
+        :param children: Children entries to be imported later.
+            This should be a map of children specifications.
+            {'foo':(['bzrlib', 'foo'], None, 
+                {'bar':(['bzrlib', 'foo', 'bar'], None {})})
+            }
+        Examples:
+            import foo => name='foo' module_path='foo',
+                          member=None, children={}
+            import foo.bar => name='foo' module_path='foo', member=None,
+                              children={'bar':(['foo', 'bar'], None, {}}
+            from foo import bar => name='bar' module_path='foo', member='bar'
+                                   children={}
+            from foo import bar, baz would get translated into 2 import
+            requests. On for 'name=bar' and one for 'name=baz'
+        """
+        if member is not None:
+            assert not children, \
+                'Cannot supply both a member and children'
+
+        self._import_replacer_children = children
+        self._member = member
+        self._module_path = module_path
+
+        # Indirecting through __class__ so that children can
+        # override _import (especially our instrumented version)
+        cls = object.__getattribute__(self, '__class__')
+        ScopeReplacer.__init__(self, scope=scope, name=name,
+                               factory=cls._import)
+
+    def _import(self, scope, name):
+        children = object.__getattribute__(self, '_import_replacer_children')
+        member = object.__getattribute__(self, '_member')
+        module_path = object.__getattribute__(self, '_module_path')
+        module_python_path = '.'.join(module_path)
+        if member is not None:
+            module = __import__(module_python_path, scope, scope, [member])
+            return getattr(module, member)
+        else:
+            module = __import__(module_python_path, scope, scope, [])
+            for path in module_path[1:]:
+                module = getattr(module, path)
+
+        # Prepare the children to be imported
+        for child_name, (child_path, child_member, grandchildren) in \
+                children.iteritems():
+            # Using self.__class__, so that children get children classes
+            # instantiated. (This helps with instrumented tests)
+            cls = object.__getattribute__(self, '__class__')
+            cls(module.__dict__, name=child_name,
+                module_path=child_path, member=child_member,
+                children=grandchildren)
+        return module
+
+
+class ImportProcessor(object):
+    """Convert text that users input into lazy import requests"""
+
+    # TODO: jam 20060912 This class is probably not strict enough about
+    #       what type of text it allows. For example, you can do:
+    #       import (foo, bar), which is not allowed by python.
+    #       For now, it should be supporting a superset of python import
+    #       syntax which is all we really care about.
+
+    __slots__ = ['imports', '_lazy_import_class']
+
+    def __init__(self, lazy_import_class=None):
+        self.imports = {}
+        if lazy_import_class is None:
+            self._lazy_import_class = ImportReplacer
+        else:
+            self._lazy_import_class = lazy_import_class
+
+    def lazy_import(self, scope, text):
+        """Convert the given text into a bunch of lazy import objects.
+
+        This takes a text string, which should be similar to normal python
+        import markup.
+        """
+        self._build_map(text)
+        self._convert_imports(scope)
+
+    def _convert_imports(self, scope):
+        # Now convert the map into a set of imports
+        for name, info in self.imports.iteritems():
+            self._lazy_import_class(scope, name=name, module_path=info[0],
+                                    member=info[1], children=info[2])
+
+    def _build_map(self, text):
+        """Take a string describing imports, and build up the internal map"""
+        for line in self._canonicalize_import_text(text):
+            if line.startswith('import '):
+                self._convert_import_str(line)
+            elif line.startswith('from '):
+                self._convert_from_str(line)
+            else:
+                raise errors.InvalidImportLine(line,
+                    "doesn't start with 'import ' or 'from '")
+
+    def _convert_import_str(self, import_str):
+        """This converts a import string into an import map.
+
+        This only understands 'import foo, foo.bar, foo.bar.baz as bing'
+
+        :param import_str: The import string to process
+        """
+        assert import_str.startswith('import ')
+        import_str = import_str[len('import '):]
+
+        for path in import_str.split(','):
+            path = path.strip()
+            if not path:
+                continue
+            as_hunks = path.split(' as ')
+            if len(as_hunks) == 2:
+                # We have 'as' so this is a different style of import
+                # 'import foo.bar.baz as bing' creates a local variable
+                # named 'bing' which points to 'foo.bar.baz'
+                name = as_hunks[1].strip()
+                module_path = as_hunks[0].strip().split('.')
+                if name in self.imports:
+                    raise errors.ImportNameCollision(name)
+                # No children available in 'import foo as bar'
+                self.imports[name] = (module_path, None, {})
+            else:
+                # Now we need to handle
+                module_path = path.split('.')
+                name = module_path[0]
+                if name not in self.imports:
+                    # This is a new import that we haven't seen before
+                    module_def = ([name], None, {})
+                    self.imports[name] = module_def
+                else:
+                    module_def = self.imports[name]
+
+                cur_path = [name]
+                cur = module_def[2]
+                for child in module_path[1:]:
+                    cur_path.append(child)
+                    if child in cur:
+                        cur = cur[child][2]
+                    else:
+                        next = (cur_path[:], None, {})
+                        cur[child] = next
+                        cur = next[2]
+
+    def _convert_from_str(self, from_str):
+        """This converts a 'from foo import bar' string into an import map.
+
+        :param from_str: The import string to process
+        """
+        assert from_str.startswith('from ')
+        from_str = from_str[len('from '):]
+
+        from_module, import_list = from_str.split(' import ')
+
+        from_module_path = from_module.split('.')
+
+        for path in import_list.split(','):
+            path = path.strip()
+            if not path:
+                continue
+            as_hunks = path.split(' as ')
+            if len(as_hunks) == 2:
+                # We have 'as' so this is a different style of import
+                # 'import foo.bar.baz as bing' creates a local variable
+                # named 'bing' which points to 'foo.bar.baz'
+                name = as_hunks[1].strip()
+                module = as_hunks[0].strip()
+            else:
+                name = module = path
+            if name in self.imports:
+                raise errors.ImportNameCollision(name)
+            self.imports[name] = (from_module_path, module, {})
+
+    def _canonicalize_import_text(self, text):
+        """Take a list of imports, and split it into regularized form.
+
+        This is meant to take regular import text, and convert it to
+        the forms that the rest of the converters prefer.
+        """
+        out = []
+        cur = None
+        continuing = False
+
+        for line in text.split('\n'):
+            line = line.strip()
+            loc = line.find('#')
+            if loc != -1:
+                line = line[:loc].strip()
+
+            if not line:
+                continue
+            if cur is not None:
+                if line.endswith(')'):
+                    out.append(cur + ' ' + line[:-1])
+                    cur = None
+                else:
+                    cur += ' ' + line
+            else:
+                if '(' in line and ')' not in line:
+                    cur = line.replace('(', '')
+                else:
+                    out.append(line.replace('(', '').replace(')', ''))
+        if cur is not None:
+            raise errors.InvalidImportLine(cur, 'Unmatched parenthesis')
+        return out
+
+
+def lazy_import(scope, text, lazy_import_class=None):
+    """Create lazy imports for all of the imports in text.
+
+    This is typically used as something like:
+    from bzrlib.lazy_import import lazy_import
+    lazy_import(globals(), '''
+    from bzrlib import (
+        foo,
+        bar,
+        baz,
+        )
+    import bzrlib.branch
+    import bzrlib.transport
+    ''')
+
+    Then 'foo, bar, baz' and 'bzrlib' will exist as lazy-loaded
+    objects which will be replaced with a real object on first use.
+
+    In general, it is best to only load modules in this way. This is
+    because other objects (functions/classes/variables) are frequently
+    used without accessing a member, which means we cannot tell they
+    have been used.
+    """
+    # This is just a helper around ImportProcessor.lazy_import
+    proc = ImportProcessor(lazy_import_class=lazy_import_class)
+    return proc.lazy_import(scope, text)

=== modified file 'bzrlib/lockdir.py'
--- bzrlib/lockdir.py	2006-09-05 21:25:15 +0000
+++ bzrlib/lockdir.py	2006-09-09 00:52:19 +0000
@@ -114,7 +114,7 @@
 from bzrlib.trace import mutter
 from bzrlib.transport import Transport
 from bzrlib.osutils import rand_chars
-from bzrlib.rio import RioWriter, read_stanza, Stanza
+from bzrlib.rio import read_stanza, Stanza
 
 # XXX: At the moment there is no consideration of thread safety on LockDir
 # objects.  This should perhaps be updated - e.g. if two threads try to take a
@@ -200,14 +200,13 @@
                 # After creating the lock directory, try again
                 self.transport.mkdir(tmpname)
 
-            sio = StringIO()
-            self._prepare_info(sio)
-            sio.seek(0)
-            # append will create a new file; we use append rather than put
-            # because we don't want to write to a temporary file and rename
-            # into place, because that's going to happen to the whole
-            # directory
-            self.transport.append_file(tmpname + self.__INFO_NAME, sio)
+            info_bytes = self._prepare_info()
+            # We use put_file_non_atomic because we just created a new unique
+            # directory so we don't have to worry about files existing there.
+            # We'll rename the whole directory into place to get atomic
+            # properties
+            self.transport.put_bytes_non_atomic(tmpname + self.__INFO_NAME,
+                                                info_bytes)
 
             self.transport.rename(tmpname, self._held_dir)
             self._lock_held = True
@@ -336,7 +335,7 @@
         except NoSuchFile, e:
             return None
 
-    def _prepare_info(self, outf):
+    def _prepare_info(self):
         """Write information about a pending lock to a temporary file.
         """
         import socket
@@ -348,7 +347,7 @@
                    nonce=self.nonce,
                    user=config.user_email(),
                    )
-        RioWriter(outf).write_stanza(s)
+        return s.to_string()
 
     def _parse_info(self, info_file):
         return read_stanza(info_file.readlines()).as_dict()

=== added file 'bzrlib/memorytree.py'
--- bzrlib/memorytree.py	1970-01-01 00:00:00 +0000
+++ bzrlib/memorytree.py	2006-09-15 02:03:15 +0000
@@ -0,0 +1,235 @@
+# Copyright (C) 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""MemoryTree object.
+
+See MemoryTree for more details.
+"""
+
+
+from copy import deepcopy
+
+from bzrlib import errors, mutabletree
+from bzrlib.decorators import needs_read_lock, needs_write_lock
+from bzrlib.osutils import sha_file
+from bzrlib.mutabletree import needs_tree_write_lock
+from bzrlib.transport.memory import MemoryTransport
+
+
+class MemoryTree(mutabletree.MutableTree):
+    """A MemoryTree is a specialisation of MutableTree.
+    
+    It maintains nearly no state outside of read_lock and write_lock
+    transactions. (it keeps a reference to the branch, and its last-revision
+    only).
+    """
+
+    def __init__(self, branch, revision_id):
+        """Construct a MemoryTree for branch using revision_id."""
+        self.branch = branch
+        self.bzrdir = branch.bzrdir
+        self._branch_revision_id = revision_id
+        self._locks = 0
+        self._lock_mode = None
+
+    @needs_tree_write_lock
+    def _add(self, files, ids, kinds):
+        """See MutableTree._add."""
+        for f, file_id, kind in zip(files, ids, kinds):
+            if kind is None:
+                kind = 'file'
+            if file_id is None:
+                self._inventory.add_path(f, kind=kind)
+            else:
+                self._inventory.add_path(f, kind=kind, file_id=file_id)
+
+    def basis_tree(self):
+        """See Tree.basis_tree()."""
+        return self._basis_tree
+
+    @staticmethod
+    def create_on_branch(branch):
+        """Create a MemoryTree for branch, using the last-revision of branch."""
+        return MemoryTree(branch, branch.last_revision())
+
+    def _gather_kinds(self, files, kinds):
+        """See MutableTree._gather_kinds.
+        
+        This implementation does not care about the file kind of
+        missing files, so is a no-op.
+        """
+
+    def get_file(self, file_id):
+        """See Tree.get_file."""
+        return self._file_transport.get(self.id2path(file_id))
+
+    def get_file_sha1(self, file_id, path=None):
+        """See Tree.get_file_sha1()."""
+        if path is None:
+            path = self.id2path(file_id)
+        stream = self._file_transport.get(path)
+        return sha_file(stream)
+
+    @needs_read_lock
+    def get_parent_ids(self):
+        """See Tree.get_parent_ids.
+
+        This implementation returns the current cached value from
+            self._parent_ids.
+        """
+        return list(self._parent_ids)
+
+    def has_filename(self, filename):
+        """See Tree.has_filename()."""
+        return self._file_transport.has(filename)
+
+    def is_executable(self, file_id, path=None):
+        return self._inventory[file_id].executable
+
+    def mkdir(self, path, file_id=None):
+        """See MutableTree.mkdir()."""
+        self.add(path, file_id, 'directory')
+        if file_id is None:
+            file_id = self.path2id(path)
+        self._file_transport.mkdir(path)
+        return file_id
+
+    @needs_read_lock
+    def last_revision(self):
+        """See MutableTree.last_revision."""
+        return self._branch_revision_id
+
+    def lock_read(self):
+        """Lock the memory tree for reading.
+
+        This triggers population of data from the branch for its revision.
+        """
+        self._locks += 1
+        try:
+            if self._locks == 1:
+                self.branch.lock_read()
+                self._lock_mode = "r"
+                self._populate_from_branch()
+        except:
+            self._locks -= 1
+            raise
+
+    def lock_tree_write(self):
+        """See MutableTree.lock_tree_write()."""
+        self._locks += 1
+        try:
+            if self._locks == 1:
+                self.branch.lock_read()
+                self._lock_mode = "w"
+                self._populate_from_branch()
+            elif self._lock_mode == "r":
+                raise errors.ReadOnlyError(self)
+        except:
+            self._locks -= 1
+            raise
+
+    def lock_write(self):
+        """See MutableTree.lock_write()."""
+        self._locks += 1
+        try:
+            if self._locks == 1:
+                self.branch.lock_write()
+                self._lock_mode = "w"
+                self._populate_from_branch()
+            elif self._lock_mode == "r":
+                raise errors.ReadOnlyError(self)
+        except:
+            self._locks -= 1
+            raise
+
+    def _populate_from_branch(self):
+        """Populate the in-tree state from the branch."""
+        self._basis_tree = self.branch.repository.revision_tree(
+            self._branch_revision_id)
+        if self._branch_revision_id is None:
+            self._parent_ids = []
+        else:
+            self._parent_ids = [self._branch_revision_id]
+        self._inventory = deepcopy(self._basis_tree._inventory)
+        self._file_transport = MemoryTransport()
+        # TODO copy the revision trees content, or do it lazy, or something.
+        inventory_entries = self._inventory.iter_entries()
+        inventory_entries.next()
+        for path, entry in inventory_entries:
+            if entry.kind == 'directory':
+                self._file_transport.mkdir(path)
+            elif entry.kind == 'file':
+                self._file_transport.put_file(path,
+                    self._basis_tree.get_file(entry.file_id))
+            else:
+                raise NotImplementedError(self._populate_from_branch)
+
+    def put_file_bytes_non_atomic(self, file_id, bytes):
+        """See MutableTree.put_file_bytes_non_atomic."""
+        self._file_transport.put_bytes(self.id2path(file_id), bytes)
+
+    def unlock(self):
+        """Release a lock.
+
+        This frees all cached state when the last lock context for the tree is
+        left.
+        """
+        if self._locks == 1:
+            self._basis_tree = None
+            self._parent_ids = []
+            self._inventory = None
+            try:
+                self.branch.unlock()
+            finally:
+                self._locks = 0
+                self._lock_mode = None
+        else:
+            self._locks -= 1
+
+    @needs_tree_write_lock
+    def unversion(self, file_ids):
+        """Remove the file ids in file_ids from the current versioned set.
+
+        When a file_id is unversioned, all of its children are automatically
+        unversioned.
+
+        :param file_ids: The file ids to stop versioning.
+        :raises: NoSuchId if any fileid is not currently versioned.
+        """
+        # XXX: This should be in mutabletree, but the inventory-save action
+        # is not relevant to memory tree. Until that is done in unlock by
+        # working tree, we cannot share the implementation.
+        for file_id in file_ids:
+            if self._inventory.has_id(file_id):
+                self._inventory.remove_recursive_id(file_id)
+            else:
+                raise errors.NoSuchId(self, file_id)
+
+    def set_parent_trees(self, parents_list, allow_leftmost_as_ghost=False):
+        """See MutableTree.set_parent_trees()."""
+        if len(parents_list) == 0:
+            self._parent_ids = []
+            self._basis_tree = self.branch.repository.revisiontree(None)
+        else:
+            if parents_list[0][1] is None and not allow_leftmost_as_ghost:
+                # a ghost in the left most parent
+                raise errors.GhostRevisionUnusableHere(parents_list[0][0])
+            self._parent_ids = [parent_id for parent_id, tree in parents_list]
+            if parents_list[0][1] is None:
+                self._basis_tree = self.branch.repository.revisiontree(None)
+            else:
+                self._basis_tree = parents_list[0][1]
+            self._branch_revision_id = parents_list[0][0]

=== added file 'bzrlib/mutabletree.py'
--- bzrlib/mutabletree.py	1970-01-01 00:00:00 +0000
+++ bzrlib/mutabletree.py	2006-09-15 02:03:15 +0000
@@ -0,0 +1,188 @@
+# Copyright (C) 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""MutableTree object.
+
+See MutableTree for more details.
+"""
+
+
+from bzrlib import tree
+from bzrlib.decorators import needs_read_lock, needs_write_lock
+from bzrlib.osutils import splitpath
+from bzrlib.symbol_versioning import DEPRECATED_PARAMETER
+
+
+def needs_tree_write_lock(unbound):
+    """Decorate unbound to take out and release a tree_write lock."""
+    def tree_write_locked(self, *args, **kwargs):
+        self.lock_tree_write()
+        try:
+            return unbound(self, *args, **kwargs)
+        finally:
+            self.unlock()
+    tree_write_locked.__doc__ = unbound.__doc__
+    tree_write_locked.__name__ = unbound.__name__
+    return tree_write_locked
+
+
+class MutableTree(tree.Tree):
+    """A MutableTree is a specialisation of Tree which is able to be mutated.
+
+    Generally speaking these mutations are only possible within a lock_write
+    context, and will revert if the lock is broken abnormally - but this cannot
+    be guaranteed - depending on the exact implementation of the mutable state.
+
+    The most common form of Mutable Tree is WorkingTree, see bzrlib.workingtree.
+    For tests we also have MemoryTree which is a MutableTree whose contents are
+    entirely in memory.
+
+    For now, we are not treating MutableTree as an interface to provide
+    conformance tests for - rather we are testing MemoryTree specifically, and 
+    interface testing implementations of WorkingTree.
+
+    A mutable tree always has an associated Branch and BzrDir object - the
+    branch and bzrdir attributes.
+    """
+
+    @needs_write_lock
+    def add(self, files, ids=None, kinds=None):
+        """Add paths to the set of versioned paths.
+
+        Note that the command line normally calls smart_add instead,
+        which can automatically recurse.
+
+        This adds the files to the inventory, so that they will be
+        recorded by the next commit.
+
+        :param files: List of paths to add, relative to the base of the tree.
+        :param ids: If set, use these instead of automatically generated ids.
+            Must be the same length as the list of files, but may
+            contain None for ids that are to be autogenerated.
+        :param kinds: Optional parameter to specify the kinds to be used for
+            each file.
+
+        TODO: Perhaps callback with the ids and paths as they're added.
+        """
+        if isinstance(files, basestring):
+            assert(ids is None or isinstance(ids, basestring))
+            assert(kinds is None or isinstance(kinds, basestring))
+            files = [files]
+            if ids is not None:
+                ids = [ids]
+            if kinds is not None:
+                kinds = [kinds]
+
+        if ids is None:
+            ids = [None] * len(files)
+        else:
+            assert(len(ids) == len(files))
+
+        if kinds is None:
+            kinds = [None] * len(files)
+        else:
+            assert(len(kinds) == len(files))
+        for f in files:
+            # generic constraint checks:
+            if self.is_control_filename(f):
+                raise errors.ForbiddenControlFileError(filename=f)
+            fp = splitpath(f)
+            if len(fp) == 0:
+                raise BzrError("cannot add top-level %r" % f)
+        # fill out file kinds for all files [not needed when we stop 
+        # caring about the instantaneous file kind within a uncommmitted tree
+        #
+        self._gather_kinds(files, kinds)
+        self._add(files, ids, kinds)
+
+    def _add(self, files, ids, kinds):
+        """Helper function for add - updates the inventory."""
+        raise NotImplementedError(self._add)
+
+    @needs_write_lock
+    def commit(self, message=None, revprops=None, *args, **kwargs):
+        # avoid circular imports
+        from bzrlib import commit
+        if revprops is None:
+            revprops = {}
+        if not 'branch-nick' in revprops:
+            revprops['branch-nick'] = self.branch.nick
+        # args for wt.commit start at message from the Commit.commit method,
+        # but with branch a kwarg now, passing in args as is results in the
+        #message being used for the branch
+        args = (DEPRECATED_PARAMETER, message, ) + args
+        committed_id = commit.Commit().commit(working_tree=self,
+            revprops=revprops, *args, **kwargs)
+        return committed_id
+
+    def _gather_kinds(self, files, kinds):
+        """Helper function for add - sets the entries of kinds."""
+        raise NotImplementedError(self._gather_kinds)
+
+    @needs_read_lock
+    def last_revision(self):
+        """Return the revision id of the last commit performed in this tree.
+
+        In early tree formats the result of last_revision is the same as the
+        branch last_revision, but that is no longer the case for modern tree
+        formats.
+        
+        last_revision returns the left most parent id, or None if there are no
+        parents.
+
+        last_revision was deprecated as of 0.11. Please use get_parent_ids
+        instead.
+        """
+        raise NotImplementedError(self.last_revision)
+
+    def lock_tree_write(self):
+        """Lock the working tree for write, and the branch for read.
+
+        This is useful for operations which only need to mutate the working
+        tree. Taking out branch write locks is a relatively expensive process
+        and may fail if the branch is on read only media. So branch write locks
+        should only be taken out when we are modifying branch data - such as in
+        operations like commit, pull, uncommit and update.
+        """
+        raise NotImplementedError(self.lock_tree_write)
+
+    def lock_write(self):
+        """Lock the tree and its branch. This allows mutating calls to be made.
+
+        Some mutating methods will take out implicit write locks, but in 
+        general you should always obtain a write lock before calling mutating
+        methods on a tree.
+        """
+        raise NotImplementedError(self.lock_write)
+
+    @needs_write_lock
+    def mkdir(self, path, file_id=None):
+        """Create a directory in the tree. if file_id is None, one is assigned.
+
+        :param path: A unicode file path.
+        :param file_id: An optional file-id.
+        :return: the file id of the new directory.
+        """
+        raise NotImplementedError(self.mkdir)
+
+    def set_parent_trees(self, parents_list, allow_leftmost_as_ghost=False):
+        """Set the parents of the working tree.
+
+        :param parents_list: A list of (revision_id, tree) tuples. 
+            If tree is None, then that element is treated as an unreachable
+            parent tree - i.e. a ghost.
+        """
+        raise NotImplementedError(self.set_parent_trees)

=== modified file 'bzrlib/osutils.py'
--- bzrlib/osutils.py	2006-09-08 18:52:17 +0000
+++ bzrlib/osutils.py	2006-09-12 21:25:01 +0000
@@ -1052,7 +1052,7 @@
         _cached_user_encoding = locale.getpreferredencoding()
     except locale.Error, e:
         sys.stderr.write('bzr: warning: %s\n'
-                         '  Could not what text encoding to use.\n'
+                         '  Could not determine what text encoding to use.\n'
                          '  This error usually means your Python interpreter\n'
                          '  doesn\'t support the locale set by $LANG (%s)\n'
                          "  Continuing with ascii encoding.\n"

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2006-09-10 02:18:31 +0000
+++ bzrlib/repository.py	2006-09-14 04:10:40 +0000
@@ -1739,6 +1739,18 @@
         raise NotImplementedError(self.copy_content)
 
     def fetch(self, revision_id=None, pb=None):
+        """Fetch the content required to construct revision_id.
+
+        The content is copied from self.source to self.target.
+
+        :param revision_id: if None all content is copied, if NULL_REVISION no
+                            content is copied.
+        :param pb: optional progress bar to use for progress reports. If not
+                   provided a default one will be created.
+
+        Returns the copied revision count and the failed revisions in a tuple:
+        (copied, failures).
+        """
         raise NotImplementedError(self.fetch)
    
     @needs_read_lock
@@ -1811,18 +1823,7 @@
 
     @needs_write_lock
     def fetch(self, revision_id=None, pb=None):
-        """Fetch the content required to construct revision_id.
-
-        The content is copied from source to target.
-
-        :param revision_id: if None all content is copied, if NULL_REVISION no
-                            content is copied.
-        :param pb: optional progress bar to use for progress reports. If not
-                   provided a default one will be created.
-
-        Returns the copied revision count and the failed revisions in a tuple:
-        (copied, failures).
-        """
+        """See InterRepository.fetch()."""
         from bzrlib.fetch import GenericRepoFetcher
         mutter("Using fetch logic to copy between %s(%s) and %s(%s)",
                self.source, self.source._format, self.target, 

=== modified file 'bzrlib/revisionspec.py'
--- bzrlib/revisionspec.py	2006-08-28 16:28:10 +0000
+++ bzrlib/revisionspec.py	2006-09-03 15:58:15 +0000
@@ -224,6 +224,13 @@
         """
         return True
 
+    def get_branch(self):
+        """When the revision specifier contains a branch location, return it.
+        
+        Otherwise, return None.
+        """
+        return None
+
 
 # private API
 
@@ -273,6 +280,12 @@
     def needs_branch(self):
         return self.spec.find(':') == -1
 
+    def get_branch(self):
+        if self.spec.find(':') == -1:
+            return None
+        else:
+            return self.spec[self.spec.find(':')+1:]
+
 # Old compatibility 
 RevisionSpec_int = RevisionSpec_revno
 

=== modified file 'bzrlib/revisiontree.py'
--- bzrlib/revisiontree.py	2006-08-15 03:09:14 +0000
+++ bzrlib/revisiontree.py	2006-09-07 23:13:08 +0000
@@ -48,7 +48,11 @@
 
         A RevisionTree's parents match the revision graph.
         """
-        parent_ids = self._repository.get_revision(self._revision_id).parent_ids
+        if self._revision_id not in (None, 'null:'):
+            parent_ids = self._repository.get_revision(
+                self._revision_id).parent_ids
+        else:
+            parent_ids = []
         return parent_ids
         
     def get_revision_id(self):

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2006-09-09 15:39:00 +0000
+++ bzrlib/tests/__init__.py	2006-09-16 13:36:47 +0000
@@ -43,6 +43,7 @@
 import time
 
 
+from bzrlib import memorytree
 import bzrlib.branch
 import bzrlib.bzrdir as bzrdir
 import bzrlib.commands
@@ -94,7 +95,8 @@
                       bzrlib.merge3,
                       bzrlib.option,
                       bzrlib.osutils,
-                      bzrlib.store
+                      bzrlib.store,
+                      bzrlib.transport,
                       ]
 
 
@@ -886,7 +888,40 @@
         :param universal_newlines: Convert CRLF => LF
         """
         env_changes = kwargs.get('env_changes', {})
-
+        process = self.start_bzr_subprocess(args, env_changes=env_changes)
+        # We distinguish between retcode=None and retcode not passed.
+        supplied_retcode = kwargs.get('retcode', 0)
+        return self.finish_bzr_subprocess(process, retcode=supplied_retcode,
+            universal_newlines=kwargs.get('universal_newlines', False),
+            process_args=args)
+
+    def start_bzr_subprocess(self, process_args, env_changes=None,
+                             skip_if_plan_to_signal=False):
+        """Start bzr in a subprocess for testing.
+
+        This starts a new Python interpreter and runs bzr in there.
+        This should only be used for tests that have a justifiable need for
+        this isolation: e.g. they are testing startup time, or signal
+        handling, or early startup code, etc.  Subprocess code can't be
+        profiled or debugged so easily.
+
+        :param process_args: a list of arguments to pass to the bzr executable,
+            for example `['--version']`.
+        :param env_changes: A dictionary which lists changes to environment
+            variables. A value of None will unset the env variable.
+            The values must be strings. The change will only occur in the
+            child, so you don't need to fix the environment after running.
+        :param skip_if_plan_to_signal: raise TestSkipped when true and os.kill
+            is not available.
+
+        :returns: Popen object for the started process.
+        """
+        if skip_if_plan_to_signal:
+            if not getattr(os, 'kill', None):
+                raise TestSkipped("os.kill not available.")
+
+        if env_changes is None:
+            env_changes = {}
         old_env = {}
 
         def cleanup_environment():
@@ -898,29 +933,47 @@
                 osutils.set_or_unset_env(env_var, value)
 
         bzr_path = os.path.dirname(os.path.dirname(bzrlib.__file__))+'/bzr'
-        args = list(args)
+        if not os.path.isfile(bzr_path):
+            # We are probably installed. Assume sys.argv is the right file
+            bzr_path = sys.argv[0]
 
         try:
             # win32 subprocess doesn't support preexec_fn
             # so we will avoid using it on all platforms, just to
             # make sure the code path is used, and we don't break on win32
             cleanup_environment()
-            process = Popen([sys.executable, bzr_path]+args,
-                             stdout=PIPE, stderr=PIPE)
+            process = Popen([sys.executable, bzr_path] + list(process_args),
+                             stdin=PIPE, stdout=PIPE, stderr=PIPE)
         finally:
             restore_environment()
-            
-        out = process.stdout.read()
-        err = process.stderr.read()
-
-        if kwargs.get('universal_newlines', False):
+        return process
+
+    def finish_bzr_subprocess(self, process, retcode=0, send_signal=None,
+                              universal_newlines=False, process_args=None):
+        """Finish the execution of process.
+
+        :param process: the Popen object returned from start_bzr_subprocess.
+        :param retcode: The status code that is expected.  Defaults to 0.  If
+            None is supplied, the status code is not checked.
+        :param send_signal: an optional signal to send to the process.
+        :param universal_newlines: Convert CRLF => LF
+        :returns: (stdout, stderr)
+        """
+        if send_signal is not None:
+            os.kill(process.pid, send_signal)
+        out, err = process.communicate()
+
+        if universal_newlines:
             out = out.replace('\r\n', '\n')
             err = err.replace('\r\n', '\n')
 
-        retcode = process.wait()
-        supplied_retcode = kwargs.get('retcode', 0)
-        if supplied_retcode is not None:
-            assert supplied_retcode == retcode
+        if retcode is not None and retcode != process.returncode:
+            if process_args is None:
+                process_args = "(unknown args)"
+            mutter('Output of bzr %s:\n%s', process_args, out)
+            mutter('Error for bzr %s:\n%s', process_args, err)
+            self.fail('Command bzr %s failed with retcode %s != %s'
+                      % (process_args, retcode, process.returncode))
         return [out, err]
 
     def check_inventory_shape(self, inv, shape):
@@ -1202,7 +1255,7 @@
         return self.__server
 
     def get_url(self, relpath=None):
-        """Get a URL for the readwrite transport.
+        """Get a URL (or maybe a path) for the readwrite transport.
 
         This will either be backed by '.' or to an equivalent non-file based
         facility.
@@ -1213,7 +1266,14 @@
         if relpath is not None and relpath != '.':
             if not base.endswith('/'):
                 base = base + '/'
-            base = base + urlutils.escape(relpath)
+            # XXX: Really base should be a url; we did after all call
+            # get_url()!  But sometimes it's just a path (from
+            # LocalAbspathServer), and it'd be wrong to append urlescaped data
+            # to a non-escaped local path.
+            if base.startswith('./') or base.startswith('/'):
+                base += relpath
+            else:
+                base += urlutils.escape(relpath)
         return base
 
     def get_transport(self):
@@ -1239,20 +1299,18 @@
 
     def make_bzrdir(self, relpath, format=None):
         try:
-            url = self.get_url(relpath)
-            mutter('relpath %r => url %r', relpath, url)
-            segments = url.split('/')
-            if segments and segments[-1] not in ('', '.'):
-                parent = '/'.join(segments[:-1])
-                t = get_transport(parent)
+            # might be a relative or absolute path
+            maybe_a_url = self.get_url(relpath)
+            segments = maybe_a_url.rsplit('/', 1)
+            t = get_transport(maybe_a_url)
+            if len(segments) > 1 and segments[-1] not in ('', '.'):
                 try:
-                    t.mkdir(segments[-1])
+                    t.mkdir('.')
                 except errors.FileExists:
                     pass
             if format is None:
-                format=bzrlib.bzrdir.BzrDirFormat.get_default_format()
-            # FIXME: make this use a single transport someday. RBC 20060418
-            return format.initialize_on_transport(get_transport(relpath))
+                format = bzrlib.bzrdir.BzrDirFormat.get_default_format()
+            return format.initialize_on_transport(t)
         except errors.UninitializableFormat:
             raise TestSkipped("Format %s is not initializable." % format)
 
@@ -1261,10 +1319,25 @@
         made_control = self.make_bzrdir(relpath, format=format)
         return made_control.create_repository(shared=shared)
 
+    def make_branch_and_memory_tree(self, relpath):
+        """Create a branch on the default transport and a MemoryTree for it."""
+        b = self.make_branch(relpath)
+        return memorytree.MemoryTree.create_on_branch(b)
+
     def make_branch_and_tree(self, relpath, format=None):
         """Create a branch on the transport and a tree locally.
 
-        Returns the tree.
+        If the transport is not a LocalTransport, the Tree can't be created on
+        the transport.  In that case the working tree is created in the local
+        directory, and the returned tree's branch and repository will also be
+        accessed locally.
+
+        This will fail if the original default transport for this test
+        case wasn't backed by the working directory, as the branch won't
+        be on disk for us to open it.  
+
+        :param format: The BzrDirFormat.
+        :returns: the WorkingTree.
         """
         # TODO: always use the local disk path for the working tree,
         # this obviously requires a format that supports branch references
@@ -1274,11 +1347,14 @@
         try:
             return b.bzrdir.create_workingtree()
         except errors.NotLocalUrl:
-            # new formats - catch No tree error and create
-            # a branch reference and a checkout.
-            # old formats at that point - raise TestSkipped.
-            # TODO: rbc 20060208
-            return WorkingTreeFormat2().initialize(bzrdir.BzrDir.open(relpath))
+            # We can only make working trees locally at the moment.  If the
+            # transport can't support them, then reopen the branch on a local
+            # transport, and create the working tree there.  
+            #
+            # Possibly we should instead keep
+            # the non-disk-backed branch and create a local checkout?
+            bd = bzrdir.BzrDir.open(relpath)
+            return bd.create_workingtree()
 
     def assertIsDirectory(self, relpath, transport):
         """Assert that relpath within transport is a directory.
@@ -1405,6 +1481,7 @@
                    'bzrlib.tests.test_errors',
                    'bzrlib.tests.test_escaped_store',
                    'bzrlib.tests.test_fetch',
+                   'bzrlib.tests.test_ftp_transport',
                    'bzrlib.tests.test_gpg',
                    'bzrlib.tests.test_graph',
                    'bzrlib.tests.test_hashcache',
@@ -1414,9 +1491,11 @@
                    'bzrlib.tests.test_ignores',
                    'bzrlib.tests.test_inv',
                    'bzrlib.tests.test_knit',
+                   'bzrlib.tests.test_lazy_import',
                    'bzrlib.tests.test_lockdir',
                    'bzrlib.tests.test_lockable_files',
                    'bzrlib.tests.test_log',
+                   'bzrlib.tests.test_memorytree',
                    'bzrlib.tests.test_merge',
                    'bzrlib.tests.test_merge3',
                    'bzrlib.tests.test_merge_core',
@@ -1441,8 +1520,8 @@
                    'bzrlib.tests.test_selftest',
                    'bzrlib.tests.test_setup',
                    'bzrlib.tests.test_sftp_transport',
-                   'bzrlib.tests.test_ftp_transport',
                    'bzrlib.tests.test_smart_add',
+                   'bzrlib.tests.test_smart_transport',
                    'bzrlib.tests.test_source',
                    'bzrlib.tests.test_status',
                    'bzrlib.tests.test_store',
@@ -1455,6 +1534,7 @@
                    'bzrlib.tests.test_transform',
                    'bzrlib.tests.test_transport',
                    'bzrlib.tests.test_tree',
+                   'bzrlib.tests.test_treebuilder',
                    'bzrlib.tests.test_tsort',
                    'bzrlib.tests.test_tuned_gzip',
                    'bzrlib.tests.test_ui',

=== modified file 'bzrlib/tests/blackbox/__init__.py'
--- bzrlib/tests/blackbox/__init__.py	2006-08-25 15:01:52 +0000
+++ bzrlib/tests/blackbox/__init__.py	2006-09-13 06:45:03 +0000
@@ -79,6 +79,7 @@
                      'bzrlib.tests.blackbox.test_revision_history',
                      'bzrlib.tests.blackbox.test_revision_info',
                      'bzrlib.tests.blackbox.test_selftest',
+                     'bzrlib.tests.blackbox.test_serve',
                      'bzrlib.tests.blackbox.test_shared_repository',
                      'bzrlib.tests.blackbox.test_sign_my_commits',
                      'bzrlib.tests.blackbox.test_status',

=== modified file 'bzrlib/tests/blackbox/test_cat.py'
--- bzrlib/tests/blackbox/test_cat.py	2006-07-28 16:05:23 +0000
+++ bzrlib/tests/blackbox/test_cat.py	2006-09-11 06:07:02 +0000
@@ -31,6 +31,8 @@
         def bzr(*args, **kwargs):
             return self.run_bzr(*args, **kwargs)[0]
 
+        os.mkdir('branch')
+        os.chdir('branch')
         bzr('init')
         open('a', 'wb').write('foo\n')
         bzr('add', 'a')
@@ -51,4 +53,11 @@
         rev_id = bzr('revision-history').strip().split('\n')[-1]
 
         self.assertEquals(bzr('cat', 'a', '-r', 'revid:%s' % rev_id), 'baz\n')
-
+        
+        os.chdir('..')
+        
+        self.assertEquals(bzr('cat', 'branch/a', '-r', 'revno:1:branch'),
+                          'foo\n')
+        bzr('cat', 'a', retcode=3)
+        bzr('cat', 'a', '-r', 'revno:1:branch-that-does-not-exist', retcode=3)
+        

=== modified file 'bzrlib/tests/blackbox/test_locale.py'
--- bzrlib/tests/blackbox/test_locale.py	2006-09-09 15:39:00 +0000
+++ bzrlib/tests/blackbox/test_locale.py	2006-09-12 21:25:01 +0000
@@ -65,7 +65,7 @@
         # as the first part of the string. It may be a little tempermental
         self.assertEqualDiff("""\
 bzr: warning: unsupported locale setting
-  Could not what text encoding to use.
+  Could not determine what text encoding to use.
   This error usually means your Python interpreter
   doesn't support the locale set by $LANG (BOGUS)
   Continuing with ascii encoding.

=== modified file 'bzrlib/tests/blackbox/test_log.py'
--- bzrlib/tests/blackbox/test_log.py	2006-07-28 16:05:23 +0000
+++ bzrlib/tests/blackbox/test_log.py	2006-08-06 17:31:40 +0000
@@ -89,6 +89,25 @@
         log = self.runbzr("log -r 1..3")[0]
         self.assertEquals(self.full_log, log)
 
+    def test_log_revno_n_path(self):
+        os.mkdir('branch1')
+        os.chdir('branch1')
+        self._prepare()
+        os.chdir('..')
+        os.mkdir('branch2')
+        os.chdir('branch2')
+        self._prepare()
+        os.chdir('..')
+        log = self.runbzr("log -r revno:2:branch1..revno:3:branch2",
+                          retcode=3)[0]
+        log = self.runbzr("log -r revno:1:branch2..revno:3:branch2")[0]
+        self.assertEquals(self.full_log, log)
+        log = self.runbzr("log -r revno:1:branch2")[0]
+        self.assertTrue('revno: 1\n' in log)
+        self.assertTrue('revno: 2\n' not in log)
+        self.assertTrue('branch nick: branch2\n' in log)
+        self.assertTrue('branch nick: branch1\n' not in log)
+        
 
 class TestLogMerges(ExternalBase):
 

=== modified file 'bzrlib/tests/blackbox/test_merge.py'
--- bzrlib/tests/blackbox/test_merge.py	2006-09-05 08:14:23 +0000
+++ bzrlib/tests/blackbox/test_merge.py	2006-09-11 09:34:51 +0000
@@ -52,6 +52,7 @@
         os.mkdir('a')
         os.chdir('a')
         self.example_branch()
+        ancestor = Branch.open('.').revno()
         os.chdir('..')
         self.runbzr('branch a b')
         os.chdir('b')
@@ -79,6 +80,17 @@
         b_tip = b.last_revision()
         self.failUnless(a.branch.repository.has_revision(b_tip))
         self.assertEqual([a_tip, b_tip], a.get_parent_ids())
+        self.runbzr('revert --no-backup')
+        out, err = self.runbzr('merge -r revno:1:./hello', retcode=3)
+        self.assertTrue("Not a branch" in err)
+        self.runbzr('merge -r revno:%d:./..revno:%d:../b'
+                    %(ancestor,b.revno()))
+        self.assertEquals(a.pending_merges(), [b.last_revision()])
+        self.check_file_contents('goodbye', 'quux')
+        self.runbzr('revert --no-backup')
+        self.runbzr('merge -r revno:%d:../b'%b.revno())
+        self.assertEquals(a.pending_merges(),
+                          [b.last_revision()])
         a_tip = a.commit('merged')
         self.runbzr('merge ../b -r last:1')
         self.assertEqual([a_tip], a.get_parent_ids())

=== modified file 'bzrlib/tests/blackbox/test_selftest.py'
--- bzrlib/tests/blackbox/test_selftest.py	2006-09-07 20:29:07 +0000
+++ bzrlib/tests/blackbox/test_selftest.py	2006-09-15 02:04:56 +0000
@@ -16,6 +16,7 @@
 """UI tests for the test framework."""
 
 import os
+import signal
 import sys
 
 import bzrlib
@@ -243,6 +244,50 @@
         self.assertEqual('it sure does!\n', out)
         self.assertEqual('', err)
 
+    def test_start_and_stop_bzr_subprocess(self):
+        """We can start and perform other test actions while that process is
+        still alive.
+        """
+        process = self.start_bzr_subprocess(['--version'])
+        result = self.finish_bzr_subprocess(process)
+        self.assertContainsRe(result[0], 'is free software')
+        self.assertEqual('', result[1])
+
+    def test_start_and_stop_bzr_subprocess_with_error(self):
+        """finish_bzr_subprocess allows specification of the desired exit code.
+        """
+        process = self.start_bzr_subprocess(['--versionn'])
+        result = self.finish_bzr_subprocess(process, retcode=3)
+        self.assertEqual('', result[0])
+        self.assertContainsRe(result[1], 'unknown command')
+
+    def test_start_and_stop_bzr_subprocess_ignoring_retcode(self):
+        """finish_bzr_subprocess allows the exit code to be ignored."""
+        process = self.start_bzr_subprocess(['--versionn'])
+        result = self.finish_bzr_subprocess(process, retcode=None)
+        self.assertEqual('', result[0])
+        self.assertContainsRe(result[1], 'unknown command')
+
+    def test_start_and_stop_bzr_subprocess_with_unexpected_retcode(self):
+        """finish_bzr_subprocess raises self.failureException if the retcode is
+        not the expected one.
+        """
+        process = self.start_bzr_subprocess(['--versionn'])
+        self.assertRaises(self.failureException, self.finish_bzr_subprocess,
+                          process, retcode=0)
+        
+    def test_start_and_stop_bzr_subprocess_send_signal(self):
+        """finish_bzr_subprocess raises self.failureException if the retcode is
+        not the expected one.
+        """
+        process = self.start_bzr_subprocess(['wait-until-signalled'],
+                                            skip_if_plan_to_signal=True)
+        self.assertEqual('running\n', process.stdout.readline())
+        result = self.finish_bzr_subprocess(process, send_signal=signal.SIGINT,
+                                            retcode=3)
+        self.assertEqual('', result[0])
+        self.assertEqual('bzr: interrupted\n', result[1])
+        
 
 class TestRunBzrError(ExternalBase):
 

=== added file 'bzrlib/tests/blackbox/test_serve.py'
--- bzrlib/tests/blackbox/test_serve.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/blackbox/test_serve.py	2006-09-15 05:19:28 +0000
@@ -0,0 +1,107 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+
+"""Tests of the bzr serve command."""
+
+import signal
+
+from bzrlib.branch import Branch
+from bzrlib.bzrdir import BzrDir
+from bzrlib.tests import TestCaseWithTransport
+from bzrlib.transport import smart
+
+
+class DoesNotCloseStdOutClient(smart.SmartStreamClient):
+    """A client that doesn't close stdout upon disconnect().
+    
+    We wish to let stdout remain open so that we can see if the server writes
+    anything to stdout during its shutdown.
+    """
+
+    def disconnect(self):
+        if self._connected:
+            self._connected = False
+            # The client's out is the server's in.
+            self._out.close()
+
+
+class TestBzrServe(TestCaseWithTransport):
+
+    def test_bzr_serve_inet(self):
+        # Make a branch
+        self.make_branch('.')
+
+        # Serve that branch from the current directory
+        process = self.start_bzr_subprocess(['serve', '--inet'])
+
+        # Connect to the server
+        # We use this url because while this is no valid URL to connect to this
+        # server instance, the transport needs a URL.
+        client = DoesNotCloseStdOutClient(
+            lambda: (process.stdout, process.stdin))
+        transport = smart.SmartTransport('bzr://localhost/', client=client)
+
+        # We get a working branch
+        branch = BzrDir.open_from_transport(transport).open_branch()
+        branch.repository.get_revision_graph()
+        self.assertEqual(None, branch.last_revision())
+
+        # finish with the transport
+        del transport
+        # Disconnect the client forcefully JUST IN CASE because of __del__'s use
+        # in the smart module.
+        client.disconnect()
+
+        # Shutdown the server: the client should have disconnected cleanly and
+        # closed stdin, so the server process should shut itself down.
+        self.assertTrue(process.stdin.closed)
+        # Hide stdin from the subprocess module, so it won't fail to close it.
+        process.stdin = None
+        result = self.finish_bzr_subprocess(process, retcode=0)
+        self.assertEqual('', result[0])
+        self.assertEqual('', result[1])
+    
+    def test_bzr_serve_port(self):
+        # Make a branch
+        self.make_branch('.')
+
+        # Serve that branch from the current directory
+        process = self.start_bzr_subprocess(['serve', '--port', 'localhost:0'],
+                                            skip_if_plan_to_signal=True)
+        port_line = process.stdout.readline()
+        prefix = 'listening on port: '
+        self.assertStartsWith(port_line, prefix)
+        port = int(port_line[len(prefix):])
+
+        # Connect to the server
+        branch = Branch.open('bzr://localhost:%d/' % port)
+
+        # We get a working branch
+        branch.repository.get_revision_graph()
+        self.assertEqual(None, branch.last_revision())
+
+        # Shutdown the server
+        result = self.finish_bzr_subprocess(process, retcode=3,
+                                            send_signal=signal.SIGINT)
+        self.assertEqual('', result[0])
+        self.assertEqual('bzr: interrupted\n', result[1])
+
+    def test_bzr_serve_no_args(self):
+        """'bzr serve' with no arguments or options should not traceback."""
+        out, err = self.run_bzr_error(
+            ['bzr serve requires one of --inet or --port'], 'serve')
+

=== modified file 'bzrlib/tests/repository_implementations/test_repository.py'
--- bzrlib/tests/repository_implementations/test_repository.py	2006-09-09 18:52:57 +0000
+++ bzrlib/tests/repository_implementations/test_repository.py	2006-09-15 01:21:38 +0000
@@ -556,3 +556,29 @@
     def test_corrupt_revision_get_revision_reconcile(self):
         repo = repository.Repository.open('inventory_with_unnecessary_ghost')
         repo.get_revision_reconcile('ghost')
+
+
+class TestEscaping(TestCaseWithTransport):
+    """Test that repositories can be stored correctly on VFAT transports.
+    
+    Makes sure we have proper escaping of invalid characters, etc.
+
+    It'd be better to test all operations on the FakeVFATTransportDecorator,
+    but working trees go straight to the os not through the Transport layer.
+    Therefore we build some history first in the regular way and then 
+    check it's safe to access for vfat.
+    """
+
+    def test_on_vfat(self):
+        FOO_ID = 'foo<:>ID'
+        REV_ID = 'revid-1'
+        wt = self.make_branch_and_tree('repo')
+        self.build_tree(["repo/foo"])
+        # add file with id containing wierd characters
+        wt.add(['foo'], [FOO_ID])
+        wt.commit('this is my new commit', rev_id=REV_ID)
+        # now access over vfat; should be safe
+        branch = bzrdir.BzrDir.open('vfat+' + self.get_url('repo')).open_branch()
+        revtree = branch.repository.revision_tree(REV_ID)
+        contents = revtree.get_file_text(FOO_ID)
+        self.assertEqual(contents, 'contents of repo/foo\n')

=== modified file 'bzrlib/tests/test_branch.py'
--- bzrlib/tests/test_branch.py	2006-09-05 20:08:52 +0000
+++ bzrlib/tests/test_branch.py	2006-09-07 23:31:28 +0000
@@ -77,43 +77,6 @@
             branch.unlock()
 
 
-class TestBranchEscaping(TestCaseWithTransport):
-    """Test a branch can be correctly stored and used on a vfat-like transport
-    
-    Makes sure we have proper escaping of invalid characters, etc.
-
-    It'd be better to test all operations on the FakeVFATTransportDecorator,
-    but working trees go straight to the os not through the Transport layer.
-    Therefore we build some history first in the regular way and then 
-    check it's safe to access for vfat.
-    """
-
-    FOO_ID = 'foo<:>ID'
-    REV_ID = 'revid-1'
-
-    def setUp(self):
-        super(TestBranchEscaping, self).setUp()
-        from bzrlib.repository import RepositoryFormatKnit1
-        bzrdir = BzrDirMetaFormat1().initialize(self.get_url())
-        repo = RepositoryFormatKnit1().initialize(bzrdir)
-        branch = bzrdir.create_branch()
-        wt = bzrdir.create_workingtree()
-        self.build_tree_contents([("foo", "contents of foo")])
-        # add file with id containing wierd characters
-        wt.add(['foo'], [self.FOO_ID])
-        wt.commit('this is my new commit', rev_id=self.REV_ID)
-
-    def test_branch_on_vfat(self):
-        from bzrlib.transport.fakevfat import FakeVFATTransportDecorator
-        # now access over vfat; should be safe
-        transport = FakeVFATTransportDecorator('vfat+' + self.get_url())
-        bzrdir, junk = BzrDir.open_containing_from_transport(transport)
-        branch = bzrdir.open_branch()
-        revtree = branch.repository.revision_tree(self.REV_ID)
-        contents = revtree.get_file_text(self.FOO_ID)
-        self.assertEqual(contents, 'contents of foo')
-
-
 class SampleBranchFormat(bzrlib.branch.BranchFormat):
     """A sample format
 

=== modified file 'bzrlib/tests/test_bundle.py'
--- bzrlib/tests/test_bundle.py	2006-08-17 14:16:46 +0000
+++ bzrlib/tests/test_bundle.py	2006-09-07 23:13:08 +0000
@@ -19,7 +19,7 @@
 import sys
 import tempfile
 
-from bzrlib import inventory
+from bzrlib import inventory, treebuilder
 from bzrlib.builtins import merge
 from bzrlib.bzrdir import BzrDir
 from bzrlib.bundle.apply_bundle import install_bundle, merge_bundle
@@ -785,11 +785,14 @@
         bundle = self.get_valid_bundle(None, 'white-4')
 
     def test_alt_timezone_bundle(self):
-        self.tree1 = self.make_branch_and_tree('b1')
+        self.tree1 = self.make_branch_and_memory_tree('b1')
         self.b1 = self.tree1.branch
+        builder = treebuilder.TreeBuilder()
 
-        self.build_tree(['b1/newfile'])
-        self.tree1.add(['newfile'])
+        self.tree1.lock_write()
+        builder.start_tree(self.tree1)
+        builder.build(['newfile'])
+        builder.finish_tree()
 
         # Asia/Colombo offset = 5 hours 30 minutes
         self.tree1.commit('non-hour offset timezone', rev_id='tz-1',
@@ -801,6 +804,7 @@
         self.assertEqual('Mon 2006-07-10 20:51:26.000000000 +0530', rev.date)
         self.assertEqual(19800, rev.timezone)
         self.assertEqual(1152544886.0, rev.timestamp)
+        self.tree1.unlock()
 
     def test_bundle_root_id(self):
         self.tree1 = self.make_branch_and_tree('b1')

=== modified file 'bzrlib/tests/test_bzrdir.py'
--- bzrlib/tests/test_bzrdir.py	2006-09-05 20:58:28 +0000
+++ bzrlib/tests/test_bzrdir.py	2006-09-14 04:55:30 +0000
@@ -392,6 +392,28 @@
             get_transport(self.get_readonly_url('g/p/q')))
         self.assertEqual('g/p/q', relpath)
 
+    def test_open_from_transport(self):
+        # transport pointing at bzrdir should give a bzrdir with root transport
+        # set to the given transport
+        control = bzrdir.BzrDir.create(self.get_url())
+        transport = get_transport(self.get_url())
+        opened_bzrdir = bzrdir.BzrDir.open_from_transport(transport)
+        self.assertEqual(transport.base, opened_bzrdir.root_transport.base)
+        self.assertIsInstance(opened_bzrdir, bzrdir.BzrDir)
+        
+    def test_open_from_transport_no_bzrdir(self):
+        transport = get_transport(self.get_url())
+        self.assertRaises(NotBranchError, bzrdir.BzrDir.open_from_transport,
+                          transport)
+
+    def test_open_from_transport_bzrdir_in_parent(self):
+        control = bzrdir.BzrDir.create(self.get_url())
+        transport = get_transport(self.get_url())
+        transport.mkdir('subdir')
+        transport = transport.clone('subdir')
+        self.assertRaises(NotBranchError, bzrdir.BzrDir.open_from_transport,
+                          transport)
+
 
 class TestMeta1DirFormat(TestCaseWithTransport):
     """Tests specific to the meta1 dir format."""

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2006-09-07 07:31:51 +0000
+++ bzrlib/tests/test_errors.py	2006-09-14 05:51:28 +0000
@@ -61,6 +61,10 @@
         errors.BzrNewError.__init__(self, foo=foo, bar=bar)
 
 
+class ErrorWithBadFormat(errors.BzrNewError):
+    """One format specifier: %(thing)s"""
+
+
 class TestErrorFormatting(TestCase):
     
     def test_always_str(self):
@@ -72,3 +76,12 @@
         # returned from e.__str__(), and it has non ascii characters
         s = str(e)
         self.assertEqual('Pass through \xc2\xb5 and bar', s)
+
+    def test_mismatched_format_args(self):
+        # Even though ErrorWithBadFormat's format string does not match the
+        # arguments we constructing it with, we can still stringify an instance
+        # of this exception. The resulting string will say its unprintable.
+        e = ErrorWithBadFormat(not_thing='x')
+        self.assertStartsWith(
+            str(e), 'Unprintable exception ErrorWithBadFormat(')
+

=== modified file 'bzrlib/tests/test_http.py'
--- bzrlib/tests/test_http.py	2006-07-14 19:25:24 +0000
+++ bzrlib/tests/test_http.py	2006-09-14 06:28:00 +0000
@@ -19,6 +19,8 @@
 
 # TODO: Should be renamed to bzrlib.transport.http.tests?
 
+import socket
+
 import bzrlib
 from bzrlib.errors import DependencyNotPresent
 from bzrlib.tests import TestCase, TestSkipped
@@ -68,9 +70,6 @@
         self.assertRaises(ValueError,
             t.abspath,
             '.bzr/')
-        self.assertRaises(ValueError,
-            t.abspath,
-            '/.bzr')
 
     def test_http_root_urls(self):
         """Construction of URLs from server root"""
@@ -135,6 +134,14 @@
         TestCaseWithWebserver.setUp(self)
         self._prep_tree()
 
+    def test_has_on_bogus_host(self):
+        import urllib2
+        # Get a random address, so that we can be sure there is no
+        # http handler there.
+        s = socket.socket()
+        s.bind(('localhost', 0))
+        t = self._transport('http://%s:%s/' % s.getsockname())
+        self.assertRaises(urllib2.URLError, t.has, 'foo/bar')
 
 
 class TestHttpConnections_pycurl(TestCaseWithWebserver, TestHttpMixins):

=== added file 'bzrlib/tests/test_lazy_import.py'
--- bzrlib/tests/test_lazy_import.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_lazy_import.py	2006-09-13 02:16:02 +0000
@@ -0,0 +1,1043 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Test the lazy_import functionality."""
+
+import os
+import sys
+
+from bzrlib import (
+    errors,
+    lazy_import,
+    osutils,
+    )
+from bzrlib.tests import TestCase, TestCaseInTempDir
+
+
+class InstrumentedReplacer(lazy_import.ScopeReplacer):
+    """Track what actions are done"""
+
+    @staticmethod
+    def use_actions(actions):
+        InstrumentedReplacer.actions = actions
+
+    def _replace(self):
+        InstrumentedReplacer.actions.append('_replace')
+        return lazy_import.ScopeReplacer._replace(self)
+
+    def __getattribute__(self, attr):
+        InstrumentedReplacer.actions.append(('__getattribute__', attr))
+        return lazy_import.ScopeReplacer.__getattribute__(self, attr)
+
+    def __call__(self, *args, **kwargs):
+        InstrumentedReplacer.actions.append(('__call__', args, kwargs))
+        return lazy_import.ScopeReplacer.__call__(self, *args, **kwargs)
+
+
+class InstrumentedImportReplacer(lazy_import.ImportReplacer):
+
+    @staticmethod
+    def use_actions(actions):
+        InstrumentedImportReplacer.actions = actions
+
+    def _import(self, scope, name):
+        InstrumentedImportReplacer.actions.append(('_import', name))
+        return lazy_import.ImportReplacer._import(self, scope, name)
+
+    def _replace(self):
+        InstrumentedImportReplacer.actions.append('_replace')
+        return lazy_import.ScopeReplacer._replace(self)
+
+    def __getattribute__(self, attr):
+        InstrumentedImportReplacer.actions.append(('__getattribute__', attr))
+        return lazy_import.ScopeReplacer.__getattribute__(self, attr)
+
+    def __call__(self, *args, **kwargs):
+        InstrumentedImportReplacer.actions.append(('__call__', args, kwargs))
+        return lazy_import.ScopeReplacer.__call__(self, *args, **kwargs)
+
+
+class TestClass(object):
+    """Just a simple test class instrumented for the test cases"""
+
+    class_member = 'class_member'
+
+    @staticmethod
+    def use_actions(actions):
+        TestClass.actions = actions
+
+    def __init__(self):
+        TestClass.actions.append('init')
+
+    def foo(self, x):
+        TestClass.actions.append(('foo', x))
+        return 'foo'
+
+
+class TestScopeReplacer(TestCase):
+    """Test the ability of the replacer to put itself into the correct scope.
+
+    In these tests we use the global scope, because we cannot replace
+    variables in the local scope. This means that we need to be careful
+    and not have the replacing objects use the same name, or we would
+    get collisions.
+    """
+
+    def test_object(self):
+        """ScopeReplacer can create an instance in local scope.
+        
+        An object should appear in globals() by constructing a ScopeReplacer,
+        and it will be replaced with the real object upon the first request.
+        """
+        actions = []
+        InstrumentedReplacer.use_actions(actions)
+        TestClass.use_actions(actions)
+
+        def factory(replacer, scope, name):
+            actions.append('factory')
+            return TestClass()
+
+        try:
+            test_obj1
+        except NameError:
+            # test_obj1 shouldn't exist yet
+            pass
+        else:
+            self.fail('test_obj1 was not supposed to exist yet')
+
+        orig_globals = set(globals().keys())
+
+        InstrumentedReplacer(scope=globals(), name='test_obj1',
+                             factory=factory)
+
+        new_globals = set(globals().keys())
+
+        # We can't use isinstance() because that uses test_obj1.__class__
+        # and that goes through __getattribute__ which would activate
+        # the replacement
+        self.assertEqual(InstrumentedReplacer,
+                         object.__getattribute__(test_obj1, '__class__'))
+        self.assertEqual('foo', test_obj1.foo(1))
+        self.assertIsInstance(test_obj1, TestClass)
+        self.assertEqual('foo', test_obj1.foo(2))
+        self.assertEqual([('__getattribute__', 'foo'),
+                          '_replace',
+                          'factory',
+                          'init',
+                          ('foo', 1),
+                          ('foo', 2),
+                         ], actions)
+
+    def test_replace_side_effects(self):
+        """Creating a new object should only create one entry in globals.
+
+        And only that entry even after replacement.
+        """
+        try:
+            test_scope1
+        except NameError:
+            # test_scope1 shouldn't exist yet
+            pass
+        else:
+            self.fail('test_scope1 was not supposed to exist yet')
+
+        # ignore the logged actions
+        TestClass.use_actions([])
+
+        def factory(replacer, scope, name):
+            return TestClass()
+
+        orig_globals = set(globals().keys())
+
+        lazy_import.ScopeReplacer(scope=globals(), name='test_scope1',
+                                  factory=factory)
+
+        new_globals = set(globals().keys())
+
+        self.assertEqual(lazy_import.ScopeReplacer,
+                         object.__getattribute__(test_scope1, '__class__'))
+        self.assertEqual('foo', test_scope1.foo(1))
+        self.assertIsInstance(test_scope1, TestClass)
+
+        final_globals = set(globals().keys())
+
+        self.assertEqual(set(['test_scope1']), new_globals - orig_globals)
+        self.assertEqual(set(), orig_globals - new_globals)
+        self.assertEqual(set(), final_globals - new_globals)
+        self.assertEqual(set(), new_globals - final_globals)
+
+    def test_class(self):
+        actions = []
+        InstrumentedReplacer.use_actions(actions)
+        TestClass.use_actions(actions)
+
+        def factory(replacer, scope, name):
+            actions.append('factory')
+            return TestClass
+
+        try:
+            test_class1
+        except NameError:
+            # test_class2 shouldn't exist yet
+            pass
+        else:
+            self.fail('test_class1 was not supposed to exist yet')
+
+        InstrumentedReplacer(scope=globals(), name='test_class1',
+                             factory=factory)
+
+        self.assertEqual('class_member', test_class1.class_member)
+        self.assertEqual(test_class1, TestClass)
+        self.assertEqual([('__getattribute__', 'class_member'),
+                          '_replace',
+                          'factory',
+                         ], actions)
+
+    def test_call_class(self):
+        actions = []
+        InstrumentedReplacer.use_actions(actions)
+        TestClass.use_actions(actions)
+
+        def factory(replacer, scope, name):
+            actions.append('factory')
+            return TestClass
+
+        try:
+            test_class2
+        except NameError:
+            # test_class2 shouldn't exist yet
+            pass
+        else:
+            self.fail('test_class2 was not supposed to exist yet')
+
+        InstrumentedReplacer(scope=globals(), name='test_class2',
+                             factory=factory)
+
+        self.failIf(test_class2 is TestClass)
+        obj = test_class2()
+        self.assertIs(test_class2, TestClass)
+        self.assertIsInstance(obj, TestClass)
+        self.assertEqual('class_member', obj.class_member)
+        self.assertEqual([('__call__', (), {}),
+                          '_replace',
+                          'factory',
+                          'init',
+                         ], actions)
+
+    def test_call_func(self):
+        actions = []
+        InstrumentedReplacer.use_actions(actions)
+
+        def func(a, b, c=None):
+            actions.append('func')
+            return (a, b, c)
+
+        def factory(replacer, scope, name):
+            actions.append('factory')
+            return func
+
+        try:
+            test_func1
+        except NameError:
+            # test_func1 shouldn't exist yet
+            pass
+        else:
+            self.fail('test_func1 was not supposed to exist yet')
+        InstrumentedReplacer(scope=globals(), name='test_func1',
+                             factory=factory)
+
+        self.failIf(test_func1 is func)
+        val = test_func1(1, 2, c='3')
+        self.assertIs(test_func1, func)
+
+        self.assertEqual((1,2,'3'), val)
+        self.assertEqual([('__call__', (1,2), {'c':'3'}),
+                          '_replace',
+                          'factory',
+                          'func',
+                         ], actions)
+
+    def test_other_variable(self):
+        """Test when a ScopeReplacer is assigned to another variable.
+
+        This test could be updated if we find a way to trap '=' rather
+        than just giving a belated exception.
+        ScopeReplacer only knows about the variable it was created as,
+        so until the object is replaced, it is illegal to pass it to
+        another variable. (Though discovering this may take a while)
+        """
+        actions = []
+        InstrumentedReplacer.use_actions(actions)
+        TestClass.use_actions(actions)
+
+        def factory(replacer, scope, name):
+            actions.append('factory')
+            return TestClass()
+
+        try:
+            test_obj2
+        except NameError:
+            # test_obj2 shouldn't exist yet
+            pass
+        else:
+            self.fail('test_obj2 was not supposed to exist yet')
+
+        InstrumentedReplacer(scope=globals(), name='test_obj2',
+                             factory=factory)
+
+        self.assertEqual(InstrumentedReplacer,
+                         object.__getattribute__(test_obj2, '__class__'))
+        # This is technically not allowed, but we don't have a way to
+        # test it until later.
+        test_obj3 = test_obj2
+        self.assertEqual(InstrumentedReplacer,
+                         object.__getattribute__(test_obj2, '__class__'))
+        self.assertEqual(InstrumentedReplacer,
+                         object.__getattribute__(test_obj3, '__class__'))
+        
+        # The first use of the alternate variable causes test_obj2 to
+        # be replaced.
+        self.assertEqual('foo', test_obj3.foo(1))
+        # test_obj2 has been replaced, but the ScopeReplacer has no
+        # idea of test_obj3
+        self.assertEqual(TestClass,
+                         object.__getattribute__(test_obj2, '__class__'))
+        self.assertEqual(InstrumentedReplacer,
+                         object.__getattribute__(test_obj3, '__class__'))
+        # We should be able to access test_obj2 attributes normally
+        self.assertEqual('foo', test_obj2.foo(2))
+        self.assertEqual('foo', test_obj2.foo(3))
+
+        # However, the next access on test_obj3 should raise an error
+        # because only now are we able to detect the problem.
+        self.assertRaises(errors.IllegalUseOfScopeReplacer,
+                          getattr, test_obj3, 'foo')
+        
+        # However, the 
+        self.assertEqual([('__getattribute__', 'foo'),
+                          '_replace',
+                          'factory',
+                          'init',
+                          ('foo', 1),
+                          ('foo', 2),
+                          ('foo', 3),
+                          ('__getattribute__', 'foo'),
+                          '_replace',
+                         ], actions)
+
+
+class ImportReplacerHelper(TestCaseInTempDir):
+    """Test the ability to have a lazily imported module or object"""
+
+    def setUp(self):
+        TestCaseInTempDir.setUp(self)
+        self.create_modules()
+        base_path = self.test_dir + '/base'
+
+        self.actions = []
+        InstrumentedImportReplacer.use_actions(self.actions)
+
+        original_import = __import__
+        def instrumented_import(mod, scope1, scope2, fromlist):
+            self.actions.append(('import', mod, fromlist))
+            return original_import(mod, scope1, scope2, fromlist)
+
+        def cleanup():
+            if base_path in sys.path:
+                sys.path.remove(base_path)
+            __builtins__['__import__'] = original_import
+        self.addCleanup(cleanup)
+        sys.path.append(base_path)
+        __builtins__['__import__'] = instrumented_import
+
+    def create_modules(self):
+        """Create some random modules to be imported.
+
+        Each entry has a random suffix, and the full names are saved
+
+        These are setup as follows:
+         base/ <= used to ensure not in default search path
+            root-XXX/
+                __init__.py <= This will contain var1, func1
+                mod-XXX.py <= This will contain var2, func2
+                sub-XXX/
+                    __init__.py <= Contains var3, func3
+                    submoda-XXX.py <= contains var4, func4
+                    submodb-XXX.py <= containse var5, func5
+        """
+        rand_suffix = osutils.rand_chars(4)
+        root_name = 'root_' + rand_suffix
+        mod_name = 'mod_' + rand_suffix
+        sub_name = 'sub_' + rand_suffix
+        submoda_name = 'submoda_' + rand_suffix
+        submodb_name = 'submodb_' + rand_suffix
+
+        os.mkdir('base')
+        root_path = osutils.pathjoin('base', root_name)
+        os.mkdir(root_path)
+        root_init = osutils.pathjoin(root_path, '__init__.py')
+        f = open(osutils.pathjoin(root_path, '__init__.py'), 'wb')
+        try:
+            f.write('var1 = 1\ndef func1(a):\n  return a\n')
+        finally:
+            f.close()
+        mod_path = osutils.pathjoin(root_path, mod_name + '.py')
+        f = open(mod_path, 'wb')
+        try:
+            f.write('var2 = 2\ndef func2(a):\n  return a\n')
+        finally:
+            f.close()
+
+        sub_path = osutils.pathjoin(root_path, sub_name)
+        os.mkdir(sub_path)
+        f = open(osutils.pathjoin(sub_path, '__init__.py'), 'wb')
+        try:
+            f.write('var3 = 3\ndef func3(a):\n  return a\n')
+        finally:
+            f.close()
+        submoda_path = osutils.pathjoin(sub_path, submoda_name + '.py')
+        f = open(submoda_path, 'wb')
+        try:
+            f.write('var4 = 4\ndef func4(a):\n  return a\n')
+        finally:
+            f.close()
+        submodb_path = osutils.pathjoin(sub_path, submodb_name + '.py')
+        f = open(submodb_path, 'wb')
+        try:
+            f.write('var5 = 5\ndef func5(a):\n  return a\n')
+        finally:
+            f.close()
+        self.root_name = root_name
+        self.mod_name = mod_name
+        self.sub_name = sub_name
+        self.submoda_name = submoda_name
+        self.submodb_name = submodb_name
+
+
+class TestImportReplacerHelper(ImportReplacerHelper):
+
+    def test_basic_import(self):
+        """Test that a real import of these modules works"""
+        sub_mod_path = '.'.join([self.root_name, self.sub_name,
+                                  self.submoda_name])
+        root = __import__(sub_mod_path, globals(), locals(), [])
+        self.assertEqual(1, root.var1)
+        self.assertEqual(3, getattr(root, self.sub_name).var3)
+        self.assertEqual(4, getattr(getattr(root, self.sub_name),
+                                    self.submoda_name).var4)
+
+        mod_path = '.'.join([self.root_name, self.mod_name])
+        root = __import__(mod_path, globals(), locals(), [])
+        self.assertEqual(2, getattr(root, self.mod_name).var2)
+
+        self.assertEqual([('import', sub_mod_path, []),
+                          ('import', mod_path, []),
+                         ], self.actions)
+
+
+class TestImportReplacer(ImportReplacerHelper):
+
+    def test_import_root(self):
+        """Test 'import root-XXX as root1'"""
+        try:
+            root1
+        except NameError:
+            # root1 shouldn't exist yet
+            pass
+        else:
+            self.fail('root1 was not supposed to exist yet')
+
+        # This should replicate 'import root-xxyyzz as root1'
+        InstrumentedImportReplacer(scope=globals(), name='root1',
+                                   module_path=[self.root_name],
+                                   member=None, children={})
+
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root1, '__class__'))
+        self.assertEqual(1, root1.var1)
+        self.assertEqual('x', root1.func1('x'))
+
+        self.assertEqual([('__getattribute__', 'var1'),
+                          '_replace',
+                          ('_import', 'root1'),
+                          ('import', self.root_name, []),
+                         ], self.actions)
+
+    def test_import_mod(self):
+        """Test 'import root-XXX.mod-XXX as mod2'"""
+        try:
+            mod1
+        except NameError:
+            # mod1 shouldn't exist yet
+            pass
+        else:
+            self.fail('mod1 was not supposed to exist yet')
+
+        mod_path = self.root_name + '.' + self.mod_name
+        InstrumentedImportReplacer(scope=globals(), name='mod1',
+                                   module_path=[self.root_name, self.mod_name],
+                                   member=None, children={})
+
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(mod1, '__class__'))
+        self.assertEqual(2, mod1.var2)
+        self.assertEqual('y', mod1.func2('y'))
+
+        self.assertEqual([('__getattribute__', 'var2'),
+                          '_replace',
+                          ('_import', 'mod1'),
+                          ('import', mod_path, []),
+                         ], self.actions)
+
+    def test_import_mod_from_root(self):
+        """Test 'from root-XXX import mod-XXX as mod2'"""
+        try:
+            mod2
+        except NameError:
+            # mod2 shouldn't exist yet
+            pass
+        else:
+            self.fail('mod2 was not supposed to exist yet')
+
+        InstrumentedImportReplacer(scope=globals(), name='mod2',
+                                   module_path=[self.root_name],
+                                   member=self.mod_name, children={})
+
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(mod2, '__class__'))
+        self.assertEqual(2, mod2.var2)
+        self.assertEqual('y', mod2.func2('y'))
+
+        self.assertEqual([('__getattribute__', 'var2'),
+                          '_replace',
+                          ('_import', 'mod2'),
+                          ('import', self.root_name, [self.mod_name]),
+                         ], self.actions)
+
+    def test_import_root_and_mod(self):
+        """Test 'import root-XXX.mod-XXX' remapping both to root3.mod3"""
+        try:
+            root3
+        except NameError:
+            # root3 shouldn't exist yet
+            pass
+        else:
+            self.fail('root3 was not supposed to exist yet')
+
+        InstrumentedImportReplacer(scope=globals(),
+            name='root3', module_path=[self.root_name], member=None,
+            children={'mod3':([self.root_name, self.mod_name], None, {})})
+
+        # So 'root3' should be a lazy import
+        # and once it is imported, mod3 should also be lazy until
+        # actually accessed.
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root3, '__class__'))
+        self.assertEqual(1, root3.var1)
+
+        # There is a mod3 member, but it is also lazy
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root3.mod3, '__class__'))
+        self.assertEqual(2, root3.mod3.var2)
+
+        mod_path = self.root_name + '.' + self.mod_name
+        self.assertEqual([('__getattribute__', 'var1'),
+                          '_replace',
+                          ('_import', 'root3'),
+                          ('import', self.root_name, []),
+                          ('__getattribute__', 'var2'),
+                          '_replace',
+                          ('_import', 'mod3'),
+                          ('import', mod_path, []),
+                         ], self.actions)
+
+    def test_import_root_and_root_mod(self):
+        """Test that 'import root, root.mod' can be done.
+
+        The second import should re-use the first one, and just add
+        children to be imported.
+        """
+        try:
+            root4
+        except NameError:
+            # root4 shouldn't exist yet
+            pass
+        else:
+            self.fail('root4 was not supposed to exist yet')
+
+        InstrumentedImportReplacer(scope=globals(),
+            name='root4', module_path=[self.root_name], member=None,
+            children={})
+
+        # So 'root4' should be a lazy import
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root4, '__class__'))
+
+        # Lets add a new child to be imported on demand
+        # This syntax of using object.__getattribute__ is the correct method
+        # for accessing the _import_replacer_children member
+        children = object.__getattribute__(root4, '_import_replacer_children')
+        children['mod4'] = ([self.root_name, self.mod_name], None, {})
+
+        # Accessing root4.mod4 should import root, but mod should stay lazy
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root4.mod4, '__class__'))
+        self.assertEqual(2, root4.mod4.var2)
+
+        mod_path = self.root_name + '.' + self.mod_name
+        self.assertEqual([('__getattribute__', 'mod4'),
+                          '_replace',
+                          ('_import', 'root4'),
+                          ('import', self.root_name, []),
+                          ('__getattribute__', 'var2'),
+                          '_replace',
+                          ('_import', 'mod4'),
+                          ('import', mod_path, []),
+                         ], self.actions)
+
+    def test_import_root_sub_submod(self):
+        """Test import root.mod, root.sub.submoda, root.sub.submodb
+        root should be a lazy import, with multiple children, who also
+        have children to be imported.
+        And when root is imported, the children should be lazy, and
+        reuse the intermediate lazy object.
+        """
+        try:
+            root5
+        except NameError:
+            # root5 shouldn't exist yet
+            pass
+        else:
+            self.fail('root5 was not supposed to exist yet')
+
+        InstrumentedImportReplacer(scope=globals(),
+            name='root5', module_path=[self.root_name], member=None,
+            children={'mod5':([self.root_name, self.mod_name], None, {}),
+                      'sub5':([self.root_name, self.sub_name], None,
+                            {'submoda5':([self.root_name, self.sub_name,
+                                         self.submoda_name], None, {}),
+                             'submodb5':([self.root_name, self.sub_name,
+                                          self.submodb_name], None, {})
+                            }),
+                     })
+
+        # So 'root5' should be a lazy import
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root5, '__class__'))
+
+        # Accessing root5.mod5 should import root, but mod should stay lazy
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root5.mod5, '__class__'))
+        # root5.sub5 should still be lazy, but not re-import root5
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root5.sub5, '__class__'))
+
+        # Accessing root5.sub5.submoda5 should import sub5, but not either
+        # of the sub objects (they should be available as lazy objects
+        self.assertEqual(InstrumentedImportReplacer,
+                     object.__getattribute__(root5.sub5.submoda5, '__class__'))
+        self.assertEqual(InstrumentedImportReplacer,
+                     object.__getattribute__(root5.sub5.submodb5, '__class__'))
+
+        # This should import mod5
+        self.assertEqual(2, root5.mod5.var2)
+        # These should import submoda5 and submodb5
+        self.assertEqual(4, root5.sub5.submoda5.var4)
+        self.assertEqual(5, root5.sub5.submodb5.var5)
+
+        mod_path = self.root_name + '.' + self.mod_name
+        sub_path = self.root_name + '.' + self.sub_name
+        submoda_path = sub_path + '.' + self.submoda_name
+        submodb_path = sub_path + '.' + self.submodb_name
+
+        self.assertEqual([('__getattribute__', 'mod5'),
+                          '_replace',
+                          ('_import', 'root5'),
+                          ('import', self.root_name, []),
+                          ('__getattribute__', 'submoda5'),
+                          '_replace',
+                          ('_import', 'sub5'),
+                          ('import', sub_path, []),
+                          ('__getattribute__', 'var2'),
+                          '_replace',
+                          ('_import', 'mod5'),
+                          ('import', mod_path, []),
+                          ('__getattribute__', 'var4'),
+                          '_replace',
+                          ('_import', 'submoda5'),
+                          ('import', submoda_path, []),
+                          ('__getattribute__', 'var5'),
+                          '_replace',
+                          ('_import', 'submodb5'),
+                          ('import', submodb_path, []),
+                         ], self.actions)
+
+
+class TestConvertImportToMap(TestCase):
+    """Directly test the conversion from import strings to maps"""
+
+    def check(self, expected, import_strings):
+        proc = lazy_import.ImportProcessor()
+        for import_str in import_strings:
+            proc._convert_import_str(import_str)
+        self.assertEqual(expected, proc.imports,
+                         'Import of %r was not converted correctly'
+                         ' %s != %s' % (import_strings, expected,
+                                        proc.imports))
+
+    def test_import_one(self):
+        self.check({'one':(['one'], None, {}),
+                   }, ['import one'])
+
+    def test_import_one_two(self):
+        one_two_map = {'one':(['one'], None,
+                              {'two':(['one', 'two'], None, {}),
+                              }),
+                      }
+        self.check(one_two_map, ['import one.two'])
+        self.check(one_two_map, ['import one, one.two'])
+        self.check(one_two_map, ['import one', 'import one.two'])
+        self.check(one_two_map, ['import one.two', 'import one'])
+
+    def test_import_one_two_three(self):
+        one_two_three_map = {
+            'one':(['one'], None,
+                   {'two':(['one', 'two'], None,
+                           {'three':(['one', 'two', 'three'], None, {}),
+                           }),
+                   }),
+        }
+        self.check(one_two_three_map, ['import one.two.three'])
+        self.check(one_two_three_map, ['import one, one.two.three'])
+        self.check(one_two_three_map, ['import one',
+                                              'import one.two.three'])
+        self.check(one_two_three_map, ['import one.two.three',
+                                              'import one'])
+
+    def test_import_one_as_x(self):
+        self.check({'x':(['one'], None, {}),
+                          }, ['import one as x'])
+
+    def test_import_one_two_as_x(self):
+        self.check({'x':(['one', 'two'], None, {}),
+                   }, ['import one.two as x'])
+
+    def test_import_mixed(self):
+        mixed = {'x':(['one', 'two'], None, {}),
+                 'one':(['one'], None,
+                       {'two':(['one', 'two'], None, {}),
+                       }),
+                }
+        self.check(mixed, ['import one.two as x, one.two'])
+        self.check(mixed, ['import one.two as x', 'import one.two'])
+        self.check(mixed, ['import one.two', 'import one.two as x'])
+
+    def test_import_with_as(self):
+        self.check({'fast':(['fast'], None, {})}, ['import fast'])
+
+
+class TestFromToMap(TestCase):
+    """Directly test the conversion of 'from foo import bar' syntax"""
+
+    def check_result(self, expected, from_strings):
+        proc = lazy_import.ImportProcessor()
+        for from_str in from_strings:
+            proc._convert_from_str(from_str)
+        self.assertEqual(expected, proc.imports,
+                         'Import of %r was not converted correctly'
+                         ' %s != %s' % (from_strings, expected, proc.imports))
+
+    def test_from_one_import_two(self):
+        self.check_result({'two':(['one'], 'two', {})},
+                          ['from one import two'])
+
+    def test_from_one_import_two_as_three(self):
+        self.check_result({'three':(['one'], 'two', {})},
+                          ['from one import two as three'])
+
+    def test_from_one_import_two_three(self):
+        two_three_map = {'two':(['one'], 'two', {}),
+                         'three':(['one'], 'three', {}),
+                        }
+        self.check_result(two_three_map,
+                          ['from one import two, three'])
+        self.check_result(two_three_map,
+                          ['from one import two',
+                           'from one import three'])
+
+    def test_from_one_two_import_three(self):
+        self.check_result({'three':(['one', 'two'], 'three', {})},
+                          ['from one.two import three'])
+
+
+class TestCanonicalize(TestCase):
+    """Test that we can canonicalize import texts"""
+
+    def check(self, expected, text):
+        proc = lazy_import.ImportProcessor()
+        parsed = proc._canonicalize_import_text(text)
+        self.assertEqual(expected, parsed,
+                         'Incorrect parsing of text:\n%s\n%s\n!=\n%s'
+                         % (text, expected, parsed))
+
+    def test_import_one(self):
+        self.check(['import one'], 'import one')
+        self.check(['import one'], '\nimport one\n\n')
+
+    def test_import_one_two(self):
+        self.check(['import one, two'], 'import one, two')
+        self.check(['import one, two'], '\nimport one, two\n\n')
+
+    def test_import_one_as_two_as(self):
+        self.check(['import one as x, two as y'], 'import one as x, two as y')
+        self.check(['import one as x, two as y'],
+                   '\nimport one as x, two as y\n')
+
+    def test_from_one_import_two(self):
+        self.check(['from one import two'], 'from one import two')
+        self.check(['from one import two'], '\nfrom one import two\n\n')
+        self.check(['from one import two'], '\nfrom one import (two)\n')
+        self.check(['from one import  two '], '\nfrom one import (\n\ttwo\n)\n')
+
+    def test_multiple(self):
+        self.check(['import one', 'import two, three', 'from one import four'],
+                   'import one\nimport two, three\nfrom one import four')
+        self.check(['import one', 'import two, three', 'from one import four'],
+                   'import one\nimport (two, three)\nfrom one import four')
+        self.check(['import one', 'import two, three', 'from one import four'],
+                   'import one\n'
+                   'import two, three\n'
+                   'from one import four')
+        self.check(['import one',
+                    'import two, three', 'from one import  four, '],
+                   'import one\n'
+                   'import two, three\n'
+                   'from one import (\n'
+                   '    four,\n'
+                   '    )\n'
+                   )
+
+    def test_missing_trailing(self):
+        proc = lazy_import.ImportProcessor()
+        self.assertRaises(errors.InvalidImportLine,
+                          proc._canonicalize_import_text,
+                          "from foo import (\n  bar\n")
+
+
+class TestImportProcessor(TestCase):
+    """Test that ImportProcessor can turn import texts into lazy imports"""
+
+    def check(self, expected, text):
+        proc = lazy_import.ImportProcessor()
+        proc._build_map(text)
+        self.assertEqual(expected, proc.imports,
+                         'Incorrect processing of:\n%s\n%s\n!=\n%s'
+                         % (text, expected, proc.imports))
+
+    def test_import_one(self):
+        exp = {'one':(['one'], None, {})}
+        self.check(exp, 'import one')
+        self.check(exp, '\nimport one\n')
+
+    def test_import_one_two(self):
+        exp = {'one':(['one'], None,
+                      {'two':(['one', 'two'], None, {}),
+                      }),
+              }
+        self.check(exp, 'import one.two')
+        self.check(exp, 'import one, one.two')
+        self.check(exp, 'import one\nimport one.two')
+
+    def test_import_as(self):
+        exp = {'two':(['one'], None, {})}
+        self.check(exp, 'import one as two')
+
+    def test_import_many(self):
+        exp = {'one':(['one'], None,
+                      {'two':(['one', 'two'], None,
+                              {'three':(['one', 'two', 'three'], None, {}),
+                              }),
+                       'four':(['one', 'four'], None, {}),
+                      }),
+               'five':(['one', 'five'], None, {}),
+              }
+        self.check(exp, 'import one.two.three, one.four, one.five as five')
+        self.check(exp, 'import one.five as five\n'
+                        'import one\n'
+                        'import one.two.three\n'
+                        'import one.four\n')
+
+    def test_from_one_import_two(self):
+        exp = {'two':(['one'], 'two', {})}
+        self.check(exp, 'from one import two\n')
+        self.check(exp, 'from one import (\n'
+                        '    two,\n'
+                        '    )\n')
+
+    def test_from_one_import_two(self):
+        exp = {'two':(['one'], 'two', {})}
+        self.check(exp, 'from one import two\n')
+        self.check(exp, 'from one import (two)\n')
+        self.check(exp, 'from one import (two,)\n')
+        self.check(exp, 'from one import two as two\n')
+        self.check(exp, 'from one import (\n'
+                        '    two,\n'
+                        '    )\n')
+
+    def test_from_many(self):
+        exp = {'two':(['one'], 'two', {}),
+               'three':(['one', 'two'], 'three', {}),
+               'five':(['one', 'two'], 'four', {}),
+              }
+        self.check(exp, 'from one import two\n'
+                        'from one.two import three, four as five\n')
+        self.check(exp, 'from one import two\n'
+                        'from one.two import (\n'
+                        '    three,\n'
+                        '    four as five,\n'
+                        '    )\n')
+
+    def test_mixed(self):
+        exp = {'two':(['one'], 'two', {}),
+               'three':(['one', 'two'], 'three', {}),
+               'five':(['one', 'two'], 'four', {}),
+               'one':(['one'], None,
+                      {'two':(['one', 'two'], None, {}),
+                      }),
+              }
+        self.check(exp, 'from one import two\n'
+                        'from one.two import three, four as five\n'
+                        'import one.two')
+        self.check(exp, 'from one import two\n'
+                        'from one.two import (\n'
+                        '    three,\n'
+                        '    four as five,\n'
+                        '    )\n'
+                        'import one\n'
+                        'import one.two\n')
+
+    def test_incorrect_line(self):
+        proc = lazy_import.ImportProcessor()
+        self.assertRaises(errors.InvalidImportLine,
+                          proc._build_map, 'foo bar baz')
+        self.assertRaises(errors.InvalidImportLine,
+                          proc._build_map, 'improt foo')
+        self.assertRaises(errors.InvalidImportLine,
+                          proc._build_map, 'importfoo')
+        self.assertRaises(errors.InvalidImportLine,
+                          proc._build_map, 'fromimport')
+
+    def test_name_collision(self):
+        proc = lazy_import.ImportProcessor()
+        proc._build_map('import foo')
+
+        # All of these would try to create an object with the
+        # same name as an existing object.
+        self.assertRaises(errors.ImportNameCollision,
+                          proc._build_map, 'import bar as foo')
+        self.assertRaises(errors.ImportNameCollision,
+                          proc._build_map, 'from foo import bar as foo')
+        self.assertRaises(errors.ImportNameCollision,
+                          proc._build_map, 'from bar import foo')
+
+
+class TestLazyImportProcessor(ImportReplacerHelper):
+
+    def test_root(self):
+        try:
+            root6
+        except NameError:
+            pass # root6 should not be defined yet
+        else:
+            self.fail('root6 was not supposed to exist yet')
+
+        text = 'import %s as root6' % (self.root_name,)
+        proc = lazy_import.ImportProcessor(InstrumentedImportReplacer)
+        proc.lazy_import(scope=globals(), text=text)
+
+        # So 'root6' should be a lazy import
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root6, '__class__'))
+
+        self.assertEqual(1, root6.var1)
+        self.assertEqual('x', root6.func1('x'))
+
+        self.assertEqual([('__getattribute__', 'var1'),
+                          '_replace',
+                          ('_import', 'root6'),
+                          ('import', self.root_name, []),
+                         ], self.actions)
+
+    def test_import_deep(self):
+        """Test import root.mod, root.sub.submoda, root.sub.submodb
+        root should be a lazy import, with multiple children, who also
+        have children to be imported.
+        And when root is imported, the children should be lazy, and
+        reuse the intermediate lazy object.
+        """
+        try:
+            submoda7
+        except NameError:
+            pass # submoda7 should not be defined yet
+        else:
+            self.fail('submoda7 was not supposed to exist yet')
+
+        text = """\
+import %(root_name)s.%(sub_name)s.%(submoda_name)s as submoda7
+""" % self.__dict__
+        proc = lazy_import.ImportProcessor(InstrumentedImportReplacer)
+        proc.lazy_import(scope=globals(), text=text)
+
+        # So 'submoda7' should be a lazy import
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(submoda7, '__class__'))
+
+        # This should import submoda7
+        self.assertEqual(4, submoda7.var4)
+
+        sub_path = self.root_name + '.' + self.sub_name
+        submoda_path = sub_path + '.' + self.submoda_name
+
+        self.assertEqual([('__getattribute__', 'var4'),
+                          '_replace',
+                          ('_import', 'submoda7'),
+                          ('import', submoda_path, []),
+                         ], self.actions)
+
+    def test_lazy_import(self):
+        """Smoke test that lazy_import() does the right thing"""
+        try:
+            root8
+        except NameError:
+            pass # root8 should not be defined yet
+        else:
+            self.fail('root8 was not supposed to exist yet')
+        lazy_import.lazy_import(globals(),
+                                'import %s as root8' % (self.root_name,),
+                                lazy_import_class=InstrumentedImportReplacer)
+
+        self.assertEqual(InstrumentedImportReplacer,
+                         object.__getattribute__(root8, '__class__'))
+
+        self.assertEqual(1, root8.var1)
+        self.assertEqual(1, root8.var1)
+        self.assertEqual(1, root8.func1(1))
+
+        self.assertEqual([('__getattribute__', 'var1'),
+                          '_replace',
+                          ('_import', 'root8'),
+                          ('import', self.root_name, []),
+                         ], self.actions)

=== added file 'bzrlib/tests/test_memorytree.py'
--- bzrlib/tests/test_memorytree.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_memorytree.py	2006-09-15 02:03:15 +0000
@@ -0,0 +1,142 @@
+# Copyright (C) 2006 Canonical Ltd
+# Authors:  Robert Collins <robert.collins@canonical.com>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for the MemoryTree class."""
+
+from bzrlib import errors
+from bzrlib.memorytree import MemoryTree
+from bzrlib.tests import TestCaseWithTransport
+from bzrlib.treebuilder import TreeBuilder
+
+
+class TestMemoryTree(TestCaseWithTransport):
+    
+    def test_create_on_branch(self):
+        """Creating a mutable tree on a trivial branch works."""
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        self.assertEqual(branch.bzrdir, tree.bzrdir)
+        self.assertEqual(branch, tree.branch)
+        self.assertEqual([], tree.get_parent_ids())
+    
+    def test_create_on_branch_with_content(self):
+        """Creating a mutable tree on a non-trivial branch works."""
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        # build some content
+        tree.lock_write()
+        builder = TreeBuilder()
+        builder.start_tree(tree)
+        builder.build(['foo'])
+        builder.finish_tree()
+        rev_id = tree.commit('first post')
+        tree.unlock()
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_read()
+        self.assertEqual([rev_id], tree.get_parent_ids())
+        self.assertEqual('contents of foo\n',
+            tree.get_file(tree.path2id('foo')).read())
+        tree.unlock()
+
+    def test_lock_tree_write(self):
+        """Check we can lock_tree_write and unlock MemoryTrees."""
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_tree_write()
+        tree.unlock()
+
+    def test_lock_tree_write_after_read_fails(self):
+        """Check that we error when trying to upgrade a read lock to write."""
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_read()
+        self.assertRaises(errors.ReadOnlyError, tree.lock_tree_write)
+        tree.unlock()
+
+    def test_lock_write(self):
+        """Check we can lock_write and unlock MemoryTrees."""
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_write()
+        tree.unlock()
+
+    def test_lock_write_after_read_fails(self):
+        """Check that we error when trying to upgrade a read lock to write."""
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_read()
+        self.assertRaises(errors.ReadOnlyError, tree.lock_write)
+        tree.unlock()
+
+    def test_add_with_kind(self):
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_write()
+        tree.add(['afile', 'adir'], None, ['file', 'directory'])
+        self.assertEqual('afile', tree.id2path(tree.path2id('afile')))
+        self.assertEqual('adir', tree.id2path(tree.path2id('adir')))
+        self.assertFalse(tree.has_filename('afile'))
+        self.assertFalse(tree.has_filename('adir'))
+        tree.unlock()
+
+    def test_put_new_file(self):
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_write()
+        tree.add(['foo'], ids=['foo-id'], kinds=['file'])
+        tree.put_file_bytes_non_atomic('foo-id', 'barshoom')
+        self.assertEqual('barshoom', tree.get_file('foo-id').read())
+        tree.unlock()
+
+    def test_put_existing_file(self):
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_write()
+        tree.add(['foo'], ids=['foo-id'], kinds=['file'])
+        tree.put_file_bytes_non_atomic('foo-id', 'first-content')
+        tree.put_file_bytes_non_atomic('foo-id', 'barshoom')
+        self.assertEqual('barshoom', tree.get_file('foo-id').read())
+        tree.unlock()
+
+    def test_commit_trivial(self):
+        """Smoke test for commit on a MemoryTree.
+
+        Becamse of commits design and layering, if this works, all commit
+        logic should work quite reliably.
+        """
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_write()
+        tree.add(['foo'], ids=['foo-id'], kinds=['file'])
+        tree.put_file_bytes_non_atomic('foo-id', 'barshoom')
+        revision_id = tree.commit('message baby')
+        # the parents list for the tree should have changed.
+        self.assertEqual([revision_id], tree.get_parent_ids())
+        tree.unlock()
+        # and we should have a revision that is accessible outside the tree lock
+        revtree = tree.branch.repository.revision_tree(revision_id)
+        self.assertEqual('barshoom', revtree.get_file('foo-id').read())
+
+    def test_unversion(self):
+        """Some test for unversion of a memory tree."""
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        tree.lock_write()
+        tree.add(['foo'], ids=['foo-id'], kinds=['file'])
+        tree.unversion(['foo-id'])
+        self.assertFalse(tree.has_id('foo-id'))
+        tree.unlock()

=== modified file 'bzrlib/tests/test_merge.py'
--- bzrlib/tests/test_merge.py	2006-09-06 02:47:54 +0000
+++ bzrlib/tests/test_merge.py	2006-09-11 09:34:51 +0000
@@ -127,7 +127,7 @@
         transform_tree(tree, tree.branch.basis_tree())
 
     def test_ignore_zero_merge_inner(self):
-        # Test that merge_inner's ignore zero paramter is effective
+        # Test that merge_inner's ignore zero parameter is effective
         tree_a =self.make_branch_and_tree('a')
         tree_a.commit(message="hello")
         dir_b = tree_a.bzrdir.sprout('b')

=== modified file 'bzrlib/tests/test_revisiontree.py'
--- bzrlib/tests/test_revisiontree.py	2006-07-14 05:02:37 +0000
+++ bzrlib/tests/test_revisiontree.py	2006-09-07 23:13:08 +0000
@@ -47,3 +47,7 @@
             [self.rev_id],
             self.t.branch.repository.revision_tree(revid_2).get_parent_ids())
         # TODO commit a merge and check it is reported correctly.
+
+        # the parents for a revision_tree(None) are []:
+        self.assertEqual([],
+            self.t.branch.repository.revision_tree(None).get_parent_ids())

=== modified file 'bzrlib/tests/test_selftest.py'
--- bzrlib/tests/test_selftest.py	2006-09-08 00:19:48 +0000
+++ bzrlib/tests/test_selftest.py	2006-09-15 01:21:38 +0000
@@ -34,11 +34,13 @@
                           TestSuite,
                           TextTestRunner,
                           )
+from bzrlib.tests.test_sftp_transport import TestCaseWithSFTPServer
 from bzrlib.tests.TestUtil import _load_module_by_name
 import bzrlib.errors as errors
 from bzrlib import symbol_versioning
 from bzrlib.symbol_versioning import zero_ten, zero_eleven
 from bzrlib.trace import note
+from bzrlib.transport.memory import MemoryServer, MemoryTransport
 from bzrlib.version import _get_bzr_source_tree
 
 
@@ -474,6 +476,21 @@
         self.assertRaises(AssertionError, self.assertIsDirectory, 'not_here', t)
 
 
+class TestTestCaseTransports(TestCaseWithTransport):
+
+    def setUp(self):
+        super(TestTestCaseTransports, self).setUp()
+        self.transport_server = MemoryServer
+
+    def test_make_bzrdir_preserves_transport(self):
+        t = self.get_transport()
+        result_bzrdir = self.make_bzrdir('subdir')
+        self.assertIsInstance(result_bzrdir.transport, 
+                              MemoryTransport)
+        # should not be on disk, should only be in memory
+        self.failIfExists('subdir')
+
+
 class TestChrootedTest(ChrootedTestCase):
 
     def test_root_is_root(self):
@@ -903,6 +920,27 @@
         self.assertIsInstance(bzrlib.bzrdir.BzrDir.open('b')._format,
                               bzrlib.bzrdir.BzrDirFormat6)
 
+    def test_make_branch_and_mutable_tree(self):
+        # we should be able to get a new branch and a mutable tree from
+        # TestCaseWithTransport
+        tree = self.make_branch_and_memory_tree('a')
+        self.assertIsInstance(tree, bzrlib.memorytree.MemoryTree)
+
+
+class TestSFTPMakeBranchAndTree(TestCaseWithSFTPServer):
+
+    def test_make_tree_for_sftp_branch(self):
+        """Transports backed by local directories create local trees."""
+
+        tree = self.make_branch_and_tree('t1')
+        base = tree.bzrdir.root_transport.base
+        self.failIf(base.startswith('sftp'),
+                'base %r is on sftp but should be local' % base)
+        self.assertEquals(tree.bzrdir.root_transport,
+                tree.branch.bzrdir.root_transport)
+        self.assertEquals(tree.bzrdir.root_transport,
+                tree.branch.repository.bzrdir.root_transport)
+
 
 class TestSelftest(TestCase):
     """Tests of bzrlib.tests.selftest."""

=== modified file 'bzrlib/tests/test_setup.py'
--- bzrlib/tests/test_setup.py	2006-08-15 13:19:12 +0000
+++ bzrlib/tests/test_setup.py	2006-09-09 16:43:09 +0000
@@ -46,6 +46,15 @@
         """
         if not os.path.isfile('setup.py'):
             raise TestSkipped('There is no setup.py file in current directory')
+        try:
+            import distutils.sysconfig
+            makefile_path = distutils.sysconfig.get_makefile_filename()
+            if not os.path.exists(makefile_path):
+                raise TestSkipped('You must have the python Makefile installed to run this test.'
+                                  ' Usually this can be found by installing "python-dev"')
+        except ImportError:
+            raise TestSkipped('You must have distutils installed to run this test.'
+                              ' Usually this can be found by installing "python-dev"')
         self.log('test_build running in %s' % os.getcwd())
         install_dir = tempfile.mkdtemp()
         # setup.py must be run from the root source directory, but the tests

=== modified file 'bzrlib/tests/test_sftp_transport.py'
--- bzrlib/tests/test_sftp_transport.py	2006-08-25 07:31:08 +0000
+++ bzrlib/tests/test_sftp_transport.py	2006-09-16 06:44:47 +0000
@@ -21,9 +21,10 @@
 
 import bzrlib.bzrdir as bzrdir
 import bzrlib.errors as errors
-from bzrlib.osutils import pathjoin, lexists
+from bzrlib.osutils import pathjoin, lexists, set_or_unset_env
 from bzrlib.tests import TestCaseWithTransport, TestCase, TestSkipped
 import bzrlib.transport
+from bzrlib.transport import get_transport
 import bzrlib.transport.http
 from bzrlib.workingtree import WorkingTree
 
@@ -115,16 +116,18 @@
         ### FIXME decide and then test for all transports. RBC20051208
 
 
-class SFTPTransportTestRelative(TestCaseWithSFTPServer):
+class SFTPTransportTestRelativeRoot(TestCaseWithSFTPServer):
     """Test the SFTP transport with homedir based relative paths."""
 
     def setUp(self):
         self._get_remote_is_absolute = False
-        super(SFTPTransportTestRelative, self).setUp()
+        super(SFTPTransportTestRelativeRoot, self).setUp()
 
     def test__remote_path_relative_root(self):
         # relative paths are preserved
         t = self.get_transport('')
+        # the remote path should be ''
+        self.assertEqual('', t._path)
         self.assertEqual('a', t._remote_path('a'))
 
 
@@ -192,8 +195,37 @@
                               clone_from=fake)
             self.fail('expected exception not raised')
         except TransportError, e:
-            self.assertEquals(str(e), 
-                    'Transport error: ~janneke: invalid port number ')
+            self.assertEquals(str(e),
+                    'Transport error: '
+                    'invalid port number ~janneke in url:\n'
+                    'sftp://lilypond.org:~janneke/public_html/bzr/gub ')
+
+    def test_get_paramiko_vendor(self):
+        """Test that if no 'ssh' is available we get builtin paramiko"""
+        from bzrlib.transport import ssh
+        # set '.' as the only location in the path, forcing no 'ssh' to exist
+        orig_vendor = ssh._ssh_vendor
+        orig_path = set_or_unset_env('PATH', '.')
+        try:
+            # No vendor defined yet, query for one
+            ssh._ssh_vendor = None
+            vendor = ssh._get_ssh_vendor()
+            self.assertIsInstance(vendor, ssh.ParamikoVendor)
+        finally:
+            set_or_unset_env('PATH', orig_path)
+            ssh._ssh_vendor = orig_vendor
+
+    def test_abspath_root_sibling_server(self):
+        from bzrlib.transport.sftp import SFTPSiblingAbsoluteServer
+        server = SFTPSiblingAbsoluteServer()
+        server.setUp()
+        try:
+            transport = get_transport(server.get_url())
+            self.assertFalse(transport.abspath('/').endswith('/~/'))
+            self.assertTrue(transport.abspath('/').endswith('/'))
+            del transport
+        finally:
+            server.tearDown()
 
 
 class SFTPBranchTest(TestCaseWithSFTPServer):
@@ -403,6 +435,8 @@
 
     def setUp(self):
         TestCase.setUp(self)
+        if not paramiko_loaded:
+            raise TestSkipped('you must have paramiko to run this test')
 
     def test_delay(self):
         from bzrlib.transport.sftp import SocketDelay

=== added file 'bzrlib/tests/test_smart_transport.py'
--- bzrlib/tests/test_smart_transport.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_smart_transport.py	2006-09-15 12:31:16 +0000
@@ -0,0 +1,309 @@
+# Copyright (C) 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for smart transport"""
+
+# all of this deals with byte strings so this is safe
+from cStringIO import StringIO
+import subprocess
+import sys
+
+import bzrlib
+from bzrlib import (
+        bzrdir,
+        errors,
+        tests,
+        )
+from bzrlib.transport import (
+        get_transport,
+        local,
+        memory,
+        smart,
+        )
+
+
+class SmartClientTests(tests.TestCase):
+
+    def test_construct_smart_stream_client(self):
+        # make a new client; this really wants a connector function that returns
+        # two fifos or sockets but the constructor should not do any IO
+        client = smart.SmartStreamClient(None)
+
+
+class TCPClientTests(tests.TestCaseWithTransport):
+
+    def setUp(self):
+        super(TCPClientTests, self).setUp()
+        # We're allowed to set  the transport class here, so that we don't use
+        # the default or a parameterized class, but rather use the
+        # TestCaseWithTransport infrastructure to set up a smart server and
+        # transport.
+        self.transport_server = smart.SmartTCPServer_for_testing
+
+    def test_plausible_url(self):
+        self.assert_(self.get_url().startswith('bzr://'))
+
+    def test_probe_transport(self):
+        t = self.get_transport()
+        self.assertIsInstance(t, smart.SmartTransport)
+
+    def test_get_client_from_transport(self):
+        t = self.get_transport()
+        client = t.get_smart_client()
+        self.assertIsInstance(client, smart.SmartStreamClient)
+
+
+class BasicSmartTests(tests.TestCase):
+    
+    def test_smart_query_version(self):
+        """Feed a canned query version to a server"""
+        to_server = StringIO('hello\n')
+        from_server = StringIO()
+        server = smart.SmartStreamServer(to_server, from_server, local.LocalTransport('file:///'))
+        server._serve_one_request()
+        self.assertEqual('ok\0011\n',
+                         from_server.getvalue())
+
+    def test_canned_get_response(self):
+        transport = memory.MemoryTransport('memory:///')
+        transport.put_bytes('testfile', 'contents\nof\nfile\n')
+        to_server = StringIO('get\001./testfile\n')
+        from_server = StringIO()
+        server = smart.SmartStreamServer(to_server, from_server, transport)
+        server._serve_one_request()
+        self.assertEqual('ok\n'
+                         '17\n'
+                         'contents\nof\nfile\n'
+                         'done\n',
+                         from_server.getvalue())
+
+    def test_get_error_unexpected(self):
+        """Error reported by server with no specific representation"""
+        class FlakyTransport(object):
+            def get_bytes(self, path):
+                raise Exception("some random exception from inside server")
+        server = smart.SmartTCPServer(backing_transport=FlakyTransport())
+        server.start_background_thread()
+        try:
+            transport = smart.SmartTCPTransport(server.get_url())
+            try:
+                transport.get('something')
+            except errors.TransportError, e:
+                self.assertContainsRe(str(e), 'some random exception')
+            else:
+                self.fail("get did not raise expected error")
+        finally:
+            server.stop_background_thread()
+
+    def test_server_subprocess(self):
+        """Talk to a server started as a subprocess
+        
+        This is similar to running it over ssh, except that it runs in the same machine 
+        without ssh intermediating.
+        """
+        args = [sys.executable, sys.argv[0], 'serve', '--inet']
+        child = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
+                                 close_fds=True)
+        conn = smart.SmartStreamClient(lambda: (child.stdout, child.stdin))
+        conn.query_version()
+        conn.query_version()
+        conn.disconnect()
+        returncode = child.wait()
+        self.assertEquals(0, returncode)
+
+
+class SmartTCPTests(tests.TestCase):
+    """Tests for connection to TCP server.
+    
+    All of these tests are run with a server running on another thread serving
+    a MemoryTransport, and a connection to it already open.
+    """
+
+    def setUp(self):
+        super(SmartTCPTests, self).setUp()
+        self.backing_transport = memory.MemoryTransport()
+        self.server = smart.SmartTCPServer(self.backing_transport)
+        self.server.start_background_thread()
+        self.transport = smart.SmartTCPTransport(self.server.get_url())
+
+    def tearDown(self):
+        if getattr(self, 'transport', None):
+            self.transport.disconnect()
+        if getattr(self, 'server', None):
+            self.server.stop_background_thread()
+        super(SmartTCPTests, self).tearDown()
+        
+    def test_start_tcp_server(self):
+        url = self.server.get_url()
+        self.assertContainsRe(url, r'^bzr://127\.0\.0\.1:[0-9]{2,}/')
+
+    def test_smart_transport_has(self):
+        """Checking for file existence over smart."""
+        self.backing_transport.put_bytes("foo", "contents of foo\n")
+        self.assertTrue(self.transport.has("foo"))
+        self.assertFalse(self.transport.has("non-foo"))
+
+    def test_smart_transport_get(self):
+        """Read back a file over smart."""
+        self.backing_transport.put_bytes("foo", "contents\nof\nfoo\n")
+        fp = self.transport.get("foo")
+        self.assertEqual('contents\nof\nfoo\n', fp.read())
+        
+    def test_get_error_enoent(self):
+        """Error reported from server getting nonexistent file."""
+        # The path in a raised NoSuchFile exception should be the precise path
+        # asked for by the client. This gives meaningful and unsurprising errors
+        # for users.
+        try:
+            self.transport.get('not%20a%20file')
+        except errors.NoSuchFile, e:
+            self.assertEqual('not%20a%20file', e.path)
+        else:
+            self.fail("get did not raise expected error")
+
+    def test_simple_clone_conn(self):
+        """Test that cloning reuses the same connection."""
+        # we create a real connection not a loopback one, but it will use the
+        # same server and pipes
+        conn2 = self.transport.clone('.')
+        self.assertTrue(self.transport._client is conn2._client)
+
+    def test__remote_path(self):
+        self.assertEquals('/foo/bar',
+                          self.transport._remote_path('foo/bar'))
+
+    def test_clone_changes_base(self):
+        """Cloning transport produces one with a new base location"""
+        conn2 = self.transport.clone('subdir')
+        self.assertEquals(self.transport.base + 'subdir/',
+                          conn2.base)
+
+    def test_open_dir(self):
+        """Test changing directory"""
+        transport = self.transport
+        self.backing_transport.mkdir('toffee')
+        self.backing_transport.mkdir('toffee/apple')
+        self.assertEquals('/toffee', transport._remote_path('toffee'))
+        toffee_trans = transport.clone('toffee')
+        # Check that each transport has only the contents of its directory
+        # directly visible. If state was being held in the wrong object, it's
+        # conceivable that cloning a transport would alter the state of the
+        # cloned-from transport.
+        self.assertTrue(transport.has('toffee'))
+        self.assertFalse(toffee_trans.has('toffee'))
+        self.assertFalse(transport.has('apple'))
+        self.assertTrue(toffee_trans.has('apple'))
+
+    def test_open_bzrdir(self):
+        """Open an existing bzrdir over smart transport"""
+        transport = self.transport
+        t = self.backing_transport
+        bzrdir.BzrDirFormat.get_default_format().initialize_on_transport(t)
+        result_dir = bzrdir.BzrDir.open_containing_from_transport(transport)
+
+
+class SmartServerTests(tests.TestCaseWithTransport):
+    """Test that call directly into the server logic, bypassing the network."""
+
+    def test_hello(self):
+        server = smart.SmartServer(None)
+        response = server.dispatch_command('hello', ())
+        self.assertEqual(('ok', '1'), response.args)
+        self.assertEqual(None, response.body)
+        
+    def test_get_bundle(self):
+        from bzrlib.bundle import serializer
+        wt = self.make_branch_and_tree('.')
+        self.build_tree_contents([('hello', 'hello world')])
+        wt.add('hello')
+        rev_id = wt.commit('add hello')
+        
+        server = smart.SmartServer(self.get_transport())
+        response = server.dispatch_command('get_bundle', ('.', rev_id))
+        bundle = serializer.read_bundle(StringIO(response.body))
+
+
+class SmartTransportRegistration(tests.TestCase):
+
+    def test_registration(self):
+        t = get_transport('bzr+ssh://example.com/path')
+        self.assertIsInstance(t, smart.SmartSSHTransport)
+        self.assertEqual('example.com', t._host)
+
+
+class FakeClient(smart.SmartStreamClient):
+    """Emulate a client for testing a transport's use of the client."""
+
+    def __init__(self):
+        smart.SmartStreamClient.__init__(self, None)
+        self._calls = []
+
+    def _call(self, *args):
+        self._calls.append(('_call', args))
+        return ('ok', )
+
+    def _recv_bulk(self):
+        return 'bar'
+
+
+class TestSmartTransport(tests.TestCase):
+        
+    def test_use_connection_factory(self):
+        # We want to be able to pass a client as a parameter to SmartTransport.
+        client = FakeClient()
+        transport = smart.SmartTransport('bzr://localhost/', client=client)
+
+        # We want to make sure the client is used when the first remote
+        # method is called.  No method should have been called yet.
+        self.assertEqual([], client._calls)
+
+        # Now call a method that should result in a single request.
+        self.assertEqual('bar', transport.get_bytes('foo'))
+        # The only call to _call should have been to get /foo.
+        self.assertEqual([('_call', ('get', '/foo'))], client._calls)
+
+
+# TODO: Client feature that does get_bundle and then installs that into a
+# branch; this can be used in place of the regular pull/fetch operation when
+# coming from a smart server.
+#
+# TODO: Eventually, want to do a 'branch' command by fetching the whole
+# history as one big bundle.  How?  
+#
+# The branch command does 'br_from.sprout', which tries to preserve the same
+# format.  We don't necessarily even want that.  
+#
+# It might be simpler to handle cmd_pull first, which does a simpler fetch()
+# operation from one branch into another.  It already has some code for
+# pulling from a bundle, which it does by trying to see if the destination is
+# a bundle file.  So it seems the logic for pull ought to be:
+# 
+#  - if it's a smart server, get a bundle from there and install that
+#  - if it's a bundle, install that
+#  - if it's a branch, pull from there
+#
+# Getting a bundle from a smart server is a bit different from reading a
+# bundle from a URL:
+#
+#  - we can reasonably remember the URL we last read from 
+#  - you can specify a revision number to pull, and we need to pass it across
+#    to the server as a limit on what will be requested
+#
+# TODO: Given a URL, determine whether it is a smart server or not (or perhaps
+# otherwise whether it's a bundle?)  Should this be a property or method of
+# the transport?  For the ssh protocol, we always know it's a smart server.
+# For http, we potentially need to probe.  But if we're explicitly given
+# bzr+http:// then we can skip that for now. 

=== modified file 'bzrlib/tests/test_transport.py'
--- bzrlib/tests/test_transport.py	2006-09-05 21:25:15 +0000
+++ bzrlib/tests/test_transport.py	2006-09-13 12:32:47 +0000
@@ -168,11 +168,17 @@
     def test_clone(self):
         transport = MemoryTransport()
         self.assertTrue(isinstance(transport, MemoryTransport))
+        self.assertEqual("memory:///", transport.clone("/").base)
 
     def test_abspath(self):
         transport = MemoryTransport()
         self.assertEqual("memory:///relpath", transport.abspath('relpath'))
 
+    def test_abspath_of_root(self):
+        transport = MemoryTransport()
+        self.assertEqual("memory:///", transport.base)
+        self.assertEqual("memory:///", transport.abspath('/'))
+
     def test_relpath(self):
         transport = MemoryTransport()
 

=== modified file 'bzrlib/tests/test_transport_implementations.py'
--- bzrlib/tests/test_transport_implementations.py	2006-09-08 18:52:17 +0000
+++ bzrlib/tests/test_transport_implementations.py	2006-09-16 13:36:47 +0000
@@ -26,18 +26,19 @@
 import sys
 
 from bzrlib import (
+    errors,
     osutils,
     urlutils,
     )
 from bzrlib.errors import (DirectoryNotEmpty, NoSuchFile, FileExists,
-                           LockError, PathError,
+                           LockError, NoSmartServer, PathError,
                            TransportNotPossible, ConnectionError,
                            InvalidURL)
 from bzrlib.osutils import getcwd
 from bzrlib.symbol_versioning import zero_eleven
 from bzrlib.tests import TestCaseInTempDir, TestSkipped
 from bzrlib.tests.test_transport import TestTransportImplementation
-from bzrlib.transport import memory
+from bzrlib.transport import memory, smart
 import bzrlib.transport
 
 
@@ -91,6 +92,13 @@
         self.assertEqual(False, t.has_any(['c', 'c', 'c']))
         self.assertEqual(True, t.has_any(['b', 'b', 'b']))
 
+    def test_has_root_works(self):
+        current_transport = self.get_transport()
+        # import pdb;pdb.set_trace()
+        self.assertTrue(current_transport.has('/'))
+        root = current_transport.clone('/')
+        self.assertTrue(root.has(''))
+
     def test_get(self):
         t = self.get_transport()
 
@@ -144,6 +152,11 @@
                              t.put, 'b', StringIO('file-like\ncontents\n'))
         self.check_transport_contents('file-like\ncontents\n', t, 'b')
 
+        self.assertRaises(NoSuchFile,
+            self.applyDeprecated,
+            zero_eleven,
+            t.put, 'path/doesnt/exist/c', StringIO('contents'))
+
     def test_put_bytes(self):
         t = self.get_transport()
 
@@ -400,6 +413,37 @@
         self.check_transport_contents('diff\ncontents for\na\n', t, 'a')
         self.check_transport_contents('another contents\nfor d\n', t, 'd')
 
+    def test_put_permissions(self):
+        t = self.get_transport()
+
+        if t.is_readonly():
+            return
+        if not t._can_roundtrip_unix_modebits():
+            # Can't roundtrip, so no need to run this test
+            return
+        self.applyDeprecated(zero_eleven, t.put, 'mode644',
+                             StringIO('test text\n'), mode=0644)
+        self.assertTransportMode(t, 'mode644', 0644)
+        self.applyDeprecated(zero_eleven, t.put, 'mode666',
+                             StringIO('test text\n'), mode=0666)
+        self.assertTransportMode(t, 'mode666', 0666)
+        self.applyDeprecated(zero_eleven, t.put, 'mode600',
+                             StringIO('test text\n'), mode=0600)
+        self.assertTransportMode(t, 'mode600', 0600)
+        # Yes, you can put a file such that it becomes readonly
+        self.applyDeprecated(zero_eleven, t.put, 'mode400',
+                             StringIO('test text\n'), mode=0400)
+        self.assertTransportMode(t, 'mode400', 0400)
+        self.applyDeprecated(zero_eleven, t.put_multi,
+                             [('mmode644', StringIO('text\n'))], mode=0644)
+        self.assertTransportMode(t, 'mmode644', 0644)
+
+        # The default permissions should be based on the current umask
+        umask = osutils.get_umask()
+        self.applyDeprecated(zero_eleven, t.put, 'nomode',
+                             StringIO('test text\n'), mode=None)
+        self.assertTransportMode(t, 'nomode', 0666 & ~umask)
+        
     def test_mkdir(self):
         t = self.get_transport()
 
@@ -751,6 +795,8 @@
         t.mkdir('adir/asubdir')
         t.mkdir('bdir')
         t.mkdir('bdir/bsubdir')
+        # any kind of PathError would be OK, though we normally expect
+        # DirectoryNotEmpty
         self.assertRaises(PathError, t.rename, 'bdir', 'adir')
         # nothing was changed so it should still be as before
         self.assertTrue(t.has('bdir/bsubdir'))
@@ -841,7 +887,11 @@
         # TODO: test copy_multi
 
     def test_connection_error(self):
-        """ConnectionError is raised when connection is impossible"""
+        """ConnectionError is raised when connection is impossible.
+        
+        The error may be raised from either the constructor or the first
+        operation on the transport.
+        """
         try:
             url = self._server.get_bogus_url()
         except NotImplementedError:
@@ -992,6 +1042,44 @@
         self.failUnless(t2.has('d'))
         self.failUnless(t3.has('b/d'))
 
+    def test_clone_to_root(self):
+        orig_transport = self.get_transport()
+        # Repeatedly go up to a parent directory until we're at the root
+        # directory of this transport
+        root_transport = orig_transport
+        new_transport = root_transport.clone("..")
+        # as we are walking up directories, the path must be must be 
+        # growing less, except at the top
+        self.assertTrue(len(new_transport.base) < len(root_transport.base)
+            or new_transport.base == root_transport.base)
+        while new_transport.base != root_transport.base:
+            root_transport = new_transport
+            new_transport = root_transport.clone("..")
+            # as we are walking up directories, the path must be must be 
+            # growing less, except at the top
+            self.assertTrue(len(new_transport.base) < len(root_transport.base)
+                or new_transport.base == root_transport.base)
+
+        # Cloning to "/" should take us to exactly the same location.
+        self.assertEqual(root_transport.base, orig_transport.clone("/").base)
+        # the abspath of "/" from the original transport should be the same
+        # as the base at the root:
+        self.assertEqual(orig_transport.abspath("/"), root_transport.base)
+
+        # At the root, the URL must still end with / as its a directory
+        self.assertEqual(root_transport.base[-1], '/')
+
+    def test_clone_from_root(self):
+        """At the root, cloning to a simple dir should just do string append."""
+        orig_transport = self.get_transport()
+        root_transport = orig_transport.clone('/')
+        self.assertEqual(root_transport.base + '.bzr/',
+            root_transport.clone('.bzr').base)
+
+    def test_base_url(self):
+        t = self.get_transport()
+        self.assertEqual('/', t.base[-1])
+
     def test_relpath(self):
         t = self.get_transport()
         self.assertEqual('', t.relpath(t.base))
@@ -1020,11 +1108,15 @@
         # specific test cases.
         transport = self.get_transport()
         
-        # disabled because some transports might normalize urls in generating
-        # the abspath - eg http+pycurl-> just http -- mbp 20060308 
         self.assertEqual(transport.base + 'relpath',
                          transport.abspath('relpath'))
 
+        # This should work without raising an error.
+        transport.abspath("/")
+
+        # the abspath of "/" and "/foo/.." should result in the same location
+        self.assertEqual(transport.abspath("/"), transport.abspath("/foo/.."))
+
     def test_local_abspath(self):
         transport = self.get_transport()
         try:
@@ -1156,23 +1248,37 @@
         self.check_transport_contents('bar', transport2, 'foo')
 
     def test_lock_write(self):
+        """Test transport-level write locks.
+
+        These are deprecated and transports may decline to support them.
+        """
         transport = self.get_transport()
         if transport.is_readonly():
             self.assertRaises(TransportNotPossible, transport.lock_write, 'foo')
             return
         transport.put_bytes('lock', '')
-        lock = transport.lock_write('lock')
+        try:
+            lock = transport.lock_write('lock')
+        except TransportNotPossible:
+            return
         # TODO make this consistent on all platforms:
         # self.assertRaises(LockError, transport.lock_write, 'lock')
         lock.unlock()
 
     def test_lock_read(self):
+        """Test transport-level read locks.
+
+        These are deprecated and transports may decline to support them.
+        """
         transport = self.get_transport()
         if transport.is_readonly():
             file('lock', 'w').close()
         else:
             transport.put_bytes('lock', '')
-        lock = transport.lock_read('lock')
+        try:
+            lock = transport.lock_read('lock')
+        except TransportNotPossible:
+            return
         # TODO make this consistent on all platforms:
         # self.assertRaises(LockError, transport.lock_read, 'lock')
         lock.unlock()
@@ -1202,3 +1308,36 @@
         self.assertEqual(d[1], (9, '9'))
         self.assertEqual(d[2], (0, '0'))
         self.assertEqual(d[3], (3, '34'))
+
+    def test_get_smart_client(self):
+        """All transports must either give a smart client, or know they can't.
+
+        For some transports such as http this might depend on probing to see 
+        what's actually present on the other end.  (But we can adjust for that 
+        in the future.)
+        """
+        transport = self.get_transport()
+        try:
+            client = transport.get_smart_client()
+            # XXX: should be a more general class
+            self.assertIsInstance(client, smart.SmartStreamClient)
+        except NoSmartServer:
+            # as long as we got it we're fine
+            pass
+
+    def test_readv_short_read(self):
+        transport = self.get_transport()
+        if transport.is_readonly():
+            file('a', 'w').write('0123456789')
+        else:
+            transport.put_bytes('a', '01234567890')
+
+        # This is intentionally reading off the end of the file
+        # since we are sure that it cannot get there
+        self.assertListRaises((errors.ShortReadvError, AssertionError),
+                              transport.readv, 'a', [(1,1), (8,10)])
+
+        # This is trying to seek past the end of the file, it should
+        # also raise a special error
+        self.assertListRaises(errors.ShortReadvError,
+                              transport.readv, 'a', [(12,2)])

=== added file 'bzrlib/tests/test_treebuilder.py'
--- bzrlib/tests/test_treebuilder.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_treebuilder.py	2006-09-15 22:46:02 +0000
@@ -0,0 +1,94 @@
+# Copyright (C) 2006 Canonical Ltd
+# Authors:  Robert Collins <robert.collins@canonical.com>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for the TreeBuilder helper class."""
+
+from bzrlib import errors
+from bzrlib.memorytree import MemoryTree
+from bzrlib.tests import TestCaseWithTransport
+from bzrlib.treebuilder import TreeBuilder
+
+
+class FakeTree(object):
+    """A pretend tree to test the calls made by TreeBuilder."""
+
+    def __init__(self):
+        self._calls = []
+
+    def lock_tree_write(self):
+        self._calls.append("lock_tree_write")
+
+    def unlock(self):
+        self._calls.append("unlock")
+
+
+class TestFakeTree(TestCaseWithTransport):
+    
+    def testFakeTree(self):
+        """Check that FakeTree works as required for the TreeBuilder tests."""
+        tree = FakeTree()
+        self.assertEqual([], tree._calls)
+        tree.lock_tree_write()
+        self.assertEqual(["lock_tree_write"], tree._calls)
+        tree.unlock()
+        self.assertEqual(["lock_tree_write", "unlock"], tree._calls)
+
+
+class TestTreeBuilder(TestCaseWithTransport):
+    
+    def test_create(self):
+        builder = TreeBuilder()
+
+    def test_start_tree_locks_write(self):
+        builder = TreeBuilder()
+        tree = FakeTree()
+        builder.start_tree(tree)
+        self.assertEqual(["lock_tree_write"], tree._calls)
+
+    def test_start_tree_when_started_fails(self):
+        builder = TreeBuilder()
+        tree = FakeTree()
+        builder.start_tree(tree)
+        self.assertRaises(errors.AlreadyBuilding, builder.start_tree, tree)
+        
+    def test_finish_tree_not_started_errors(self):
+        builder = TreeBuilder()
+        self.assertRaises(errors.NotBuilding, builder.finish_tree)
+
+    def test_finish_tree_unlocks(self):
+        builder = TreeBuilder()
+        tree = FakeTree()
+        builder.start_tree(tree)
+        builder.finish_tree()
+        self.assertEqual(["lock_tree_write", "unlock"], tree._calls)
+
+    def test_build_tree_not_started_errors(self):
+        builder = TreeBuilder()
+        self.assertRaises(errors.NotBuilding, builder.build, "foo")
+
+    def test_build_tree(self):
+        """Test building works using a MemoryTree."""
+        branch = self.make_branch('branch')
+        tree = MemoryTree.create_on_branch(branch)
+        builder = TreeBuilder()
+        builder.start_tree(tree)
+        builder.build(['foo', "bar/", "bar/file"])
+        self.assertEqual('contents of foo\n',
+            tree.get_file(tree.path2id('foo')).read())
+        self.assertEqual('contents of bar/file\n',
+            tree.get_file(tree.path2id('bar/file')).read())
+        builder.finish_tree()

=== modified file 'bzrlib/tests/test_workingtree.py'
--- bzrlib/tests/test_workingtree.py	2006-09-10 22:50:11 +0000
+++ bzrlib/tests/test_workingtree.py	2006-09-15 02:03:15 +0000
@@ -25,12 +25,12 @@
 from bzrlib.bzrdir import BzrDir
 from bzrlib.errors import NotBranchError, NotVersionedError
 from bzrlib.lockdir import LockDir
+from bzrlib.mutabletree import needs_tree_write_lock
 from bzrlib.osutils import pathjoin, getcwd, has_symlinks
 from bzrlib.tests import TestCase, TestCaseWithTransport, TestSkipped
 from bzrlib.trace import mutter
 from bzrlib.transport import get_transport
 from bzrlib.workingtree import (
-    needs_tree_write_lock,
     TreeEntry,
     TreeDirectory,
     TreeFile,

=== modified file 'bzrlib/tests/workingtree_implementations/__init__.py'
--- bzrlib/tests/workingtree_implementations/__init__.py	2006-09-07 07:31:51 +0000
+++ bzrlib/tests/workingtree_implementations/__init__.py	2006-09-07 23:31:28 +0000
@@ -63,6 +63,8 @@
         'bzrlib.tests.workingtree_implementations.test_merge_from_branch',
         'bzrlib.tests.workingtree_implementations.test_parents',
         'bzrlib.tests.workingtree_implementations.test_pull',
+        'bzrlib.tests.workingtree_implementations.test_put_file',
+        'bzrlib.tests.workingtree_implementations.test_mkdir',
         'bzrlib.tests.workingtree_implementations.test_unversion',
         'bzrlib.tests.workingtree_implementations.test_workingtree',
         ]

=== added file 'bzrlib/tests/workingtree_implementations/test_mkdir.py'
--- bzrlib/tests/workingtree_implementations/test_mkdir.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/workingtree_implementations/test_mkdir.py	2006-09-07 23:13:08 +0000
@@ -0,0 +1,34 @@
+# (C) 2006 Canonical Ltd
+# Authors:  Robert Collins <robert.collins@canonical.com>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for interface conformance of 'workingtree.put_mkdir'"""
+
+from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
+
+
+class TestMkdir(TestCaseWithWorkingTree):
+
+    def test_mkdir_no_id(self):
+        t = self.make_branch_and_tree('t1')
+        file_id = t.mkdir('path')
+        self.assertEqual('directory', t.kind(file_id))
+
+    def test_mkdir_with_id(self):
+        t = self.make_branch_and_tree('t1')
+        file_id = t.mkdir('path', 'my-id')
+        self.assertEqual('my-id', file_id)
+        self.assertEqual('directory', t.kind(file_id))

=== added file 'bzrlib/tests/workingtree_implementations/test_put_file.py'
--- bzrlib/tests/workingtree_implementations/test_put_file.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/workingtree_implementations/test_put_file.py	2006-09-07 23:13:08 +0000
@@ -0,0 +1,37 @@
+# (C) 2006 Canonical Ltd
+# Authors:  Robert Collins <robert.collins@canonical.com>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for interface conformance of 'workingtree.put_file*'"""
+
+from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
+
+
+class TestPutFileBytesNonAtomic(TestCaseWithWorkingTree):
+
+    def test_put_new_file(self):
+        t = self.make_branch_and_tree('t1')
+        t.add(['foo'], ids=['foo-id'], kinds=['file'])
+        t.put_file_bytes_non_atomic('foo-id', 'barshoom')
+        self.assertEqual('barshoom', t.get_file('foo-id').read())
+
+    def test_put_existing_file(self):
+        t = self.make_branch_and_tree('t1')
+        t.add(['foo'], ids=['foo-id'], kinds=['file'])
+        t.put_file_bytes_non_atomic('foo-id', 'first-content')
+        t.put_file_bytes_non_atomic('foo-id', 'barshoom')
+        self.assertEqual('barshoom', t.get_file('foo-id').read())
+

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2006-09-08 18:52:17 +0000
+++ bzrlib/transport/__init__.py	2006-09-16 02:14:14 +0000
@@ -165,7 +165,8 @@
             port = int(port)
         except ValueError:
             # TODO: Should this be ConnectionError?
-            raise errors.TransportError('%s: invalid port number' % port)
+            raise errors.TransportError(
+                'invalid port number %s in url:\n%s' % (port, url))
     host = urllib.unquote(host)
 
     path = urllib.unquote(path)
@@ -198,6 +199,8 @@
     implementations can do pipelining.
     In general implementations should support having a generator or a list
     as an argument (ie always iterate, never index)
+
+    :ivar base: Base URL for the transport; should always end in a slash.
     """
 
     # implementations can override this if it is more efficient
@@ -304,13 +307,60 @@
 
     def abspath(self, relpath):
         """Return the full url to the given relative path.
-        This can be supplied with a string or a list
 
-        XXX: Robert Collins 20051016 - is this really needed in the public
-             interface ?
+        :param relpath: a string of a relative path
         """
+
+        # XXX: Robert Collins 20051016 - is this really needed in the public
+        # interface ?
         raise NotImplementedError(self.abspath)
 
+    def _combine_paths(self, base_path, relpath):
+        """Transform a Transport-relative path to a remote absolute path.
+
+        This does not handle substitution of ~ but does handle '..' and '.'
+        components.
+
+        Examples::
+
+            >>> t = Transport('/')
+            >>> t._combine_paths('/home/sarah', 'project/foo')
+            '/home/sarah/project/foo'
+            >>> t._combine_paths('/home/sarah', '../../etc')
+            '/etc'
+
+        :param base_path: urlencoded path for the transport root; typically a 
+             URL but need not contain scheme/host/etc.
+        :param relpath: relative url string for relative part of remote path.
+        :return: urlencoded string for final path.
+        """
+        # FIXME: share the common code across more transports; variants of
+        # this likely occur in http and sftp too.
+        #
+        # TODO: Also need to consider handling of ~, which might vary between
+        # transports?
+        if not isinstance(relpath, str):
+            raise errors.InvalidURL("not a valid url: %r" % relpath)
+        if relpath.startswith('/'):
+            base_parts = []
+        else:
+            base_parts = base_path.split('/')
+        if len(base_parts) > 0 and base_parts[-1] == '':
+            base_parts = base_parts[:-1]
+        for p in relpath.split('/'):
+            if p == '..':
+                if len(base_parts) == 0:
+                    # In most filesystems, a request for the parent
+                    # of root, just returns root.
+                    continue
+                base_parts.pop()
+            elif p == '.':
+                continue # No-op
+            elif p != '':
+                base_parts.append(p)
+        path = '/'.join(base_parts)
+        return path
+
     def relpath(self, abspath):
         """Return the local path portion from a given absolute path.
 
@@ -342,6 +392,8 @@
         Note that some transports MAY allow querying on directories, but this
         is not part of the protocol.  In other words, the results of 
         t.has("a_directory_name") are undefined.
+
+        :rtype: bool
         """
         raise NotImplementedError(self.has)
 
@@ -378,6 +430,7 @@
         """Get the file at the given relative path.
 
         :param relpath: The relative path to the file
+        :rtype: File-like object.
         """
         raise NotImplementedError(self.get)
 
@@ -388,6 +441,13 @@
         """
         return self.get(relpath).read()
 
+    def get_smart_client(self):
+        """Return a smart client for this transport if possible.
+
+        :raises NoSmartServer: if no smart server client is available.
+        """
+        raise errors.NoSmartServer(self.base)
+
     def readv(self, relpath, offsets):
         """Get parts of the file at the given relative path.
 
@@ -398,9 +458,9 @@
             return
 
         fp = self.get(relpath)
-        return self._seek_and_read(fp, offsets)
+        return self._seek_and_read(fp, offsets, relpath)
 
-    def _seek_and_read(self, fp, offsets):
+    def _seek_and_read(self, fp, offsets, relpath='<unknown>'):
         """An implementation of readv that uses fp.seek and fp.read.
 
         This uses _coalesce_offsets to issue larger reads and fewer seeks.
@@ -428,6 +488,9 @@
             #       benchmarked.
             fp.seek(c_offset.start)
             data = fp.read(c_offset.length)
+            if len(data) < c_offset.length:
+                raise errors.ShortReadvError(relpath, c_offset.start,
+                            c_offset.length, actual=len(data))
             for suboffset, subsize in c_offset.ranges:
                 key = (c_offset.start+suboffset, subsize)
                 data_map[key] = data[suboffset:suboffset+subsize]
@@ -634,11 +697,15 @@
         return self.append_file(relpath, f, mode=mode)
 
     def append_file(self, relpath, f, mode=None):
-        """Append the text in the file-like object to the supplied location.
-
-        returns the length of relpath before the content was written to it.
-        
-        If the file does not exist, it is created with the supplied mode.
+        """Append bytes from a file-like object to a file at relpath.
+
+        The file is created if it does not already exist.
+
+        :param f: a file-like object of the bytes to append.
+        :param mode: Unix mode for newly created files.  This is not used for
+            existing files.
+
+        :returns: the length of relpath before the content was written to it.
         """
         symbol_versioning.warn('Transport %s should implement append_file,'
                                ' rather than implementing append() as of'
@@ -648,11 +715,16 @@
         return self.append(relpath, f, mode=mode)
 
     def append_bytes(self, relpath, bytes, mode=None):
-        """Append the text in the string object to the supplied location.
-
-        returns the length of relpath before the content was written to it.
-        
-        If the file does not exist, it is created with the supplied mode.
+        """Append bytes to a file at relpath.
+
+        The file is created if it does not already exist.
+
+        :type f: str
+        :param f: a string of the bytes to append.
+        :param mode: Unix mode for newly created files.  This is not used for
+            existing files.
+
+        :returns: the length of relpath before the content was written to it.
         """
         assert isinstance(bytes, str), \
             'bytes must be a plain string, not %s' % type(bytes)
@@ -848,26 +920,37 @@
         WARNING: many transports do not support this, so trying avoid using
         it if at all possible.
         """
-        raise errors.TransportNotPossible("This transport has not "
+        raise errors.TransportNotPossible("Transport %r has not "
                                           "implemented list_dir "
                                           "(but must claim to be listable "
-                                          "to trigger this error).")
+                                          "to trigger this error)."
+                                          % (self))
 
     def lock_read(self, relpath):
         """Lock the given file for shared (read) access.
-        WARNING: many transports do not support this, so trying avoid using it
+
+        WARNING: many transports do not support this, so trying avoid using it.
+        These methods may be removed in the future.
+
+        Transports may raise TransportNotPossible if OS-level locks cannot be
+        taken over this transport.  
 
         :return: A lock object, which should contain an unlock() function.
         """
-        raise NotImplementedError(self.lock_read)
+        raise errors.TransportNotPossible("transport locks not supported on %s" % self)
 
     def lock_write(self, relpath):
         """Lock the given file for exclusive (write) access.
-        WARNING: many transports do not support this, so trying avoid using it
+
+        WARNING: many transports do not support this, so trying avoid using it.
+        These methods may be removed in the future.
+
+        Transports may raise TransportNotPossible if OS-level locks cannot be
+        taken over this transport.
 
         :return: A lock object, which should contain an unlock() function.
         """
-        raise NotImplementedError(self.lock_write)
+        raise errors.TransportNotPossible("transport locks not supported on %s" % self)
 
     def is_readonly(self):
         """Return true if this connection cannot be written to."""
@@ -1083,6 +1166,12 @@
 register_lazy_transport('memory://', 'bzrlib.transport.memory', 'MemoryTransport')
 register_lazy_transport('readonly+', 'bzrlib.transport.readonly', 'ReadonlyTransportDecorator')
 register_lazy_transport('fakenfs+', 'bzrlib.transport.fakenfs', 'FakeNFSTransportDecorator')
-register_lazy_transport('vfat+', 
+register_lazy_transport('vfat+',
                         'bzrlib.transport.fakevfat',
                         'FakeVFATTransportDecorator')
+register_lazy_transport('bzr://',
+                        'bzrlib.transport.smart',
+                        'SmartTCPTransport')
+register_lazy_transport('bzr+ssh://',
+                        'bzrlib.transport.smart',
+                        'SmartSSHTransport')

=== modified file 'bzrlib/transport/decorator.py'
--- bzrlib/transport/decorator.py	2006-09-05 21:20:24 +0000
+++ bzrlib/transport/decorator.py	2006-09-13 13:43:21 +0000
@@ -88,6 +88,9 @@
         """See Transport.get()."""
         return self._decorated.get(relpath)
 
+    def get_smart_client(self):
+        return self._decorated.get_smart_client()
+
     def has(self, relpath):
         """See Transport.has()."""
         return self._decorated.has(relpath)

=== modified file 'bzrlib/transport/ftp.py'
--- bzrlib/transport/ftp.py	2006-09-08 18:52:17 +0000
+++ bzrlib/transport/ftp.py	2006-09-15 01:12:18 +0000
@@ -196,15 +196,13 @@
     def _abspath(self, relpath):
         assert isinstance(relpath, basestring)
         relpath = urlutils.unescape(relpath)
-        relpath_parts = relpath.split('/')
-        if len(relpath_parts) > 1:
-            if relpath_parts[0] == '':
-                raise ValueError("path %r within branch %r seems to be absolute"
-                                 % (relpath, self._path))
-        basepath = self._path.split('/')
+        if relpath.startswith('/'):
+            basepath = []
+        else:
+            basepath = self._path.split('/')
         if len(basepath) > 0 and basepath[-1] == '':
             basepath = basepath[:-1]
-        for p in relpath_parts:
+        for p in relpath.split('/'):
             if p == '..':
                 if len(basepath) == 0:
                     # In most filesystems, a request for the parent

=== modified file 'bzrlib/transport/http/__init__.py'
--- bzrlib/transport/http/__init__.py	2006-09-08 18:52:17 +0000
+++ bzrlib/transport/http/__init__.py	2006-09-16 02:14:14 +0000
@@ -169,19 +169,17 @@
         else:
             # TODO: Don't call this with an array - no magic interfaces
             relpath_parts = relpath[:]
-        if len(relpath_parts) > 1:
-            # TODO: Check that the "within branch" part of the
-            # error messages below is relevant in all contexts
-            if relpath_parts[0] == '':
-                raise ValueError("path %r within branch %r seems to be absolute"
-                                 % (relpath, self._path))
-            # read only transports never manipulate directories
-            if self.is_readonly() and relpath_parts[-1] == '':
+        if relpath.startswith('/'):
+            basepath = []
+        else:
+            # Except for the root, no trailing slashes are allowed
+            if len(relpath_parts) > 1 and relpath_parts[-1] == '':
                 raise ValueError("path %r within branch %r seems to be a directory"
                                  % (relpath, self._path))
-        basepath = self._path.split('/')
-        if len(basepath) > 0 and basepath[-1] == '':
-            basepath = basepath[:-1]
+            basepath = self._path.split('/')
+            if len(basepath) > 0 and basepath[-1] == '':
+                basepath = basepath[:-1]
+
         for p in relpath_parts:
             if p == '..':
                 if len(basepath) == 0:
@@ -254,7 +252,9 @@
             f.seek(start, (start < 0) and 2 or 0)
             start = f.tell()
             data = f.read(size)
-            assert len(data) == size
+            if len(data) != size:
+                raise errors.ShortReadvError(relpath, start, size,
+                                             actual=len(data))
             yield start, data
 
     @staticmethod

=== modified file 'bzrlib/transport/http/_urllib.py'
--- bzrlib/transport/http/_urllib.py	2006-09-08 18:46:29 +0000
+++ bzrlib/transport/http/_urllib.py	2006-09-12 10:54:46 +0000
@@ -72,9 +72,9 @@
                     e.errno, errno.errorcode.get(e.errno), path)
                 if e.errno == errno.ENOENT:
                     raise NoSuchFile(path, extra=e)
-            raise ConnectionError(msg = "Error retrieving %s: %s" 
-                             % (self.abspath(relpath), str(e)),
-                             orig_error=e)
+            raise ConnectionError(msg = "Error retrieving %s: %s"
+                                  % (self.abspath(relpath), str(e)),
+                                  orig_error=e)
 
     def _get_url_impl(self, url, method, ranges, tail_amount=0):
         """Actually pass get request into urllib
@@ -113,10 +113,13 @@
             f.read()
             f.close()
             return True
+        except urllib2.HTTPError, e:
+            mutter('url error code: %s, for has url: %r', e.code, abspath)
+            if e.code == 404:
+                return False
+            raise
         except urllib2.URLError, e:
-            mutter('url error code: %s for has url: %r', e.code, abspath)
-            if e.code == 404:
-                return False
+            mutter('url error: %s, for has url: %r', e.reason, abspath)
             raise
         except IOError, e:
             mutter('io error: %s %s for has url: %r',

=== modified file 'bzrlib/transport/memory.py'
--- bzrlib/transport/memory.py	2006-09-05 21:25:15 +0000
+++ bzrlib/transport/memory.py	2006-09-16 13:36:47 +0000
@@ -20,7 +20,6 @@
 so this is primarily useful for testing.
 """
 
-from copy import copy
 import os
 import errno
 import re
@@ -59,7 +58,9 @@
         if url[-1] != '/':
             url = url + '/'
         super(MemoryTransport, self).__init__(url)
-        self._cwd = url[url.find(':') + 3:]
+        split = url.find(':') + 3
+        self._scheme = url[:split]
+        self._cwd = url[split:]
         # dictionaries from absolute path to file mode
         self._dirs = {'/':None}
         self._files = {}
@@ -67,20 +68,10 @@
 
     def clone(self, offset=None):
         """See Transport.clone()."""
-        if offset is None or offset == '':
-            return copy(self)
-        segments = offset.split('/')
-        cwdsegments = self._cwd.split('/')[:-1]
-        while len(segments):
-            segment = segments.pop(0)
-            if segment == '.':
-                continue
-            if segment == '..':
-                if len(cwdsegments) > 1:
-                    cwdsegments.pop()
-                continue
-            cwdsegments.append(segment)
-        url = self.base[:self.base.find(':') + 3] + '/'.join(cwdsegments) + '/'
+        path = self._combine_paths(self._cwd, offset)
+        if len(path) == 0 or path[-1] != '/':
+            path += '/'
+        url = self._scheme + path
         result = MemoryTransport(url)
         result._dirs = self._dirs
         result._files = self._files
@@ -236,6 +227,10 @@
         relpath = urlutils.unescape(relpath)
         if relpath.find('..') != -1:
             raise AssertionError('relpath contains ..')
+        if relpath == '':
+            return '/'
+        if relpath[0] == '/':
+            return relpath
         if relpath == '.':
             if (self._cwd == '/'):
                 return self._cwd

=== modified file 'bzrlib/transport/sftp.py'
--- bzrlib/transport/sftp.py	2006-09-08 18:52:17 +0000
+++ bzrlib/transport/sftp.py	2006-09-16 13:36:47 +0000
@@ -17,26 +17,35 @@
 
 """Implementation of Transport over SFTP, using paramiko."""
 
+# TODO: Remove the transport-based lock_read and lock_write methods.  They'll
+# then raise TransportNotPossible, which will break remote access to any
+# formats which rely on OS-level locks.  That should be fine as those formats
+# are pretty old, but these combinations may have to be removed from the test
+# suite.  Those formats all date back to 0.7; so we should be able to remove
+# these methods when we officially drop support for those formats.
+
 import errno
 import os
 import random
 import select
 import socket
 import stat
-import subprocess
 import sys
 import time
 import urllib
 import urlparse
 import weakref
 
-from bzrlib.errors import (FileExists, 
+from bzrlib import (
+    errors,
+    urlutils,
+    )
+from bzrlib.errors import (FileExists,
                            NoSuchFile, PathNotChild,
                            TransportError,
-                           LockError, 
+                           LockError,
                            PathError,
                            ParamikoNotPresent,
-                           UnknownSSH,
                            )
 from bzrlib.osutils import pathjoin, fancy_rename, getcwd
 from bzrlib.trace import mutter, warning
@@ -47,7 +56,6 @@
     ssh,
     Transport,
     )
-import bzrlib.urlutils as urlutils
 
 try:
     import paramiko
@@ -85,8 +93,15 @@
 
 
 class SFTPLock(object):
-    """This fakes a lock in a remote location."""
+    """This fakes a lock in a remote location.
+    
+    A present lock is indicated just by the existence of a file.  This
+    doesn't work well on all transports and they are only used in 
+    deprecated storage formats.
+    """
+    
     __slots__ = ['path', 'lock_path', 'lock_file', 'transport']
+
     def __init__(self, path, transport):
         assert isinstance(transport, SFTPTransport)
 
@@ -119,7 +134,69 @@
             pass
 
 
-class SFTPTransport(Transport):
+class SFTPUrlHandling(Transport):
+    """Mix-in that does common handling of SSH/SFTP URLs."""
+
+    def __init__(self, base):
+        self._parse_url(base)
+        base = self._unparse_url(self._path)
+        if base[-1] != '/':
+            base += '/'
+        super(SFTPUrlHandling, self).__init__(base)
+
+    def _parse_url(self, url):
+        (self._scheme,
+         self._username, self._password,
+         self._host, self._port, self._path) = self._split_url(url)
+
+    def _unparse_url(self, path):
+        """Return a URL for a path relative to this transport.
+        """
+        path = urllib.quote(path)
+        # handle homedir paths
+        if not path.startswith('/'):
+            path = "/~/" + path
+        netloc = urllib.quote(self._host)
+        if self._username is not None:
+            netloc = '%s@%s' % (urllib.quote(self._username), netloc)
+        if self._port is not None:
+            netloc = '%s:%d' % (netloc, self._port)
+        return urlparse.urlunparse((self._scheme, netloc, path, '', '', ''))
+
+    def _split_url(self, url):
+        (scheme, username, password, host, port, path) = split_url(url)
+        ## assert scheme == 'sftp'
+
+        # the initial slash should be removed from the path, and treated
+        # as a homedir relative path (the path begins with a double slash
+        # if it is absolute).
+        # see draft-ietf-secsh-scp-sftp-ssh-uri-03.txt
+        # RBC 20060118 we are not using this as its too user hostile. instead
+        # we are following lftp and using /~/foo to mean '~/foo'.
+        # handle homedir paths
+        if path.startswith('/~/'):
+            path = path[3:]
+        elif path == '/~':
+            path = ''
+        return (scheme, username, password, host, port, path)
+
+    def abspath(self, relpath):
+        """Return the full url to the given relative path.
+        
+        @param relpath: the relative path or path components
+        @type relpath: str or list
+        """
+        return self._unparse_url(self._remote_path(relpath))
+    
+    def _remote_path(self, relpath):
+        """Return the path to be passed along the sftp protocol for relpath.
+        
+        :param relpath: is a urlencoded string.
+        """
+        return self._combine_paths(self._path, relpath)
+
+
+class SFTPTransport(SFTPUrlHandling):
     """Transport implementation for SFTP access."""
 
     _do_prefetch = _default_do_prefetch
@@ -141,11 +218,6 @@
     _max_request_size = 32768
 
     def __init__(self, base, clone_from=None):
-        assert base.startswith('sftp://')
-        self._parse_url(base)
-        base = self._unparse_url()
-        if base[-1] != '/':
-            base += '/'
         super(SFTPTransport, self).__init__(base)
         if clone_from is None:
             self._sftp_connect()
@@ -171,45 +243,117 @@
         else:
             return SFTPTransport(self.abspath(offset), self)
 
-    def abspath(self, relpath):
-        """
-        Return the full url to the given relative path.
-        
-        @param relpath: the relative path or path components
-        @type relpath: str or list
-        """
-        return self._unparse_url(self._remote_path(relpath))
-    
     def _remote_path(self, relpath):
         """Return the path to be passed along the sftp protocol for relpath.
         
         relpath is a urlencoded string.
+
+        :return: a path prefixed with / for regular abspath-based urls, or a
+            path that does not begin with / for urls which begin with /~/.
         """
-        # FIXME: share the common code across transports
+        # how does this work? 
+        # it processes relpath with respect to 
+        # our state:
+        # firstly we create a path to evaluate: 
+        # if relpath is an abspath or homedir path, its the entire thing
+        # otherwise we join our base with relpath
+        # then we eliminate all empty segments (double //'s) outside the first
+        # two elements of the list. This avoids problems with trailing 
+        # slashes, or other abnormalities.
+        # finally we evaluate the entire path in a single pass
+        # '.'s are stripped,
+        # '..' result in popping the left most already 
+        # processed path (which can never be empty because of the check for
+        # abspath and homedir meaning that its not, or that we've used our
+        # path. If the pop would pop the root, we ignore it.
+
+        # Specific case examinations:
+        # remove the special casefor ~: if the current root is ~/ popping of it
+        # = / thus our seed for a ~ based path is ['', '~']
+        # and if we end up with [''] then we had basically ('', '..') (which is
+        # '/..' so we append '' if the length is one, and assert that the first
+        # element is still ''. Lastly, if we end with ['', '~'] as a prefix for
+        # the output, we've got a homedir path, so we strip that prefix before
+        # '/' joining the resulting list.
+        #
+        # case one: '/' -> ['', ''] cannot shrink
+        # case two: '/' + '../foo' -> ['', 'foo'] (take '', '', '..', 'foo')
+        #           and pop the second '' for the '..', append 'foo'
+        # case three: '/~/' -> ['', '~', ''] 
+        # case four: '/~/' + '../foo' -> ['', '~', '', '..', 'foo'],
+        #           and we want to get '/foo' - the empty path in the middle
+        #           needs to be stripped, then normal path manipulation will 
+        #           work.
+        # case five: '/..' ['', '..'], we want ['', '']
+        #            stripping '' outside the first two is ok
+        #            ignore .. if its too high up
+        #
+        # lastly this code is possibly reusable by FTP, but not reusable by
+        # local paths: ~ is resolvable correctly, nor by HTTP or the smart
+        # server: ~ is resolved remotely.
+        # 
+        # however, a version of this that acts on self.base is possible to be
+        # written which manipulates the URL in canonical form, and would be
+        # reusable for all transports, if a flag for allowing ~/ at all was
+        # provided.
         assert isinstance(relpath, basestring)
-        relpath = urlutils.unescape(relpath).split('/')
-        basepath = self._path.split('/')
-        if len(basepath) > 0 and basepath[-1] == '':
-            basepath = basepath[:-1]
-
-        for p in relpath:
-            if p == '..':
-                if len(basepath) == 0:
-                    # In most filesystems, a request for the parent
-                    # of root, just returns root.
-                    continue
-                basepath.pop()
-            elif p == '.':
-                continue # No-op
-            else:
-                basepath.append(p)
-
-        path = '/'.join(basepath)
-        # mutter('relpath => remotepath %s => %s', relpath, path)
+        relpath = urlutils.unescape(relpath)
+
+        # case 1)
+        if relpath.startswith('/'):
+            # abspath - normal split is fine.
+            current_path = relpath.split('/')
+        elif relpath.startswith('~/'):
+            # root is homedir based: normal split and prefix '' to remote the
+            # special case
+            current_path = [''].extend(relpath.split('/'))
+        else:
+            # root is from the current directory:
+            if self._path.startswith('/'):
+                # abspath, take the regular split
+                current_path = []
+            else:
+                # homedir based, add the '', '~' not present in self._path
+                current_path = ['', '~']
+            # add our current dir
+            current_path.extend(self._path.split('/'))
+            # add the users relpath
+            current_path.extend(relpath.split('/'))
+        # strip '' segments that are not in the first one - the leading /.
+        to_process = current_path[:1]
+        for segment in current_path[1:]:
+            if segment != '':
+                to_process.append(segment)
+
+        # process '.' and '..' segments into output_path.
+        output_path = []
+        for segment in to_process:
+            if segment == '..':
+                # directory pop. Remove a directory 
+                # as long as we are not at the root
+                if len(output_path) > 1:
+                    output_path.pop()
+                # else: pass
+                # cannot pop beyond the root, so do nothing
+            elif segment == '.':
+                continue # strip the '.' from the output.
+            else:
+                # this will append '' to output_path for the root elements,
+                # which is appropriate: its why we strip '' in the first pass.
+                output_path.append(segment)
+
+        # check output special cases:
+        if output_path == ['']:
+            # [''] -> ['', '']
+            output_path = ['', '']
+        elif output_path[:2] == ['', '~']:
+            # ['', '~', ...] -> ...
+            output_path = output_path[2:]
+        path = '/'.join(output_path)
         return path
 
     def relpath(self, abspath):
-        username, password, host, port, path = self._split_url(abspath)
+        scheme, username, password, host, port, path = self._split_url(abspath)
         error = []
         if (username != self._username):
             error.append('username mismatch')
@@ -263,13 +407,13 @@
             fp = self._sftp.file(path, mode='rb')
             readv = getattr(fp, 'readv', None)
             if readv:
-                return self._sftp_readv(fp, offsets)
+                return self._sftp_readv(fp, offsets, relpath)
             mutter('seek and read %s offsets', len(offsets))
-            return self._seek_and_read(fp, offsets)
+            return self._seek_and_read(fp, offsets, relpath)
         except (IOError, paramiko.SSHException), e:
             self._translate_io_exception(e, path, ': error retrieving')
 
-    def _sftp_readv(self, fp, offsets):
+    def _sftp_readv(self, fp, offsets, relpath='<unknown>'):
         """Use the readv() member of fp to do async readv.
 
         And then read them using paramiko.readv(). paramiko.readv()
@@ -362,10 +506,16 @@
                 yield cur_offset_and_size[0], this_data
                 cur_offset_and_size = offset_stack.next()
 
+            # We read a coalesced entry, so mark it as done
+            cur_coalesced = None
             # Now that we've read all of the data for this coalesced section
             # on to the next
             cur_coalesced = cur_coalesced_stack.next()
 
+        if cur_coalesced is not None:
+            raise errors.ShortReadvError(relpath, cur_coalesced.start,
+                cur_coalesced.length, len(data))
+
     def put_file(self, relpath, f, mode=None):
         """
         Copy the file-like object into the location.
@@ -673,41 +823,6 @@
         # that we have taken the lock.
         return SFTPLock(relpath, self)
 
-    def _unparse_url(self, path=None):
-        if path is None:
-            path = self._path
-        path = urllib.quote(path)
-        # handle homedir paths
-        if not path.startswith('/'):
-            path = "/~/" + path
-        netloc = urllib.quote(self._host)
-        if self._username is not None:
-            netloc = '%s@%s' % (urllib.quote(self._username), netloc)
-        if self._port is not None:
-            netloc = '%s:%d' % (netloc, self._port)
-        return urlparse.urlunparse(('sftp', netloc, path, '', '', ''))
-
-    def _split_url(self, url):
-        (scheme, username, password, host, port, path) = split_url(url)
-        assert scheme == 'sftp'
-
-        # the initial slash should be removed from the path, and treated
-        # as a homedir relative path (the path begins with a double slash
-        # if it is absolute).
-        # see draft-ietf-secsh-scp-sftp-ssh-uri-03.txt
-        # RBC 20060118 we are not using this as its too user hostile. instead
-        # we are following lftp and using /~/foo to mean '~/foo'.
-        # handle homedir paths
-        if path.startswith('/~/'):
-            path = path[3:]
-        elif path == '/~':
-            path = ''
-        return (username, password, host, port, path)
-
-    def _parse_url(self, url):
-        (self._username, self._password,
-         self._host, self._port, self._path) = self._split_url(url)
-
     def _sftp_connect(self):
         """Connect to the remote sftp server.
         After this, self._sftp should have a valid connection (or
@@ -1004,7 +1119,6 @@
         # Re-import these as locals, so that they're still accessible during
         # interpreter shutdown (when all module globals get set to None, leading
         # to confusing errors like "'NoneType' object has no attribute 'error'".
-        import socket, errno
         class FakeChannel(object):
             def get_transport(self):
                 return self

=== added file 'bzrlib/transport/smart.py'
--- bzrlib/transport/smart.py	1970-01-01 00:00:00 +0000
+++ bzrlib/transport/smart.py	2006-09-15 23:48:16 +0000
@@ -0,0 +1,928 @@
+# Copyright (C) 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Smart-server protocol, client and server.
+
+Requests are sent as a command and list of arguments, followed by optional
+bulk body data.  Responses are similarly a response and list of arguments,
+followed by bulk body data. ::
+
+  SEP := '\001'
+    Fields are separated by Ctrl-A.
+  BULK_DATA := CHUNK+ TRAILER
+    Chunks can be repeated as many times as necessary.
+  CHUNK := CHUNK_LEN CHUNK_BODY
+  CHUNK_LEN := DIGIT+ NEWLINE
+    Gives the number of bytes in the following chunk.
+  CHUNK_BODY := BYTE[chunk_len]
+  TRAILER := SUCCESS_TRAILER | ERROR_TRAILER
+  SUCCESS_TRAILER := 'done' NEWLINE
+  ERROR_TRAILER := 
+
+Paths are passed across the network.  The client needs to see a namespace that
+includes any repository that might need to be referenced, and the client needs
+to know about a root directory beyond which it cannot ascend.
+
+Servers run over ssh will typically want to be able to access any path the user 
+can access.  Public servers on the other hand (which might be over http, ssh
+or tcp) will typically want to restrict access to only a particular directory 
+and its children, so will want to do a software virtual root at that level.
+In other words they'll want to rewrite incoming paths to be under that level
+(and prevent escaping using ../ tricks.)
+
+URLs that include ~ should probably be passed across to the server verbatim
+and the server can expand them.  This will proably not be meaningful when 
+limited to a directory?
+"""
+
+
+
+# TODO: A plain integer from query_version is too simple; should give some
+# capabilities too?
+
+# TODO: Server should probably catch exceptions within itself and send them
+# back across the network.  (But shouldn't catch KeyboardInterrupt etc)
+# Also needs to somehow report protocol errors like bad requests.  Need to
+# consider how we'll handle error reporting, e.g. if we get halfway through a
+# bulk transfer and then something goes wrong.
+
+# TODO: Standard marker at start of request/response lines?
+
+# TODO: Make each request and response self-validatable, e.g. with checksums.
+#
+# TODO: get/put objects could be changed to gradually read back the data as it
+# comes across the network
+#
+# TODO: What should the server do if it hits an error and has to terminate?
+#
+# TODO: is it useful to allow multiple chunks in the bulk data?
+#
+# TODO: If we get an exception during transmission of bulk data we can't just
+# emit the exception because it won't be seen.
+#   John proposes:  I think it would be worthwhile to have a header on each
+#   chunk, that indicates it is another chunk. Then you can send an 'error'
+#   chunk as long as you finish the previous chunk.
+#
+# TODO: Clone method on Transport; should work up towards parent directory;
+# unclear how this should be stored or communicated to the server... maybe
+# just pass it on all relevant requests?
+#
+# TODO: Better name than clone() for changing between directories.  How about
+# open_dir or change_dir or chdir?
+#
+# TODO: Is it really good to have the notion of current directory within the
+# connection?  Perhaps all Transports should factor out a common connection
+# from the thing that has the directory context?
+#
+# TODO: Pull more things common to sftp and ssh to a higher level.
+#
+# TODO: The server that manages a connection should be quite small and retain
+# minimum state because each of the requests are supposed to be stateless.
+# Then we can write another implementation that maps to http.
+#
+# TODO: What to do when a client connection is garbage collected?  Maybe just
+# abruptly drop the connection?
+#
+# TODO: Server in some cases will need to restrict access to files outside of
+# a particular root directory.  LocalTransport doesn't do anything to stop you
+# ascending above the base directory, so we need to prevent paths
+# containing '..' in either the server or transport layers.  (Also need to
+# consider what happens if someone creates a symlink pointing outside the 
+# directory tree...)
+#
+# TODO: Server should rebase absolute paths coming across the network to put
+# them under the virtual root, if one is in use.  LocalTransport currently
+# doesn't do that; if you give it an absolute path it just uses it.
+# 
+# XXX: Arguments can't contain newlines or ascii; possibly we should e.g.
+# urlescape them instead.  Indeed possibly this should just literally be
+# http-over-ssh.
+#
+# FIXME: This transport, with several others, has imperfect handling of paths
+# within urls.  It'd probably be better for ".." from a root to raise an error
+# rather than return the same directory as we do at present.
+#
+# TODO: Rather than working at the Transport layer we want a Branch,
+# Repository or BzrDir objects that talk to a server.
+#
+# TODO: Probably want some way for server commands to gradually produce body
+# data rather than passing it as a string; they could perhaps pass an
+# iterator-like callback that will gradually yield data; it probably needs a
+# close() method that will always be closed to do any necessary cleanup.
+#
+# TODO: Split the actual smart server from the ssh encoding of it.
+#
+# TODO: Perhaps support file-level readwrite operations over the transport
+# too.
+#
+# TODO: SmartBzrDir class, proxying all Branch etc methods across to another
+# branch doing file-level operations.
+#
+# TODO: jam 20060915 _decode_tuple is acting directly on input over
+#       the socket, and it assumes everything is UTF8 sections separated
+#       by \001. Which means a request like '\002' Will abort the connection
+#       because of a UnicodeDecodeError. It does look like invalid data will
+#       kill the SmartStreamServer, but only with an abort + exception, and 
+#       the overall server shouldn't die.
+
+from cStringIO import StringIO
+import errno
+import os
+import socket
+import sys
+import tempfile
+import threading
+import urllib
+import urlparse
+
+from bzrlib import (
+    bzrdir,
+    errors,
+    revision,
+    transport,
+    trace,
+    urlutils,
+    )
+from bzrlib.bundle.serializer import write_bundle
+from bzrlib.trace import mutter
+from bzrlib.transport import local
+
+# must do this otherwise urllib can't parse the urls properly :(
+for scheme in ['ssh', 'bzr', 'bzr+loopback', 'bzr+ssh']:
+    transport.register_urlparse_netloc_protocol(scheme)
+del scheme
+
+
+def _recv_tuple(from_file):
+    req_line = from_file.readline()
+    return _decode_tuple(req_line)
+
+
+def _decode_tuple(req_line):
+    if req_line == None or req_line == '':
+        return None
+    if req_line[-1] != '\n':
+        raise errors.SmartProtocolError("request %r not terminated" % req_line)
+    return tuple((a.decode('utf-8') for a in req_line[:-1].split('\x01')))
+
+
+def _send_tuple(to_file, args):
+    # XXX: this will be inefficient.  Just ask Robert.
+    to_file.write('\x01'.join((a.encode('utf-8') for a in args)) + '\n')
+    to_file.flush()
+
+
+class SmartProtocolBase(object):
+    """Methods common to client and server"""
+
+    def _send_bulk_data(self, body):
+        """Send chunked body data"""
+        assert isinstance(body, str)
+        self._out.write('%d\n' % len(body))
+        self._out.write(body)
+        self._out.write('done\n')
+        self._out.flush()
+
+    # TODO: this only actually accomodates a single block; possibly should support
+    # multiple chunks?
+    def _recv_bulk(self):
+        chunk_len = self._in.readline()
+        try:
+            chunk_len = int(chunk_len)
+        except ValueError:
+            raise errors.SmartProtocolError("bad chunk length line %r" % chunk_len)
+        bulk = self._in.read(chunk_len)
+        if len(bulk) != chunk_len:
+            raise errors.SmartProtocolError("short read fetching bulk data chunk")
+        self._recv_trailer()
+        return bulk
+
+    def _recv_tuple(self):
+        return _recv_tuple(self._in)
+
+    def _recv_trailer(self):
+        resp = self._recv_tuple()
+        if resp == ('done', ):
+            return
+        else:
+            self._translate_error(resp)
+
+
+class SmartStreamServer(SmartProtocolBase):
+    """Handles smart commands coming over a stream.
+
+    The stream may be a pipe connected to sshd, or a tcp socket, or an
+    in-process fifo for testing.
+
+    One instance is created for each connected client; it can serve multiple
+    requests in the lifetime of the connection.
+
+    The server passes requests through to an underlying backing transport, 
+    which will typically be a LocalTransport looking at the server's filesystem.
+    """
+
+    def __init__(self, in_file, out_file, backing_transport):
+        """Construct new server.
+
+        :param in_file: Python file from which requests can be read.
+        :param out_file: Python file to write responses.
+        :param backing_transport: Transport for the directory served.
+        """
+        self._in = in_file
+        self._out = out_file
+        self.smart_server = SmartServer(backing_transport)
+        # server can call back to us to get bulk data - this is not really
+        # ideal, they should get it per request instead
+        self.smart_server._recv_body = self._recv_bulk
+
+    def _recv_tuple(self):
+        """Read a request from the client and return as a tuple.
+        
+        Returns None at end of file (if the client closed the connection.)
+        """
+        return _recv_tuple(self._in)
+
+    def _send_tuple(self, args):
+        """Send response header"""
+        return _send_tuple(self._out, args)
+
+    def _send_error_and_disconnect(self, exception):
+        self._send_tuple(('error', str(exception)))
+        self._out.flush()
+        ## self._out.close()
+        ## self._in.close()
+
+    def _serve_one_request(self):
+        """Read one request from input, process, send back a response.
+        
+        :return: False if the server should terminate, otherwise None.
+        """
+        req_args = self._recv_tuple()
+        if req_args == None:
+            # client closed connection
+            return False  # shutdown server
+        try:
+            response = self.smart_server.dispatch_command(req_args[0], req_args[1:])
+            self._send_tuple(response.args)
+            if response.body is not None:
+                self._send_bulk_data(response.body)
+        except KeyboardInterrupt:
+            raise
+        except Exception, e:
+            # everything else: pass to client, flush, and quit
+            self._send_error_and_disconnect(e)
+            return False
+
+    def serve(self):
+        """Serve requests until the client disconnects."""
+        # Keep a reference to stderr because the sys module's globals get set to
+        # None during interpreter shutdown.
+        from sys import stderr
+        try:
+            while self._serve_one_request() != False:
+                pass
+        except Exception, e:
+            stderr.write("%s terminating on exception %s\n" % (self, e))
+            raise
+
+
+class SmartServerResponse(object):
+    """Response generated by SmartServer."""
+
+    def __init__(self, args, body=None):
+        self.args = args
+        self.body = body
+
+
+class SmartServer(object):
+    """Protocol logic for smart server.
+    
+    This doesn't handle serialization at all, it just processes requests and
+    creates responses.
+    """
+
+    # TODO: Better way of representing the body for commands that take it,
+    # and allow it to be streamed into the server.
+    
+    def __init__(self, backing_transport):
+        self._backing_transport = backing_transport
+        
+    def do_hello(self):
+        """Answer a version request with my version."""
+        return SmartServerResponse(('ok', '1'))
+
+    def do_has(self, relpath):
+        r = self._backing_transport.has(relpath) and 'yes' or 'no'
+        return SmartServerResponse((r,))
+
+    def do_get(self, relpath):
+        backing_bytes = self._backing_transport.get_bytes(relpath)
+        return SmartServerResponse(('ok',), backing_bytes)
+
+    def _deserialise_optional_mode(self, mode):
+        if mode == '':
+            return None
+        else:
+            return int(mode)
+
+    def do_append(self, relpath, mode):
+        old_length = self._backing_transport.append_bytes(
+            relpath, self._recv_body(), self._deserialise_optional_mode(mode))
+        return SmartServerResponse(('appended', '%d' % old_length))
+
+    def do_delete(self, relpath):
+        self._backing_transport.delete(relpath)
+
+    def do_iter_files_recursive(self, abspath):
+        # XXX: the path handling needs some thought.
+        #relpath = self._backing_transport.relpath(abspath)
+        transport = self._backing_transport.clone(abspath)
+        filenames = transport.iter_files_recursive()
+        return SmartServerResponse(('names',) + tuple(filenames))
+
+    def do_list_dir(self, relpath):
+        filenames = self._backing_transport.list_dir(relpath)
+        return SmartServerResponse(('names',) + tuple(filenames))
+
+    def do_mkdir(self, relpath, mode):
+        self._backing_transport.mkdir(relpath,
+                                      self._deserialise_optional_mode(mode))
+
+    def do_move(self, rel_from, rel_to):
+        self._backing_transport.move(rel_from, rel_to)
+
+    def do_put(self, relpath, mode):
+        self._backing_transport.put_bytes(relpath,
+                self._recv_body(),
+                self._deserialise_optional_mode(mode))
+
+    def do_rename(self, rel_from, rel_to):
+        self._backing_transport.rename(rel_from, rel_to)
+
+    def do_rmdir(self, relpath):
+        self._backing_transport.rmdir(relpath)
+
+    def do_stat(self, relpath):
+        stat = self._backing_transport.stat(relpath)
+        return SmartServerResponse(('stat', str(stat.st_size), oct(stat.st_mode)))
+        
+    def do_get_bundle(self, path, revision_id):
+        # open transport relative to our base
+        t = self._backing_transport.clone(path)
+        control, extra_path = bzrdir.BzrDir.open_containing_from_transport(t)
+        repo = control.open_repository()
+        tmpf = tempfile.TemporaryFile()
+        base_revision = revision.NULL_REVISION
+        write_bundle(repo, revision_id, base_revision, tmpf)
+        tmpf.seek(0)
+        return SmartServerResponse((), tmpf.read())
+
+    def dispatch_command(self, cmd, args):
+        func = getattr(self, 'do_' + cmd, None)
+        if func is None:
+            raise errors.SmartProtocolError("bad request %r" % (cmd,))
+        try:
+            result = func(*args)
+            if result is None: 
+                result = SmartServerResponse(('ok',))
+            return result
+        except errors.NoSuchFile, e:
+            return SmartServerResponse(('NoSuchFile', e.path))
+        except errors.FileExists, e:
+            return SmartServerResponse(('FileExists', e.path))
+        except errors.DirectoryNotEmpty, e:
+            return SmartServerResponse(('DirectoryNotEmpty', e.path))
+        except UnicodeError, e:
+            # If it is a DecodeError, than most likely we are starting
+            # with a plain string
+            str_or_unicode = e.object
+            if isinstance(str_or_unicode, unicode):
+                val = u'u:' + str_or_unicode
+            else:
+                val = u's:' + str_or_unicode.encode('base64')
+            # This handles UnicodeEncodeError or UnicodeDecodeError
+            return SmartServerResponse((e.__class__.__name__,
+                    e.encoding, val, str(e.start), str(e.end), e.reason))
+
+
+class SmartTCPServer(object):
+    """Listens on a TCP socket and accepts connections from smart clients"""
+
+    def __init__(self, backing_transport=None, host='127.0.0.1', port=0):
+        """Construct a new server.
+
+        To actually start it running, call either start_background_thread or
+        serve.
+
+        :param host: Name of the interface to listen on.
+        :param port: TCP port to listen on, or 0 to allocate a transient port.
+        """
+        if backing_transport is None:
+            backing_transport = memory.MemoryTransport()
+        self._server_socket = socket.socket()
+        self._server_socket.bind((host, port))
+        self.port = self._server_socket.getsockname()[1]
+        self._server_socket.listen(1)
+        self._server_socket.settimeout(1)
+        self.backing_transport = backing_transport
+
+    def serve(self):
+        # let connections timeout so that we get a chance to terminate
+        # Keep a reference to the exceptions we want to catch because the socket
+        # module's globals get set to None during interpreter shutdown.
+        from socket import timeout as socket_timeout
+        from socket import error as socket_error
+        self._should_terminate = False
+        while not self._should_terminate:
+            try:
+                self.accept_and_serve()
+            except socket_timeout:
+                # just check if we're asked to stop
+                pass
+            except socket_error, e:
+                trace.warning("client disconnected: %s", e)
+                pass
+
+    def get_url(self):
+        """Return the url of the server"""
+        return "bzr://%s:%d/" % self._server_socket.getsockname()
+
+    def accept_and_serve(self):
+        conn, client_addr = self._server_socket.accept()
+        conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
+        from_client = conn.makefile('r')
+        to_client = conn.makefile('w')
+        handler = SmartStreamServer(from_client, to_client,
+                self.backing_transport)
+        connection_thread = threading.Thread(None, handler.serve, name='smart-server-child')
+        connection_thread.setDaemon(True)
+        connection_thread.start()
+
+    def start_background_thread(self):
+        self._server_thread = threading.Thread(None,
+                self.serve,
+                name='server-' + self.get_url())
+        self._server_thread.setDaemon(True)
+        self._server_thread.start()
+
+    def stop_background_thread(self):
+        self._should_terminate = True
+        # self._server_socket.close()
+        # we used to join the thread, but it's not really necessary; it will
+        # terminate in time
+        ## self._server_thread.join()
+
+
+class SmartTCPServer_for_testing(SmartTCPServer):
+    """Server suitable for use by transport tests.
+    
+    This server is backed by the process's cwd.
+    """
+
+    def __init__(self):
+        self._homedir = os.getcwd()
+        # The server is set up by default like for ssh access: the client
+        # passes filesystem-absolute paths; therefore the server must look
+        # them up relative to the root directory.  it might be better to act
+        # a public server and have the server rewrite paths into the test
+        # directory.
+        SmartTCPServer.__init__(self, transport.get_transport("file:///"))
+        
+    def setUp(self):
+        """Set up server for testing"""
+        self.start_background_thread()
+
+    def tearDown(self):
+        self.stop_background_thread()
+
+    def get_url(self):
+        """Return the url of the server"""
+        host, port = self._server_socket.getsockname()
+        # XXX: I think this is likely to break on windows -- self._homedir will
+        # have backslashes (and maybe a drive letter?).
+        #  -- Andrew Bennetts, 2006-08-29
+        return "bzr://%s:%d%s" % (host, port, urlutils.escape(self._homedir))
+
+    def get_bogus_url(self):
+        """Return a URL which will fail to connect"""
+        return 'bzr://127.0.0.1:1/'
+
+
+class SmartStat(object):
+
+    def __init__(self, size, mode):
+        self.st_size = size
+        self.st_mode = mode
+
+
+class SmartTransport(transport.Transport):
+    """Connection to a smart server.
+
+    The connection holds references to pipes that can be used to send requests
+    to the server.
+
+    The connection has a notion of the current directory to which it's
+    connected; this is incorporated in filenames passed to the server.
+    
+    This supports some higher-level RPC operations and can also be treated 
+    like a Transport to do file-like operations.
+
+    The connection can be made over a tcp socket, or (in future) an ssh pipe
+    or a series of http requests.  There are concrete subclasses for each
+    type: SmartTCPTransport, etc.
+    """
+
+    def __init__(self, url, clone_from=None, client=None):
+        """Constructor.
+
+        :param client: ignored when clone_from is not None.
+        """
+        ### Technically super() here is faulty because Transport's __init__
+        ### fails to take 2 parameters, and if super were to choose a silly
+        ### initialisation order things would blow up. 
+        if not url.endswith('/'):
+            url += '/'
+        super(SmartTransport, self).__init__(url)
+        self._scheme, self._username, self._password, self._host, self._port, self._path = \
+                transport.split_url(url)
+        if clone_from is None:
+            if client is None:
+                self._client = SmartStreamClient(self._connect_to_server)
+            else:
+                self._client = client
+        else:
+            # credentials may be stripped from the base in some circumstances
+            # as yet to be clearly defined or documented, so copy them.
+            self._username = clone_from._username
+            # reuse same connection
+            self._client = clone_from._client
+
+    def abspath(self, relpath):
+        """Return the full url to the given relative path.
+        
+        @param relpath: the relative path or path components
+        @type relpath: str or list
+        """
+        return self._unparse_url(self._remote_path(relpath))
+    
+    def clone(self, relative_url):
+        """Make a new SmartTransport related to me, sharing the same connection.
+
+        This essentially opens a handle on a different remote directory.
+        """
+        if relative_url is None:
+            return self.__class__(self.base, self)
+        else:
+            return self.__class__(self.abspath(relative_url), self)
+
+    def is_readonly(self):
+        """Smart server transport can do read/write file operations."""
+        return False
+                                                   
+    def get_smart_client(self):
+        return self._client
+                                                   
+    def _unparse_url(self, path):
+        """Return URL for a path.
+
+        :see: SFTPUrlHandling._unparse_url
+        """
+        # TODO: Eventually it should be possible to unify this with
+        # SFTPUrlHandling._unparse_url?
+        if path == '':
+            path = '/'
+        path = urllib.quote(path)
+        netloc = urllib.quote(self._host)
+        if self._username is not None:
+            netloc = '%s@%s' % (urllib.quote(self._username), netloc)
+        if self._port is not None:
+            netloc = '%s:%d' % (netloc, self._port)
+        return urlparse.urlunparse((self._scheme, netloc, path, '', '', ''))
+
+    def _remote_path(self, relpath):
+        """Returns the Unicode version of the absolute path for relpath."""
+        return self._combine_paths(self._path, relpath)
+
+    def has(self, relpath):
+        """Indicate whether a remote file of the given name exists or not.
+
+        :see: Transport.has()
+        """
+        resp = self._client._call('has', self._remote_path(relpath))
+        if resp == ('yes', ):
+            return True
+        elif resp == ('no', ):
+            return False
+        else:
+            self._translate_error(resp)
+
+    def get(self, relpath):
+        """Return file-like object reading the contents of a remote file.
+        
+        :see: Transport.get_bytes()/get_file()
+        """
+        remote = self._remote_path(relpath)
+        resp = self._client._call('get', remote)
+        if resp != ('ok', ):
+            self._translate_error(resp, relpath)
+        return StringIO(self._client._recv_bulk())
+
+    def _serialise_optional_mode(self, mode):
+        if mode is None:
+            return ''
+        else:
+            return '%d' % mode
+
+    def mkdir(self, relpath, mode=None):
+        resp = self._client._call('mkdir', 
+                                  self._remote_path(relpath), 
+                                  self._serialise_optional_mode(mode))
+        self._translate_error(resp)
+
+    def put_file(self, relpath, upload_file, mode=None):
+        # its not ideal to seek back, but currently put_non_atomic_file depends
+        # on transports not reading before failing - which is a faulty
+        # assumption I think - RBC 20060915
+        pos = upload_file.tell()
+        try:
+            return self.put_bytes(relpath, upload_file.read(), mode)
+        except:
+            upload_file.seek(pos)
+            raise
+
+    def put_bytes(self, relpath, upload_contents, mode=None):
+        # FIXME: upload_file is probably not safe for non-ascii characters -
+        # should probably just pass all parameters as length-delimited
+        # strings?
+        resp = self._client._call_with_upload(
+            'put',
+            (self._remote_path(relpath), self._serialise_optional_mode(mode)),
+            upload_contents)
+        self._translate_error(resp)
+
+    def append_file(self, relpath, from_file, mode=None):
+        return self.append_bytes(relpath, from_file.read(), mode)
+        
+    def append_bytes(self, relpath, bytes, mode=None):
+        resp = self._client._call_with_upload(
+            'append',
+            (self._remote_path(relpath), self._serialise_optional_mode(mode)),
+            bytes)
+        if resp[0] == 'appended':
+            return int(resp[1])
+        self._translate_error(resp)
+
+    def delete(self, relpath):
+        resp = self._client._call('delete', self._remote_path(relpath))
+        self._translate_error(resp)
+
+    def rename(self, rel_from, rel_to):
+        self._call('rename', 
+                   self._remote_path(rel_from),
+                   self._remote_path(rel_to))
+
+    def move(self, rel_from, rel_to):
+        self._call('move', 
+                   self._remote_path(rel_from),
+                   self._remote_path(rel_to))
+
+    def rmdir(self, relpath):
+        resp = self._call('rmdir', self._remote_path(relpath))
+
+    def _call(self, method, *args):
+        resp = self._client._call(method, *args)
+        self._translate_error(resp)
+
+    def _translate_error(self, resp, orig_path=None):
+        """Raise an exception from a response"""
+        what = resp[0]
+        if what == 'ok':
+            return
+        elif what == 'NoSuchFile':
+            if orig_path is not None:
+                error_path = orig_path
+            else:
+                error_path = resp[1]
+            raise errors.NoSuchFile(error_path)
+        elif what == 'error':
+            raise errors.SmartProtocolError(unicode(resp[1]))
+        elif what == 'FileExists':
+            raise errors.FileExists(resp[1])
+        elif what == 'DirectoryNotEmpty':
+            raise errors.DirectoryNotEmpty(resp[1])
+        elif what in ('UnicodeEncodeError', 'UnicodeDecodeError'):
+            encoding = str(resp[1]) # encoding must always be a string
+            val = resp[2]
+            start = int(resp[3])
+            end = int(resp[4])
+            reason = str(resp[5]) # reason must always be a string
+            if val.startswith('u:'):
+                val = val[2:]
+            elif val.startswith('s:'):
+                val = val[2:].decode('base64')
+            if what == 'UnicodeDecodeError':
+                raise UnicodeDecodeError(encoding, val, start, end, reason)
+            elif what == 'UnicodeEncodeError':
+                raise UnicodeEncodeError(encoding, val, start, end, reason)
+        else:
+            raise errors.SmartProtocolError('unexpected smart server error: %r' % (resp,))
+
+    def _send_tuple(self, args):
+        self._client._send_tuple(args)
+
+    def _recv_tuple(self):
+        return self._client._recv_tuple()
+
+    def disconnect(self):
+        self._client.disconnect()
+
+    def delete_tree(self, relpath):
+        raise errors.TransportNotPossible('readonly transport')
+
+    def stat(self, relpath):
+        resp = self._client._call('stat', self._remote_path(relpath))
+        if resp[0] == 'stat':
+            return SmartStat(int(resp[1]), int(resp[2], 8))
+        else:
+            self._translate_error(resp)
+
+    ## def lock_read(self, relpath):
+    ##     """Lock the given file for shared (read) access.
+    ##     :return: A lock object, which should be passed to Transport.unlock()
+    ##     """
+    ##     # The old RemoteBranch ignore lock for reading, so we will
+    ##     # continue that tradition and return a bogus lock object.
+    ##     class BogusLock(object):
+    ##         def __init__(self, path):
+    ##             self.path = path
+    ##         def unlock(self):
+    ##             pass
+    ##     return BogusLock(relpath)
+
+    def listable(self):
+        return True
+
+    def list_dir(self, relpath):
+        resp = self._client._call('list_dir',
+                                  self._remote_path(relpath))
+        if resp[0] == 'names':
+            return [name.encode('ascii') for name in resp[1:]]
+        else:
+            self._translate_error(resp)
+
+    def iter_files_recursive(self):
+        resp = self._client._call('iter_files_recursive',
+                                  self._remote_path(''))
+        if resp[0] == 'names':
+            return resp[1:]
+        else:
+            self._translate_error(resp)
+
+
+class SmartStreamClient(SmartProtocolBase):
+    """Connection to smart server over two streams"""
+
+    def __init__(self, connect_func):
+        self._connect_func = connect_func
+        self._connected = False
+
+    def __del__(self):
+        self.disconnect()
+
+    def _ensure_connection(self):
+        if not self._connected:
+            self._in, self._out = self._connect_func()
+            self._connected = True
+
+    def _send_tuple(self, args):
+        self._ensure_connection()
+        _send_tuple(self._out, args)
+
+    def _send_bulk_data(self, body):
+        self._ensure_connection()
+        SmartProtocolBase._send_bulk_data(self, body)
+        
+    def _recv_bulk(self):
+        self._ensure_connection()
+        return SmartProtocolBase._recv_bulk(self)
+
+    def _recv_tuple(self):
+        self._ensure_connection()
+        return SmartProtocolBase._recv_tuple(self)
+
+    def _recv_trailer(self):
+        self._ensure_connection()
+        return SmartProtocolBase._recv_trailer(self)
+
+    def disconnect(self):
+        """Close connection to the server"""
+        if self._connected:
+            self._out.close()
+            self._in.close()
+
+    def _call(self, *args):
+        self._send_tuple(args)
+        return self._recv_tuple()
+
+    def _call_with_upload(self, method, args, body):
+        """Call an rpc, supplying bulk upload data.
+
+        :param method: method name to call
+        :param args: parameter args tuple
+        :param body: upload body as a byte string
+        """
+        self._send_tuple((method,) + args)
+        self._send_bulk_data(body)
+        return self._recv_tuple()
+
+    def query_version(self):
+        """Return protocol version number of the server."""
+        # XXX: should make sure it's empty
+        self._send_tuple(('hello',))
+        resp = self._recv_tuple()
+        if resp == ('ok', '1'):
+            return 1
+        else:
+            raise errors.SmartProtocolError("bad response %r" % (resp,))
+
+
+class SmartTCPTransport(SmartTransport):
+    """Connection to smart server over plain tcp"""
+
+    def __init__(self, url, clone_from=None):
+        super(SmartTCPTransport, self).__init__(url, clone_from)
+        try:
+            self._port = int(self._port)
+        except (ValueError, TypeError), e:
+            raise errors.InvalidURL(path=url, extra="invalid port %s" % self._port)
+        self._socket = None
+
+    def _connect_to_server(self):
+        self._socket = socket.socket()
+        self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
+        result = self._socket.connect_ex((self._host, int(self._port)))
+        if result:
+            raise errors.ConnectionError("failed to connect to %s:%d: %s" %
+                    (self._host, self._port, os.strerror(result)))
+        # TODO: May be more efficient to just treat them as sockets
+        # throughout?  But what about pipes to ssh?...
+        to_server = self._socket.makefile('w')
+        from_server = self._socket.makefile('r')
+        return from_server, to_server
+
+    def disconnect(self):
+        super(SmartTCPTransport, self).disconnect()
+        # XXX: Is closing the socket as well as closing the files really
+        # necessary?
+        if self._socket is not None:
+            self._socket.close()
+
+try:
+    from bzrlib.transport import sftp
+except errors.ParamikoNotPresent:
+    # no paramiko, no SSHTransport.
+    pass
+else:
+    class SmartSSHTransport(SmartTransport):
+        """Connection to smart server over SSH."""
+
+        def __init__(self, url, clone_from=None):
+            # TODO: all this probably belongs in the parent class.
+            super(SmartSSHTransport, self).__init__(url, clone_from)
+            try:
+                if self._port is not None:
+                    self._port = int(self._port)
+            except (ValueError, TypeError), e:
+                raise errors.InvalidURL(path=url, extra="invalid port %s" % self._port)
+
+        def _connect_to_server(self):
+            # XXX: don't hardcode vendor
+            # XXX: cannot pass password to SSHSubprocess yet
+            if self._password is not None:
+                raise errors.InvalidURL("SSH smart transport doesn't handle passwords")
+            self._ssh_connection = sftp.SSHSubprocess(self._host, 'openssh',
+                    port=self._port, user=self._username,
+                    command=['bzr', 'serve', '--inet'])
+            return self._ssh_connection.get_filelike_channels()
+
+        def disconnect(self):
+            super(SmartSSHTransport, self).disconnect()
+            self._ssh_connection.close()
+
+
+def get_test_permutations():
+    """Return (transport, server) permutations for testing"""
+    return [(SmartTCPTransport, SmartTCPServer_for_testing)]

=== modified file 'bzrlib/transport/ssh.py'
--- bzrlib/transport/ssh.py	2006-08-30 13:50:53 +0000
+++ bzrlib/transport/ssh.py	2006-09-13 12:44:46 +0000
@@ -28,6 +28,7 @@
 from bzrlib.errors import (ConnectionError,
                            ParamikoNotPresent,
                            TransportError,
+                           UnknownSSH,
                            )
 
 from bzrlib.osutils import pathjoin
@@ -101,7 +102,7 @@
     # A check for putty's plink or lsh would go here.
 
     mutter('falling back to paramiko implementation')
-    _ssh_vendor = ssh.ParamikoVendor()
+    _ssh_vendor = ParamikoVendor()
     return _ssh_vendor
 
 

=== modified file 'bzrlib/tree.py'
--- bzrlib/tree.py	2006-09-09 18:52:57 +0000
+++ bzrlib/tree.py	2006-09-15 01:21:38 +0000
@@ -119,6 +119,18 @@
     def id2path(self, file_id):
         return self.inventory.id2path(file_id)
 
+    def is_control_filename(self, filename):
+        """True if filename is the name of a control file in this tree.
+        
+        :param filename: A filename within the tree. This is a relative path
+        from the root of this tree.
+
+        This is true IF and ONLY IF the filename is part of the meta data
+        that bzr controls in this tree. I.E. a random .bzr directory placed
+        on disk will not be a control file for this tree.
+        """
+        return self.bzrdir.is_control_filename(filename)
+
     def iter_entries_by_dir(self):
         """Walk the tree in 'by_dir' order.
 
@@ -135,6 +147,10 @@
     def _get_inventory(self):
         return self._inventory
     
+    def get_file(self, file_id):
+        """Return a file object for the file file_id in the tree."""
+        raise NotImplementedError(self.get_file)
+    
     def get_file_by_path(self, path):
         return self.get_file(self._inventory.path2id(path))
 
@@ -160,6 +176,9 @@
                      "file is actually %s" % fp['sha1'],
                      "store is probably damaged/corrupt"])
 
+    def path2id(self, path):
+        """Return the id for path in this tree."""
+        return self._inventory.path2id(path)
 
     def print_file(self, file_id):
         """Print file with id `file_id` to stdout."""

=== added file 'bzrlib/treebuilder.py'
--- bzrlib/treebuilder.py	1970-01-01 00:00:00 +0000
+++ bzrlib/treebuilder.py	2006-09-15 02:03:15 +0000
@@ -0,0 +1,76 @@
+# Copyright (C) 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""TreeBuilder helper class.
+
+TreeBuilders are used to build trees of various shapres or properties. This 
+can be extremely useful in testing for instance.
+"""
+
+from bzrlib import errors
+
+
+class TreeBuilder(object):
+    """A TreeBuilder allows the creation of specific content in one tree at a
+    time.
+    """
+
+    def __init__(self):
+        """Construct a TreeBuilder."""
+        self._tree = None
+
+    def build(self, recipe):
+        """Build recipe into the current tree.
+
+        :param recipe: A sequence of paths. For each path, the corresponding
+            path in the current tree is created and added. If the path ends in
+            '/' then a directory is added, otherwise a regular file is added.
+        """
+        self._ensure_building()
+        for name in recipe:
+            assert isinstance(name, basestring)
+            if name[-1] == '/':
+                self._tree.mkdir(name[:-1])
+            else:
+                end = '\n'
+                content = "contents of %s%s" % (name.encode('utf-8'), end)
+                self._tree.add(name, None, 'file')
+                file_id = self._tree.path2id(name)
+                self._tree.put_file_bytes_non_atomic(file_id, content)
+
+    def _ensure_building(self):
+        """Raise NotBuilding if there is no current tree being built."""
+        if self._tree is None:
+            raise errors.NotBuilding
+            
+    def finish_tree(self):
+        """Finish building the current tree."""
+        self._ensure_building()
+        tree = self._tree
+        self._tree = None
+        tree.unlock()
+
+    def start_tree(self, tree):
+        """Start building on tree.
+        
+        :param tree: A tree to start building on. It must provide the 
+            MutableTree interface.
+        """
+        if self._tree is not None:
+            raise errors.AlreadyBuilding
+        self._tree = tree
+        self._tree.lock_tree_write()
+

=== modified file 'bzrlib/urlutils.py'
--- bzrlib/urlutils.py	2006-07-30 06:52:39 +0000
+++ bzrlib/urlutils.py	2006-09-16 06:44:47 +0000
@@ -129,6 +129,8 @@
         if m:
             # Absolute URL
             scheme = m.group('scheme')
+            # this skips .. normalisation, making http://host/../../..
+            # be rather strange.
             path = m.group('path').split('/')
         else:
             for chunk in arg.split('/'):

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2006-09-10 23:56:57 +0000
+++ bzrlib/workingtree.py	2006-09-15 02:03:15 +0000
@@ -71,6 +71,8 @@
 from bzrlib.lockable_files import LockableFiles, TransportLock
 from bzrlib.lockdir import LockDir
 from bzrlib.merge import merge_inner, transform_tree
+import bzrlib.mutabletree
+from bzrlib.mutabletree import needs_tree_write_lock
 from bzrlib.osutils import (
                             abspath,
                             compact_date,
@@ -90,6 +92,7 @@
                             )
 from bzrlib.progress import DummyProgress, ProgressPhase
 from bzrlib.revision import NULL_REVISION
+import bzrlib.revisiontree
 from bzrlib.rio import RioReader, rio_file, Stanza
 from bzrlib.symbol_versioning import (deprecated_passed,
         deprecated_method,
@@ -103,7 +106,6 @@
 from bzrlib.transport import get_transport
 from bzrlib.transport.local import LocalTransport
 from bzrlib.textui import show_status
-import bzrlib.tree
 import bzrlib.ui
 import bzrlib.xml5
 
@@ -159,19 +161,6 @@
     return gen_file_id('TREE_ROOT')
 
 
-def needs_tree_write_lock(unbound):
-    """Decorate unbound to take out and release a tree_write lock."""
-    def tree_write_locked(self, *args, **kwargs):
-        self.lock_tree_write()
-        try:
-            return unbound(self, *args, **kwargs)
-        finally:
-            self.unlock()
-    tree_write_locked.__doc__ = unbound.__doc__
-    tree_write_locked.__name__ = unbound.__name__
-    return tree_write_locked
-
-
 class TreeEntry(object):
     """An entry that implements the minimum interface used by commands.
 
@@ -227,7 +216,7 @@
         return ''
 
 
-class WorkingTree(bzrlib.tree.Tree):
+class WorkingTree(bzrlib.mutabletree.MutableTree):
     """Working copy tree.
 
     The inventory is held in the `Branch` working-inventory, and the
@@ -338,19 +327,6 @@
     def _set_inventory(self, inv):
         assert inv.root is not None
         self._inventory = inv
-        self.path2id = self._inventory.path2id
-
-    def is_control_filename(self, filename):
-        """True if filename is the name of a control file in this tree.
-        
-        :param filename: A filename within the tree. This is a relative path
-        from the root of this tree.
-
-        This is true IF and ONLY IF the filename is part of the meta data
-        that bzr controls in this tree. I.E. a random .bzr directory placed
-        on disk will not be a control file for this tree.
-        """
-        return self.bzrdir.is_control_filename(filename)
 
     @staticmethod
     def open(path=None, _unsupported=False):
@@ -564,22 +540,6 @@
             transform_tree(tree, self)
             tree.set_parent_ids([revision_id])
 
-    @needs_write_lock
-    def commit(self, message=None, revprops=None, *args, **kwargs):
-        # avoid circular imports
-        from bzrlib.commit import Commit
-        if revprops is None:
-            revprops = {}
-        if not 'branch-nick' in revprops:
-            revprops['branch-nick'] = self.branch.nick
-        # args for wt.commit start at message from the Commit.commit method,
-        # but with branch a kwarg now, passing in args as is results in the
-        #message being used for the branch
-        args = (DEPRECATED_PARAMETER, message, ) + args
-        committed_id = Commit().commit( working_tree=self, revprops=revprops,
-            *args, **kwargs)
-        return committed_id
-
     def id2abspath(self, file_id):
         return self.abspath(self.id2path(file_id))
 
@@ -622,68 +582,35 @@
             mode = os.lstat(self.abspath(path)).st_mode
             return bool(stat.S_ISREG(mode) and stat.S_IEXEC & mode)
 
-    @needs_tree_write_lock
-    def add(self, files, ids=None):
-        """Make files versioned.
-
-        Note that the command line normally calls smart_add instead,
-        which can automatically recurse.
-
-        This adds the files to the inventory, so that they will be
-        recorded by the next commit.
-
-        files
-            List of paths to add, relative to the base of the tree.
-
-        ids
-            If set, use these instead of automatically generated ids.
-            Must be the same length as the list of files, but may
-            contain None for ids that are to be autogenerated.
-
-        TODO: Perhaps have an option to add the ids even if the files do
-              not (yet) exist.
-
-        TODO: Perhaps callback with the ids and paths as they're added.
-        """
+    @needs_write_lock
+    def _add(self, files, ids, kinds):
+        """See MutableTree._add."""
         # TODO: Re-adding a file that is removed in the working copy
         # should probably put it back with the previous ID.
-        if isinstance(files, basestring):
-            assert(ids is None or isinstance(ids, basestring))
-            files = [files]
-            if ids is not None:
-                ids = [ids]
-
-        if ids is None:
-            ids = [None] * len(files)
-        else:
-            assert(len(ids) == len(files))
-
+        # the read and write working inventory should not occur in this 
+        # function - they should be part of lock_write and unlock.
         inv = self.read_working_inventory()
-        for f,file_id in zip(files, ids):
-            if self.is_control_filename(f):
-                raise errors.ForbiddenControlFileError(filename=f)
-
-            fp = splitpath(f)
-
-            if len(fp) == 0:
-                raise BzrError("cannot add top-level %r" % f)
-
-            fullpath = normpath(self.abspath(f))
-            try:
-                kind = file_kind(fullpath)
-            except OSError, e:
-                if e.errno == errno.ENOENT:
-                    raise NoSuchFile(fullpath)
-            if not InventoryEntry.versionable_kind(kind):
-                raise errors.BadFileKindError(filename=f, kind=kind)
+        for f, file_id, kind in zip(files, ids, kinds):
+            assert kind is not None
             if file_id is None:
                 inv.add_path(f, kind=kind)
             else:
                 inv.add_path(f, kind=kind, file_id=file_id)
-
         self._write_inventory(inv)
 
     @needs_tree_write_lock
+    def _gather_kinds(self, files, kinds):
+        """See MutableTree._gather_kinds."""
+        for pos, f in enumerate(files):
+            if kinds[pos] is None:
+                fullpath = normpath(self.abspath(f))
+                try:
+                    kinds[pos] = file_kind(fullpath)
+                except OSError, e:
+                    if e.errno == errno.ENOENT:
+                        raise NoSuchFile(fullpath)
+
+    @needs_write_lock
     def add_parent_tree_id(self, revision_id, allow_leftmost_as_ghost=False):
         """Add revision_id as a parent.
 
@@ -774,12 +701,7 @@
 
     @needs_tree_write_lock
     def set_parent_trees(self, parents_list, allow_leftmost_as_ghost=False):
-        """Set the parents of the working tree.
-
-        :param parents_list: A list of (revision_id, tree) tuples. 
-            If tree is None, then that element is treated as an unreachable
-            parent tree - i.e. a ghost.
-        """
+        """See MutableTree.set_parent_trees."""
         # parent trees are not used in current format trees, delegate to
         # set_parent_ids
         self.set_parent_ids([rev for (rev, tree) in parents_list],
@@ -868,6 +790,15 @@
                 merge_hashes[file_id] = hash
         return merge_hashes
 
+    @needs_write_lock
+    def mkdir(self, path, file_id=None):
+        """See MutableTree.mkdir()."""
+        if file_id is None:
+            file_id = gen_file_id(os.path.basename(path))
+        os.mkdir(self.abspath(path))
+        self.add(path, file_id, 'directory')
+        return file_id
+
     def get_symlink_target(self, file_id):
         return os.readlink(self.id2abspath(file_id))
 
@@ -1204,6 +1135,16 @@
             source.unlock()
             top_pb.finished()
 
+    @needs_write_lock
+    def put_file_bytes_non_atomic(self, file_id, bytes):
+        """See MutableTree.put_file_bytes_non_atomic."""
+        stream = file(self.id2abspath(file_id), 'wb')
+        try:
+            stream.write(bytes)
+        finally:
+            stream.close()
+        # TODO: update the hashcache here ?
+
     def extras(self):
         """Yield all unknown files in this WorkingTree.
 
@@ -1370,14 +1311,12 @@
         return file_kind(self.id2abspath(file_id))
 
     def last_revision(self):
-        """Return the last revision id of this working tree.
-
-        In early branch formats this was the same as the branch last_revision,
-        but that cannot be relied upon - for working tree operations,
-        always use tree.last_revision(). This returns the left most parent id,
-        or None if there are no parents.
-
-        This was deprecated as of 0.11. Please use get_parent_ids instead.
+        """Return the last revision of the branch for this tree.
+
+        This format tree does not support a separate marker for last-revision
+        compared to the branch.
+
+        See MutableTree.last_revision
         """
         return self._last_revision()
 
@@ -1399,14 +1338,7 @@
             raise
 
     def lock_tree_write(self):
-        """Lock the working tree for write, and the branch for read.
-
-        This is useful for operations which only need to mutate the working
-        tree. Taking out branch write locks is a relatively expensive process
-        and may fail if the branch is on read only media. So branch write locks
-        should only be taken out when we are modifying branch data - such as in
-        operations like commit, pull, uncommit and update.
-        """
+        """See MutableTree.lock_tree_write, and WorkingTree.unlock."""
         self.branch.lock_read()
         try:
             return self._control_files.lock_write()
@@ -1415,7 +1347,7 @@
             raise
 
     def lock_write(self):
-        """See Branch.lock_write, and WorkingTree.unlock."""
+        """See MutableTree.lock_write, and WorkingTree.unlock."""
         self.branch.lock_write()
         try:
             return self._control_files.lock_write()
@@ -1775,7 +1707,7 @@
 
     @needs_read_lock
     def _last_revision(self):
-        """See WorkingTree._last_revision."""
+        """See Mutable.last_revision."""
         try:
             return self._control_files.get_utf8('last-revision').read()
         except NoSuchFile:

=== modified file 'doc/index.txt'
--- doc/index.txt	2006-07-30 06:52:39 +0000
+++ doc/index.txt	2006-09-13 04:48:33 +0000
@@ -19,30 +19,34 @@
 * `Configuration <configuration.htm>`_ 
   (wiki page: http://bazaar-vcs.org/ConfiguringBzr)
 
-    Information on how to configure Bazaar
+    Information on how to configure Bazaar.
 
 * `Setting up email <setting_up_email.htm>`_
   (wiki page: http://bazaar-vcs.org/BzrSettingEmail)
 
     The various ways that exist to tell Bazaar your email address for
-    commits
+    commits.
 
 * `Plugins <plugins.htm>`_
   (wiki page: http://bazaar-vcs.org/BzrPlugins)
     
-    Information on how to use plugins in Bazaar
+    Information on how to use plugins in Bazaar.
 
 * `Specifying revision <specifying_revisions.htm>`_
   (wiki page: http://bazaar-vcs.org/BzrRevisionSpec)
 
     Description of the subarguments for ``-r`` as used by commands such as
-    **bzr log** and **bzr merge**
+    **bzr log** and **bzr merge**.
 
 * `Using aliases <using_aliases.htm>`_
   (wiki page: http://bazaar-vcs.org/CommandAliases)
 
     How to setup and use command aliases.
 
+* `Running a Bazaar server <server.htm>`_
+  
+    How to run a server to allow remote access to Bazaar branches.
+
 This document automatically created based on output of **bzr help**:
 
 * `Man page (help for bzr commands) <bzr_man.htm>`_

=== added file 'doc/server.txt'
--- doc/server.txt	1970-01-01 00:00:00 +0000
+++ doc/server.txt	2006-09-13 04:48:33 +0000
@@ -0,0 +1,72 @@
+=======================
+Running a Bazaar Server
+=======================
+
+Bazaar does not require a specialised server because it operates over HTTP, FTP
+or SFTP.  There is an optional smart server that can be invoked over SSH, from
+inetd, or in a dedicated mode.
+
+Dumb servers
+============
+
+We describe HTTP, FTP, SFTP and HTTP-WebDAV as "dumb" servers because they do
+not offer any assistance to Bazaar.  If you make a Bazaar repository available
+over any of these protocols, Bazaar will allow you to read it remotely.  Just
+enter the URL to the branch in the Bazaar command you are running.::
+
+    bzr log http://bazaar-vcs.org/bzr/bzr.dev
+
+Bazaar supports writing over FTP, SFTP and via a plugin over HTTP-WebDAV.
+
+High performance server
+=======================
+
+**In development**
+
+The high performance server is currently in development.  The version of Bazaar
+that accompanies this documentation is able to use the servers underlying
+protocol as a dumb server.  This is the first stage high performance server
+functionality to be delivered.
+
+The current high performance server **always** provides **read/write** access.
+It is not suitable for publishing of read-only mirrors.
+
+It can be configured to run in three different ways:
+
+SSH
+---
+
+Using Bazaar over SSH requires no special configuration on the server::
+
+    bzr log bzr+ssh://host/~/path/to/branch --remote-bzr-path="~/bin/bzr"
+
+inetd
+-----
+
+This example shows how to run `bzr` with a dedicated user `bzruser` for a shared
+repository in `/srv/bzr/repo` which has a branch at `/srv/bzr/repo/branchname`
+
+Running a Bazaar server from inetd requires an inetd.conf entry::
+
+    1234  stream  tcp  nowait  bzruser  /usr/bin/bzr serve --inet --directory=/srv/bzr/repo
+
+When running client commands, the URL you supply is a `bzr://` URL relative to
+the `--directory` option given in inetd.conf::
+
+    bzr log bzr://host:1234/branchname
+
+Dedicated
+---------
+
+This mode has the same path and URL behaviour as the inetd mode.  To run as a
+specific, you should use `su` or login as that user.  This example runs bzr
+serve on `localhost` port `1234`.
+
+server::
+
+    bzr serve --port=localhost:1234 --directory=/srv/bzr/repo
+    
+client::
+
+    bzr log bzr://host:1234/branchname
+

=== modified file 'profile_imports.py'
--- profile_imports.py	2006-06-19 14:40:19 +0000
+++ profile_imports.py	2006-09-12 22:02:02 +0000
@@ -115,12 +115,15 @@
     frame = sys._getframe(1)
     frame_name = frame.f_globals.get('__name__', '<unknown>')
     extra = ''
-    cur_frame = 1
     if frame_name.endswith('demandload'):
         # If this was demandloaded, we have 3 frames to ignore
         extra = ' (demandload)'
         frame = sys._getframe(4)
-        cur_frame = 4
+        frame_name = frame.f_globals.get('__name__', '<unknown>')
+    elif frame_name.endswith('lazy_import'):
+        # If this was lazily imported, we have 3 frames to ignore
+        extra = ' (lazy)'
+        frame = sys._getframe(4)
         frame_name = frame.f_globals.get('__name__', '<unknown>')
     frame_lineno = frame.f_lineno
 

